//---------------------------------------------------------------------------
// $Id: cssame.cc,v 1.1 1998/06/03 16:52:43 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>

#define _MODULE_ "libcssame.a"

#include <values.h>
#include <ctype.h>
#include <stdio.h>

#include <suif.h>
#include <par.h>
#include <odc-util.h>
#include <pfg.h>

#include "cssame.h"

//---------------------------------------------------------------------------
// cssame::buildCSSAME()
// Algorithm to build concurrent factored use-def chains. FUD chains have
// the property that each use of a variable is reached by a single
// defintion: the use-def chain for any one use is a single link to the
// unique reaching definition. 
// 
// Phi-terms are inserted into the program when there are multiple reaching
// definitions to one place. These phi-terms are pseudo-assignments that
// factor the multiple incoming reaching definitions.
//
// FUD chains are an interpretation of the static single assignment (SSA)
// form of the program. The difference is that FUD chains do not require
// the renaming of variables to satisfy the single-assignment restriction.
//---------------------------------------------------------------------------
void
cssame::buildCSSAME()
{
    _graph->find_dominators();
    _graph->find_postdominators();
    _graph->find_df();

    /* Find natural loops. Needed by cssame::computeOrdering to cluster all
     * the nodes in a loop in one big super-node. All the nodes in this
     * super-node will have the same precedence information than the
     * header. This makes the analysis less accurate but avoids computing
     * wrong precedence for nodes in a loop.
     */
    _graph->find_natural_loops();

    this->computeOrdering();
    StrictInterval_list *S = this->findMutexBodies();
    this->placePhiFunctions();
    this->placeFUDChains();
    this->placePiFunctions();
    this->rewritePiFunctions();
}


//---------------------------------------------------------------------------
// cssame::computeReachingDefs()
// Computes reaching definitions for all the variable uses in the program.
// Reaching definition lists are appended to each use and to the statement
// that contains it.
//---------------------------------------------------------------------------
void
cssame::computeReachingDefs()
{
    VarRef_list_iter ref_iter(_graph->refs());
    while (!ref_iter.is_empty()) {
	VarRef *ref = ref_iter.step();

	/* Set Marked(d) = NULL for every reference 'd' in the graph */
	ref->markedBy(NULL);

	/* Empty current reaching definition information from the reference
	 * and from the statement containing it.
	 */
	VarRef_list *empty_rdefs = new VarRef_list;
	assert(empty_rdefs);
	ref->set_reachingDefs(empty_rdefs);
	ref->set_reachedUses(empty_rdefs);

	if (ref->stmt()) {
	    instruction *instr = ref->stmt()->instr();
	    VarRef_list *stmt_rdefs;
	    
	    stmt_rdefs = (VarRef_list *)instr->peek_annote(k_reaching_defs);
	    if (stmt_rdefs == NULL) {
		stmt_rdefs = new VarRef_list;
		assert(stmt_rdefs);
		instr->append_annote(k_reaching_defs, (void *)stmt_rdefs);
	    }
	    stmt_rdefs->erase();

	    stmt_rdefs = (VarRef_list *)instr->peek_annote(k_reached_uses);
	    if (stmt_rdefs == NULL) {
		stmt_rdefs = new VarRef_list;
		assert(stmt_rdefs);
		instr->append_annote(k_reached_uses, (void *)stmt_rdefs);
	    }
	    stmt_rdefs->erase();
	}
    }

    /* Traverse all the uses following their use-def chains looking for
     * reaching definitions.
     */
    ref_iter.reset(_graph->refs());
    while (!ref_iter.is_empty()) {
	VarRef *ref = ref_iter.step();
	if (ref->isU()) {
	    this->followChain(ref->chain(), ref);
	}
    }
}


//---------------------------------------------------------------------------
// cssame::followChain()
// Follows the factored use-def links to find all possible reaching
// definitions for the given use. Adapted for explicitly parallel programs
// from "High Performance Compilers for Parallel Computing", M. Wolfe, page
// 182. The steps of the algorithm are included in the comments.
//
//	if Marked(d) = u then return endif
//	Marked(d) = u
//	if d is a definition for u then
//	    add d to Defs(u)
//	endif
//	if d is a phi-term then
//	    for each link j do
//		followChain(phi-chain(d)[j], u)
//	    endfor
//	else if d is a killing definition then
//	    /* stop here */
//	else /* follow def-def link */
//	    followChain(Chain(d), u)
//	endif
//
//---------------------------------------------------------------------------
void
cssame::followChain(VarRef *d, VarRef *u)
{
    static char *__FNAME = "cssame::followChain";

    if (d == NULL) {
	return;
    }

    assert(d->isD() || d->isPHI() || d->isPI());
    assert(u->isU());

    /* if Marked(d) = u then return endif */
    if (d->isMarkedBy(u)) {
	return;
    }

    /* Marked(d) = u */
    d->markedBy(u);

    /* if d is a definition for u then
     *    add d to Defs(u)
     * endif
     */
    if (d->isD() && d->var()->sym_id() == u->var()->sym_id()) {
	/* Add d to the list of reaching defs of u */
	u->reachingDefs()->append(d);

	/* Add u to the list of reached uses of d. This adds def-use edges
	 * to the graph.
	 */
	d->reachedUses()->append(u);

	/* Also add the reaching definition to the list of reaching
	 * definitions for the statement containing the reference.
	 */
	assert(u->stmt());
	instruction *instr = u->stmt()->instr();
	VarRef_list *stmt_rdefs;
	stmt_rdefs = (VarRef_list *)instr->peek_annote(k_reaching_defs);
	assert(stmt_rdefs);
	stmt_rdefs->append(d);

	/* Add the use to the list of reached uses for the statement
	 * containing the reference
	 */
	stmt_rdefs = (VarRef_list *)instr->peek_annote(k_reached_uses);
	assert(stmt_rdefs);
	stmt_rdefs->append(u);

	if (debug(__FNAME, 2)) {
	    printf("\n%s(): Added '", __FNAME);
	    d->print();
	    printf("' to list of reaching defs for\n");
	    printf("use ");
	    u->print();
	    printf("\nand instruction\n");
	    instr->print();
	    printf("\n");
	}
    }

    /* if d is a phi-term then
     *     for each link j do
     *        FollowChain(phi-chain(d)[j], u)
     *     endfor
     * else if d is a killing definition then
     *     stop here
     * else (follow a def-def link)
     *     FollowChain(Chain(d), u)
     * endif
     */
    if (d->isPHI() || d->isPI()) {
	VarRef_list_iter iter(((PhiTerm *)d)->phi_chain());

	while (iter.is_empty() == FALSE) {
	    VarRef *arg = iter.step();
	    this->followChain(arg, u);
	}
    } else {
	/* We do nothing. We should add extra cases for def-def links, but
	 * we still don't have them in place.
	 */
    }
}


//---------------------------------------------------------------------------
// cssame::print()
// Prints the original program in pseudo-SSA form. Only phi and pi terms
// are shown. Variables are not actually renumbered. This assumes that FUD
// chains have already been built.
//---------------------------------------------------------------------------
void
cssame::print(FILE *f)
{
    /* First sort the Phi and Pi terms in the program according to their line
     * number.
     */
    VarRef_list *sortedTerms = this->sortPhiPi();

    /* Once sorted we output the original lines from the source file. When
     * we find a phi or pi term for the current line we print the term(s)
     * that go in that line before the source line.
     */
    char *fname = source_file_name(_graph->tproc());
    int start = 1, end;
    VarRef_list_iter iter(sortedTerms);

    while (iter.is_empty() == FALSE) {
	VarRef *cref = iter.step();
	end = cref->lineno();
	char *lastline = displayFile(fname, start, end - 1, DF_LINENO, f);

	/* Only print phi and pi terms for user defined variables */
	if (cref->var()->is_userdef()) {
	    fprintf(f, "%7d ", end);
	    while (isspace(*lastline)) {
		putchar(*lastline);
		lastline++;
	    }
	    cref->print();
	    fprintf(f, ";\n");
	}

	start = end;
    }
    displayFile(fname, start, MAXINT, DF_LINENO, f);
}


//***************************************************************************
//			     Internal methods
//***************************************************************************
//---------------------------------------------------------------------------
// cssame::computeOrdering()
// Computes guaranteed execution ordering for all the nodes in the graph.
// Based on 'Concurrent Static Single Assignment Form and Constant
// Propagation for Explicitly Parallel Programs', Padua, Midkiff & Lee
// LCPC '97.
//---------------------------------------------------------------------------
void
cssame::computeOrdering()
{
    static char *__FNAME = "cssame::computeOrdering";

    /* Initialize the _prec, _prec_ct and _prec_sy bitvector sets in every
     * node of the graph.
     */
    unsigned num = _graph->num_nodes();
    unsigned i;
    for (i = 0; i < num; i++) {
	_graph->node(i)->_prec.expand(0, num);
	_graph->node(i)->_prec_ct.expand(0, num);
	_graph->node(i)->_prec_sy.expand(0, num);
    }

    /* Create a new work queue and initialize it with the successors to the
     * entry node.
     */
    pfg_node_list *queue = new pfg_node_list;
    assert(queue);
    pfg_node_list_iter node_iter(_graph->entry_node()->succs());
    while (!node_iter.is_empty()) {
	queue->append(node_iter.step());
    }

    /* while Q is not empty do
     *     n = the first entry in Q
     *     prec_old = prec(n)
     *     if Ntype(n) is Coend then
     *         prec_ct(n) = Union{(m,n) in Ect} (Prec(m) U {n})
     *     else
     *         prec_ct(n) = Intersection{(m, n) in Ect} (Prec(m) U {n})
     *     endif
     *     prec_sy(n) = Intersection{(m, n) in Esy} (Prec(m) U {n})
     *     prec(n) = prec_ct(n) U prec_sy(n)
     *     if prec_old <> prec(n) then
     *         Put control flow and synchronization successors of n in Q
     *     endif
     * endwhile
     */
    while (!queue->is_empty()) {
	pfg_node *n = queue->pop();

	if (debug(__FNAME, 2)) {
	    printf("\n%s(): Computing Prec(%d) -> loop depth: %d\n", 
		    __FNAME, n->number(), _graph->loop_depth(n));
	}

	bit_set prec_old(0, num);
	prec_old.copy(&(n->_prec));

	/* if Ntype(n) is Coend then
	 *     prec_ct(n) = Union{(m,n) in Ect} (Prec(m) U {n})
	 * else
	 *     prec_ct(n) = Intersection{(m, n) in Ect} (Prec(m) U {n})
	 * endif
	 */
	pfg_node_list_iter preds_iter(n->preds());

	if (preds_iter.is_empty() == FALSE) {
	    /* If n is coend we initialize its _prec_ct set to empty
	     * because we will be adding elements. Otherwise we initialize
	     * it to all ones because we will be intersecting elements.
	     */
	    if (n->is_parend()) {
		n->_prec_ct.clear();
	    } else {
		n->_prec_ct.universal();
	    }

	    /* Node n always goes in its own _prec_ct set */
	    n->_prec_ct.add(n->number());
	}

	while (preds_iter.is_empty() == FALSE) {
	    pfg_node *pred = preds_iter.step();

	    /* If n is a loop header we must avoid processing any
	     * predecessor for n inside the body of the loop (ie, the
	     * last node of the body). Therefore, if the predecessor
	     * pred has the same loop_depth than n and n is a loop
	     * header, we ignore the predecessor.
	     */
	    if (_graph->is_loop_begin(n) && 
		_graph->loop_depth(n) == _graph->loop_depth(pred)) {
		continue;
	    }

	    bit_set ancestor_prec(0, num);
	    ancestor_prec.copy(&(pred->_prec));
	    ancestor_prec.add(n->number());

	    if (n->is_parend()) {
		n->_prec_ct += ancestor_prec;
		if (debug(__FNAME, 2)) {
		    printf("\tAdded _prec(%d) to _prec_ct(%d): ",
			    pred->number(), n->number());
		    n->_prec_ct.print(stdout, "%d ");
		    printf("\n");
		}
	    } else {
		n->_prec_ct *= ancestor_prec;
		if (debug(__FNAME, 2)) {
		    printf("\tIntersected _prec(%d) to _prec_ct(%d): ",
			    pred->number(), n->number());
		    n->_prec_ct.print(stdout, "%d ");
		    printf("\n");
		}
	    }
	}

	/* prec_sy(n) = Intersection{(m, n) in Esy} (Prec(m) U {n}) */
	preds_iter.reset(n->spreds());
	if (preds_iter.is_empty() == FALSE) {
	    n->_prec_sy.universal();
	}
	while (preds_iter.is_empty() == FALSE) {
	    pfg_node *pred = preds_iter.step();

	    bit_set ancestor_prec(0, num);
	    ancestor_prec.copy(&(pred->_prec));
	    ancestor_prec.add(n->number());

	    n->_prec_sy *= ancestor_prec;

	    if (debug(__FNAME, 2)) {
		printf("\tIntersected _prec(%d) to _prec_sy(%d): ",
			pred->number(), n->number());
		n->_prec_sy.print(stdout, "%d ");
		printf("\n");
	    }
	}

	/* prec(n) = prec_ct(n) U prec_sy(n) */
	n->_prec.set_union(&(n->_prec_ct), &(n->_prec_sy));

	if (debug(__FNAME, 2)) {
	    printf("\tprec_old is: ");
	    prec_old.print(stdout, "%d ");
	    printf("\n\tn->_prec is: ");
	    n->_prec.print(stdout, "%d ");
	    printf("\n");
	}

	if (prec_old != n->_prec) {
	    if (debug(__FNAME, 2)) {
		printf("\tAppending control successors of node %d to work "
			"queue: ", n->number()
			);
		pfg_node_list_iter node_iter(n->succs());
		while (node_iter.is_empty() == FALSE) {
		    printf("%d ", node_iter.step()->number());
		}
		printf("\n");

		printf("\tAppending synchronization successors of node %d "
			"to work queue: ", n->number()
			);
		node_iter.reset(n->ssuccs());
		while (node_iter.is_empty() == FALSE) {
		    printf("%d ", node_iter.step()->number());
		}
		printf("\n");
	    }

	    pfg_node_list_iter node_iter(n->succs());
	    while (!node_iter.is_empty()) {
		queue->append(node_iter.step());
	    }

	    node_iter.reset(n->ssuccs());
	    while (!node_iter.is_empty()) {
		queue->append(node_iter.step());
	    }
	}
    }

    /* Traverse all the loop bodies assigning to each node n in the body the
     * same prec() set than the outermost header loop node enclosing n.
     */
    for (i = 0; i < num; i++) {
	pfg_node *node = _graph->node(i);
	if (_graph->loop_depth(node) > 0) {
	    pfg_node *header = _graph->outermostLoopHeader(node);

	    assert_msg(header, ("Node %d has a loop depth > 0. It should"
		    " have a loop header!\n", node->number()));

	    if (header != node) {

		if (debug(__FNAME, 1)) {
		    printf("\n%s(): Setting prec(%d) = prec(%d)\n",
			    __FNAME, node->number(), header->number());
		}

		node->_prec.copy(&(header->_prec));
	    }
	}
    }


    if (debug(__FNAME, 1)) {
	printf("\n%s(): Guaranteed execution ordering Prec(n)\n", __FNAME);
	unsigned i;
	for (i = 0; i < _graph->num_nodes(); i++) {
	    printf("\tPrec(%d) = ", i);
	    _graph->node(i)->_prec.print(stdout, "%d ");
	    printf("\n");
	}
	printf("\n");
    }

    delete queue;
}

//---------------------------------------------------------------------------
// cssame::sortPhiPi()
// Sorts the phi and pi terms by their line numbers.
//---------------------------------------------------------------------------
VarRef_list *
cssame::sortPhiPi()
{
    static char *__FNAME = "cssame::sortPhiPi";

    VarRef_list *sorted = new VarRef_list;
    assert(sorted);

    if (debug(__FNAME, 2)) {
	printf("%s(): Sorting list of references\n", __FNAME);
    }

    unsigned i;
    for (i = 0; i < _graph->num_nodes(); i++) {
	VarRef_list_iter iter(_graph->node(i)->refs());
	while (iter.is_empty() == FALSE) {
	    VarRef *newref = iter.step();
	    int newlineno = newref->lineno();
	    if (!newref->isPI() && !newref->isPHI()) {
		continue;
	    }

	    if (debug(__FNAME, 2)) {
		printf("\tInserting reference ");
		newref->print_brief();
		printf(" into sorted list\n");
	    }

	    boolean inserted = FALSE;
	    VarRef_list_iter sorted_iter(sorted);
	    while (sorted_iter.is_empty() == FALSE && !inserted) {
		VarRef *oldref = sorted_iter.step();
		int oldlineno = oldref->lineno();

		if (debug(__FNAME, 3)) {
		    printf("\tComparing new line %d with old line %d\n",
			    newlineno, oldlineno);
		}

		if (newlineno <= oldlineno) {
		    VarRef_list_e *oldref_e = sorted->lookup(oldref);
		    sorted->insert_before(newref, oldref_e);
		    inserted = TRUE;

		    if (debug(__FNAME, 2)) {
			printf("\tInserted before reference ");
			oldref->print_brief();
			printf("\n");
		    }
		}
	    }

	    if (!inserted) {
		sorted->append(newref);
		if (debug(__FNAME, 2)) {
		    printf("\tAppended at the end of the list\n");
		}
	    }
	}
    }

    return sorted;
}


//---------------------------------------------------------------------------
// cssame::placePhiFunctions()
// Algorithm to place phi-functions in the graph. Taken from "High
// Performance Compilers for Parallel Computing", M. Wolfe, page 175. The
// steps of the algorithm are also included in comments:
// 
// Find dominance frontier for nodes in the graph
//
// for X in V do
//    inWork(X) = Empty;
//    added(X) = Empty;
// endfor
// 
// workList = Empty;
// for M in Symbols do
//     for X in D(M) do
//         workList = workList U {X}
//         inWork(X) = M;
//         
//     endfor
//     while workList <> Empty do
//         remove some node X from workList
//         for W in DF(X) do
//             if added(W) <> M then
//                add phi-term for M at W
//                added(W) = M
//                if inWork(W) <> M then
//                    workList = workList U {W}
//                    inWork(W) = M
//                endif
//             endif
//         endfor
//     endwhile
// endfor
//---------------------------------------------------------------------------
void
cssame::placePhiFunctions()
{
    static char *__FNAME = "cssame::placePhiFunctions";

    pfg_node_list *workList = NULL;
    var_sym **added, **inWork;
    sym_node_list *symbols = _graph->tproc()->proc_syms()->symbols();
    sym_node_list_iter symbols_iter(symbols);
    unsigned x;

    if (debug(__FNAME, 2)) {
	printf("%s(): Dominance frontier for all the nodes in the graph\n",
		__FNAME);
	for (x = 0; x < _graph->num_nodes(); x++) {
	    printf("%s(): DF(%d) = { ", __FNAME, x);
	    bit_set *df = _graph->dom_frontier(_graph->node(x));
	    bit_set_iter df_iter(df);
	    while (df_iter.is_empty() == FALSE) {
		int w = df_iter.step();
		printf("%d ", w);
	    }
	    printf(" }\n");
	}
    }

    /* for X in V do
     *    inWork(X) = Empty;
     *    added(X) = Empty;
     * endfor
     */
    added = new var_sym *[_graph->num_nodes()];
    assert(added);

    inWork = new var_sym *[_graph->num_nodes()];
    assert(inWork);

    for (x = 0; x < _graph->num_nodes(); x++) {
	inWork[x] = NULL;
	added[x] = NULL;
    }

    /* workList = Empty; */
    workList = new pfg_node_list;
    assert(workList);

    /* for M in Symbols do */
    while (symbols_iter.is_empty() == FALSE) {
	sym_node *sym = symbols_iter.step();
	if (sym->is_var() == FALSE) {
	    continue;
	}

	/* for X in D(M) do
	 *    workList = workList U {X}
	 *    inWork(X) = M;
	 * endfor
	 */
	var_sym *m = (var_sym *)sym;
	VarRef_list *defs = (VarRef_list *)m->peek_annote(k_defs);
	VarRef_list_iter defs_iter(defs);
	while (defs_iter.is_empty() == FALSE) {
	    pfg_node *node = defs_iter.step()->node();
	    workList->append(node);
	    inWork[node->number()] = m;
	    if (debug(__FNAME, 1)) {
		printf("%s(): Adding node %d to work list for variable %s\n",
			__FNAME, node->number(), m->name());
	    }
	}

	/* while workList <> Empty do */
	while (workList->count() > 0) {
	    /* remove some node X from workList */
	    pfg_node *node = workList->pop();

	    /* for W in DF(X) do */
	    bit_set *df = _graph->dom_frontier(node);
	    bit_set_iter df_iter(df);
	    while (df_iter.is_empty() == FALSE) {
		/* if added(W) <> M then
		 *    add phi-term for M at W
		 *    added(W) = M
		 *    if inWork(W) <> M then
		 *        workList = workList U {W}
		 *        inWork(W) = M
		 *    endif
		 * endif
		 */
		int w = df_iter.step();
		if (added[w] != m) {
		    if (debug(__FNAME, 2)) {
			printf("cssame::placePhiFunctions(): Add phi-term for "
				"'%s' at node %d\n", m->name(), w);
		    }
		    PhiTerm *newphi = _graph->node(w)->addPhi(m);
		    _graph->refs()->append(newphi);
		    added[w] = m;
		    if (inWork[w] != m) {
			workList->append(_graph->node(w));
			inWork[w] = m;
		    }
		}
	    }
	}
    }
}


//---------------------------------------------------------------------------
// cssame::placeFUDChains()
// Link each variable use with its unique reaching definition or
// pseudo-assignment and fill in the phi-term links. Taken from "High
// Performance Compilers for Parallel Computing", M. Wolfe, page 178.
//---------------------------------------------------------------------------
void
cssame::placeFUDChains()
{
    /* Set up an iterator on the procedure's symbol table. NOTE: We should
     * also include any parent symbol table to catch variables in the
     * global scope.
     */
    sym_node_list *symbols = _graph->tproc()->proc_syms()->symbols();
    sym_node_list_iter symbols_iter(symbols);


    /* for M in Symbols do
     *     CurrDef(M) = Empty;
     * endfor
     */
    while (symbols_iter.is_empty() == FALSE) {
	sym_node *sym = symbols_iter.step();
	if (sym->is_var() == FALSE) {
	    continue;
	}

	/* Remove the existing k_currdef annotation */
	sym->get_annote(k_currdef);
    }

    /* Search(Entry) */
    this->searchFUDChains(_graph->entry_node());
}


//---------------------------------------------------------------------------
// cssame::searchFUDChains()
// Performs a depth-first traversal of the dominator tree looking for FUD
// (Factored Use-Def) chains. Taken from "High Performance Compilers for
// Parallel Computing", M. Wolfe, page 178.
//---------------------------------------------------------------------------
void
cssame::searchFUDChains(pfg_node *x)
{
    static char *__FNAME = "cssame::searchFUDChains";

    if (debug(__FNAME, 2)) {
	printf("\n\n%s(): Searching FUD chains in node %d\n", __FNAME,
		x->number());
	printf("%s(): Variables referenced in this node\n", __FNAME);
	VarRef_list_iter ref_iter(x->refs());
	while (!ref_iter.is_empty()) {
	    printf("\t");
	    ref_iter.step()->print();
	    printf("\n");
	}
    }

    /* for each variable use or def or phi-term R in X do
     *     let M be the variable referenced at R
     *     if R is a use then
     *         Chain(R) = CurrDef(M)
     *     else if R is a def or phi-term then
     *         SaveChain(R) = CurrDef(M)
     *         CurrDef(M) = R
     *     endif
     * endfor
     */
    VarRef_list_iter ref_iter(x->refs());
    while (ref_iter.is_empty() == FALSE) {
	VarRef *ref = ref_iter.step();
	var_sym *var = ref->var();

	/* Retrieve the current definition for the variable */
	VarRef *currdef = (VarRef *)var->peek_annote(k_currdef);

	if (debug(__FNAME, 2)) {
	    printf("%s(): Current definition for %s: ", __FNAME, var->name());
	    if (currdef) {
		currdef->print();
	    } else {
		printf("NONE");
	    }
	    printf("\n");
	}
	
	if (ref->isU()) {
	    ref->set_chain(currdef);
	} else if (ref->isD() || ref->isPHI()) {
	    ref->set_save_chain(currdef);
	    /* Replace the current k_currdef with a new one */
	    var->get_annote(k_currdef);
	    var->append_annote(k_currdef, (void *)ref);

	    if (debug(__FNAME, 2)) {
		printf("%s(): New definition for %s: ", __FNAME, var->name());
		ref->print();
		printf("\n");
	    }
	}
    }

    /* for Y in SUCC(X) do
     *    J = WhichPred(X -> Y)
     *    for each phi-term R in Y do
     *        let M be the variable referenced at R
     *        phi-chain(R)[J] = CurrDef(M)
     *    endfor
     * endfor
     */
    int j;
    for (j = 0; j < x->succs()->count(); j++) {
	pfg_node *y = (*(x->succs()))[j];
	VarRef_list_iter ref_iter(y->refs());

	while (ref_iter.is_empty() == FALSE) {
	    VarRef *ref = ref_iter.step();
	    if (ref->isPHI() == FALSE) {
		continue;
	    }
	    var_sym *var = ref->var();
	    VarRef *currdef = (VarRef *)var->peek_annote(k_currdef);
	    ((PhiTerm *)ref)->phi_chain()->append(currdef);
	    if (debug(__FNAME, 2)) {
		printf("%s(): Updated phi-term ", __FNAME);
		ref->print();
		printf("\n\twith argument ");
		if (currdef) {
		    currdef->print();
		} else {
		    printf("NONE");
		}
		printf("\n");
	    }
	}
    }


    /* for Y in Child(X) do	<-- Child(X) is the set of dominator
     * 	   Search(Y)                children of X in the dominator tree.
     * endfor
     */
    for (j = 0; j < (int)_graph->num_nodes(); j++) {
	pfg_node *y = _graph->node(j);
	if (_graph->immed_dom(y) == x) {
	    if (debug(__FNAME, 2)) {
		printf("%s(): Node %d -> searching dominator child %d\n",
			__FNAME, x->number(), y->number());
	    }
	    this->searchFUDChains(y);
	}
    }

    /* for each variable use or def or phi-term R in X in reverse order do
     *     let M be the variable referenced at R
     *     if R is a def or a phi-term then
     *         CurrDef(M) = SaveChain(R)
     *     endif
     * endfor
     */
    int i;
    for (i = x->refs()->count() - 1; i >= 0; i--) {
	VarRef *ref = (*(x->refs()))[i];
	if (ref->isD() || ref->isPHI()) {
	    /* Restore the current definition for the variable */
	    ref->var()->get_annote(k_currdef);
	    ref->var()->append_annote(k_currdef, (void *)ref->save_chain());
	}
    }
}


//---------------------------------------------------------------------------
// cssame::placePiFunctions()
// Places pi-terms at parallel join nodes. A parallel join node is a node x
// that has at least one incoming DU conflict edge for a variable v. This
// algorithm assumes that phi-terms have already been placed. The
// conflicts for a node X are those nodes that have X as the conflict head.
// This makes it easier when looking for all the incoming DU edges into
// each node.
//
// Adapted from "Concurrent Static Single Assignment Form and Constant
// Propagation for Explicitly Parallel Programs", J. Lee, S. P. Midkiff and
// D. A. Padua. LCPC '97.
//---------------------------------------------------------------------------
void
cssame::placePiFunctions()
{
    static char *__FNAME = "cssame::placePiFunctions";

    unsigned n;
    for (n = 0; n < _graph->num_nodes(); n++) {
	pfg_node *node = _graph->node(n);

	if (debug(__FNAME, 2)) {
	    printf("%s(): Checking conflicts for node %d\n", __FNAME, n);
	}

	if (node->conflicts()->count() == 0) {
	    continue;
	}

	/* The first argument to the PI function is the incoming control
	 * edge. Because of the way the flow graph is constructed
	 * there should only be one incoming control edge.
	 */
	assert(node->preds()->count() == 1);

	/* We traverse the incoming conflict edges to this node. For every
	 * incoming DU edge we create a new PI node for the conflicting
	 * variable if one hasn't been created already. If a PI term
	 * already exists for the variable, we add the reference to the
	 * list of arguments.
	 */
	RefConflict_list_iter conf_iter(node->conflicts());
	while (conf_iter.is_empty() == FALSE) {
	    RefConflict *conf = conf_iter.step();
	    if (conf->tail()->isD() && conf->head()->isU()) {
		var_sym *var = conf->head()->var();

		PhiTerm *newpi = node->hasPi(var);
		if (newpi == NULL) {
		    newpi = node->addPi(var);
		    _graph->refs()->append(newpi);

		    /* Since every PI term must contain an argument for the
		     * incoming control flow edge, we add the control
		     * reaching definition (cr_def) for the variable to the
		     * PI term. The incoming definition is the chain
		     * element for the conflicting reference (computed when
		     * we placed phi-terms).
		     * 
		     * A NULL incoming control definition means that this
		     * use is only affected by concurrent definitions.
		     */
		    VarRef *cr_def = conf->head()->chain();
		    if (cr_def) {
			newpi->phi_chain()->append(cr_def);
		    }
		}

		/* Before adding the arguments to the new pi-term we should
		 * compute the guaranteed ordering information. Some
		 * arguments to the pi-term might be eliminated due to
		 * guaranteed ordering between the concurrent basic blocks.
		 */
		newpi->phi_chain()->append(conf->tail());

		/* Finally, we chain the use for the var to the pi-term we
		 * just added. (ie, link this use to its reaching def)
		 */
		conf->head()->set_chain(newpi);

		if (debug(__FNAME, 1)) {
		    printf("%s(): Added pi-term for variable '%s'\n",
			    __FNAME, var->name());
		    printf("%s(): ", __FNAME);
		    newpi->print();
		    printf("\n");
		}
	    }
	}
    }
}

//---------------------------------------------------------------------------
// cssame::rewritePiFunctions()
// This method removes arguments to PI terms based on the mutex bodies
// found in the program. 
//---------------------------------------------------------------------------
void
cssame::rewritePiFunctions()
{
    static char *__FNAME = "cssame::rewritePiFunctions";

    if (debug(__FNAME, 1)) {
	printf("\n\n%s(): CSSAME before rewriting PI terms\n");
	this->print();
	printf("\n\n");

	var_sym_list_iter mutex_iter(_mutexVars);
	while (!mutex_iter.is_empty()) {
	    var_sym *var = mutex_iter.step();

	    printf("\n%s(): Mutex structure for synchronization variable %s\n",
		   __FNAME, var->name());

	    StrictInterval_list *mxstruct;
	    mxstruct = (StrictInterval_list *)var->peek_annote(k_mutex_struct);
	    StrictInterval_list_iter interval_iter(mxstruct);
	    while (!interval_iter.is_empty()) {
		StrictInterval *interval = interval_iter.step();
		printf("\t");
		interval->print();
	    }
	}
    }

    /* First step. Compute the set of definitions that reach the exit of
     * each mutex body in every mutex structure.
     */
    var_sym_list_iter mutex_iter(_mutexVars);
    while (!mutex_iter.is_empty()) {
	var_sym *var = mutex_iter.step();

	StrictInterval_list *mxstruct;
	mxstruct = (StrictInterval_list *)var->peek_annote(k_mutex_struct);

	StrictInterval_list_iter interval_iter(mxstruct);
	while (!interval_iter.is_empty()) {
	    StrictInterval *interval = interval_iter.step();
	    interval->computeExitRDefs();
	}
    }


    /* Second step. For each PI term in every body of a mutex structure,
     * remove those arguments that come from other mutex bodies and are not
     * in the set of exit reaching definitions.
     */
    mutex_iter.reset(_mutexVars);
    while (!mutex_iter.is_empty()) {
	var_sym *var = mutex_iter.step();

	StrictInterval_list *mxstruct;
	mxstruct = (StrictInterval_list *)var->peek_annote(k_mutex_struct);

	StrictInterval_list_iter interval_iter(mxstruct);
	while (!interval_iter.is_empty()) {
	    StrictInterval *interval = interval_iter.step();
	    interval->rewritePiFunctions(mxstruct);
	}
    }

    if (debug(__FNAME, 1)) {
	printf("\n\n%s(): CSSAME after rewriting PI terms\n");
	this->print();
	printf("\n\n");
    }
}
