//---------------------------------------------------------------------------
// $Id: mutex-analysis.cc,v 1.3 1998/06/03 16:52:43 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>
#include <suif.h>
#include <pfg.h>
#include <odc-util.h>

#include "cssame.h"

//---------------------------------------------------------------------------
// cssame::findMutexBodies()
//---------------------------------------------------------------------------
StrictInterval_list *
cssame::findMutexBodies()
{
    static char *__FNAME = "cssame::findMutexBodies";

    StrictInterval_list *S;

    bit_set_list *pn_set = new bit_set_list;
    assert(pn_set);

    bit_set_list *px_set = new bit_set_list;
    assert(px_set);

    _mutexVars = new var_sym_list;
    assert(_mutexVars);

    /* Create the two partitions pn and px containing the entry and exit
     * partitions for each mutex instruction in the program.
     */
    VarRef_list_iter ref_iter(_graph->refs());
    while (!ref_iter.is_empty()) {
	VarRef *ref = ref_iter.step();
	if (!ref->isMutexSync()) {
	    continue;
	}

	/* Add the entry partition to the set of entry partitions */
	bit_set *pn = (bit_set *)ref->var()->peek_annote(k_entry_sync_part);
	assert(pn);
	if (pn_set->lookup(pn) == NULL) {
	    pn_set->append(pn);
	}

	/* Add the exit partition to the set of exit partitions */
	bit_set *px = (bit_set *)ref->var()->peek_annote(k_exit_sync_part);
	assert(px);
	if (px_set->lookup(px) == NULL) {
	    px_set->append(px);
	}

	/* Add the synchronization variable to the set of synch variables. */
	if (_mutexVars->lookup(ref->var()) == NULL) {
	    _mutexVars->append(ref->var());
	}

	/* Remove any existing mutex structure associated with the
	 * synchronization object.
	 */
	delete (StrictInterval_list *)ref->var()->get_annote(k_mutex_struct);
    }

    /* Look for all the strict intervals defined by the partitions */
    S = this->findStrictIntervals(_mutexVars, pn_set, px_set);

    if (debug(__FNAME, 1)) {
	printf("\n%s(): Strict intervals found\n", __FNAME);
	StrictInterval_list_iter interval_iter(S);
	while (!interval_iter.is_empty()) {
	    interval_iter.step()->print();
	}
    }

    /* Display warnings for unmatched sync nodes left in the partitions */
    this->validateMutexBodies(pn_set, px_set);

    /* Create mutex structures for each synchronization object. A mutex
     * structure for a synchronization object is a set of mutex bodies
     * that access that object. Mutex structures are annotations attached
     * to the lock variables.
     */
    StrictInterval_list_iter interval_iter(S);
    while (!interval_iter.is_empty()) {
	StrictInterval *interval = interval_iter.step();
	var_sym *object = interval->var();

	StrictInterval_list *mxstruct;
	mxstruct = (StrictInterval_list *)object->peek_annote(k_mutex_struct);
	if (mxstruct == NULL) {
	    mxstruct = new StrictInterval_list;
	    assert(mxstruct);
	    object->append_annote(k_mutex_struct, (void *)mxstruct);
	}
	mxstruct->append(interval);
    }


    delete pn_set;
    delete px_set;

    return S;
}



//---------------------------------------------------------------------------
// cssame::findStrictIntervals()
// Given a graph and a partition of nodes this function returns a list of
// strict intervals in the graph relative to the partition. A strict
// interval is a pair (n, x) which represents the set of nodes that can be
// reached from n without passing through x.
//
// The partition of nodes P is represented by two sets of partitions:
// pn_set and px_set. Each set in pn_set is a set of nodes pn and each set
// in px_set is a set of nodes px. The reason the nodes are split up this
// way is that each set pn is supposed to be the set of entry nodes for a
// particular kind of interval (e.g., all the lock() instructions that
// access the same lock L). Thus, the partition P is defined to be:
//
//   P = pn_set U px_set = pn1 U pn2 U ... U pnN U px1 U px2 ... U pxN
//
// where N is the number of different synchronization variables. The
// algorithm is based on Masticola's definition of strict intervals in
// 'Static Detection of deadlocks in polynomial time', PhD thesis, Rutgers
// University, 1993. Note however that we do not really comply with
// Masticola's definition of strict intervals, in particular we do not
// comply with the global properties for strict intervals. Also, we do not
// return strict all strict intervals, we only look intervals in the
// corresponding pairs pn_i and px_i.
//
// We assume that the partitions pn_i and px_i correspond to the same
// object. That is, pn_i is the set of entry nodes for object 'i' and px_i
// are the set exit nodes for object 'i'.
//
// We further assume that dominator and post-dominator information has
// already been computed on the graph.
//---------------------------------------------------------------------------
StrictInterval_list *
cssame::findStrictIntervals(var_sym_list *vars_set, bit_set_list *pn_set, bit_set_list *px_set)
{
    static char *__FNAME = "cssame::findStrictIntervals";

    if (debug(__FNAME, 2)) {
	printf("%s(): Number of partitions in pn: %d, in px: %d\n",
		__FNAME, pn_set->count(), px_set->count()
	      );
    }

    assert_msg(pn_set->count() == px_set->count(), 
		("Entry and exit partitions must have the same size\n"));

    StrictInterval_list *S = new StrictInterval_list;

    /* First step. Find all the candidate strict intervals. For every set
     * of entry and exit nodes (pn_i and px_i) we intersect the dominators
     * and post-dominators for every pair (n, x) with n in pn_i and x in
     * px_i. Add to the initial set of candidates every pair (n, x) with a
     * non-empty intersection.
     *
     * foreach Pn and Px in P (s.t. Pn and Px correspond to the same symbol)
     *     foreach n in Pn
     *         foreach x in Px
     *             if n in DOM(x) ^ x in PDOM(n) then
     *                 interval->entry = n
     *                 interval->exit = x
     *                 interval->body = DOM(x) ^ PDOM(n)
     *                 S = S U {interval}
     *             endif
     *         endfor
     *     endfor
     * endfor
     */
    bit_set_list_iter pn_set_iter(pn_set);
    bit_set_list_iter px_set_iter(px_set);
    var_sym_list_iter vars_iter(vars_set);

    while (!pn_set_iter.is_empty() && !px_set_iter.is_empty()) {
	bit_set *pn = pn_set_iter.step();
	bit_set *px = px_set_iter.step();
	var_sym *var = (var_sym *)vars_iter.step();

	/* Traverse all the nodes in pn and px looking for pairs of nodes
	 * (n, x) such that n dominates x and x postdominates n.
	 */
	bit_set_iter pn_iter(pn);
	while (!pn_iter.is_empty()) {
	    pfg_node *entry = _graph->node(pn_iter.step());
	    bit_set_iter px_iter(px);
	    while (!px_iter.is_empty()) {
		pfg_node *exit = _graph->node(px_iter.step());

		/* If the entry node is in the set of dominators for the
		 * exit node and the exit node is in the set of
		 * postdominators for the entry node, we add the pair to
		 * the set of possible intervals.
		 */
		bit_set *dom_x = _graph->dominators(exit);
		bit_set *pdom_n = _graph->postdominators(entry);

		if (dom_x->contains(entry->number()) &&
		    pdom_n->contains(exit->number())) {
		    StrictInterval *interval;
		    interval = new StrictInterval(var, entry, pn, exit, px);
		    assert(interval);
		    S->append(interval);
		}
	    }
	}
    }

    if (debug(__FNAME, 2)) {
	printf("\n%s(): Initial set of strict intervals\n", __FNAME);
	StrictInterval_list_iter interval_iter(S);
	while (!interval_iter.is_empty()) {
	    interval_iter.step()->print();
	}
	printf("\n");
    }

    /* Second step. Filter out unwanted intervals by making sure that each
     * interval (n, x) doesn't include other nodes in pn U px. This avoids
     * the following problem:
     *
     *	(1)	lock x;
     *		...
     *	(2)	unlock x;
     *		...
     *	(3)	lock x;
     *		...
     *	(4)	unlock x;
     *
     * The previous pass would have created the intervals (1, 2), (1, 4)
     * and (3, 4). Obviously the interval (1, 4) is bogus, so we get rid of
     * it.
     *
     * foreach interval in S
     *     foreach n in interval->Pn U interval->Px
     *         if n <> interval->entry ^ n <> interval->exit then
     *              if n in interval->body then
     *                  remove interval from S
     *              endif
     *         endif
     *     endfor
     * endfor
     */
    StrictInterval_list *to_remove = new StrictInterval_list;
    assert(to_remove);

    StrictInterval_list_iter interval_iter(S);
    while (!interval_iter.is_empty()) {
	StrictInterval *interval = interval_iter.step();

	bit_set *nodes = new bit_set(0, _graph->num_nodes());
	nodes->set_union(interval->pn(), interval->px());

	bit_set_iter nodes_iter(nodes);
	while (!nodes_iter.is_empty()) {
	    int node = nodes_iter.step();
	    if (node != interval->entry()->number() && 
		node != interval->exit()->number()  && 
		interval->body()->contains(node))      {
		if (to_remove->lookup(interval) == NULL) {
		    to_remove->append(interval);
		}
	    }
	}

	delete nodes;
    }

    while (!to_remove->is_empty()) {
	StrictInterval *interval = to_remove->pop();

	if (debug(__FNAME, 2)) {
	    printf("%s(): Removing interval ", __FNAME);
	    interval->print();
	}

	delete S->remove(S->lookup(interval));
    }

    delete to_remove;

    /* Third step. Remove from their original partitions those nodes that
     * are in some interval in S.
     */
    interval_iter.reset(S);
    while (!interval_iter.is_empty()) {
	StrictInterval *interval = interval_iter.step();
	interval->pn()->remove(interval->entry()->number());
	interval->px()->remove(interval->exit()->number());
    }
    

    return S;
}


//---------------------------------------------------------------------------
// cssame::validateMutexBodies()
// Once the initial strict intervals have been created, we keep the
// intervals that are valid mutex bodies and report errors on the unmatched
// nodes. Unmatched nodes have been left in the original pn and px sets,
// therefore we traverse the partitions pn_set and px_set checking each
// partition pair (pn_i, px_i).
//---------------------------------------------------------------------------
void 
cssame::validateMutexBodies(bit_set_list *pn_set, bit_set_list *px_set)
{
    bit_set_list_iter pn_set_iter(pn_set);
    bit_set_list_iter px_set_iter(px_set);

    while (!pn_set_iter.is_empty() && !px_set_iter.is_empty()) {
	bit_set *pn = pn_set_iter.step();
	bit_set *px = px_set_iter.step();

	/* Traverse all the remaining nodes in pn and px looking for pairs
	 * of nodes (n, x) such that there is a path from n to x but n does
	 * not dominate x and/or x does not postdominate n.
	 */
	bit_set_iter pn_iter(pn);
	while (!pn_iter.is_empty()) {
	    pfg_node *entry = _graph->node(pn_iter.step());
	    bit_set_iter px_iter(px);
	    while (!px_iter.is_empty()) {
		pfg_node *exit = _graph->node(px_iter.step());

		if (_graph->isPath(entry, exit)) {
		    /* There is a path from entry to exit. If entry
		     * dominates the exit but the exit does not
		     * postdominate the entry, the problem is that the lock
		     * may not be released.
		     */
		    if (_graph->dominates(entry, exit) &&
			!_graph->postdominates(exit, entry)) {
			warning_line(entry->last_instr(), ("Lock may not be"
				" released"));
			pn->remove(entry->number());
			px->remove(exit->number());
		    }

		    /* If entry does not dominate the exit but the exit
		     * postdominates the entry, the problem is that the
		     * user may be releasing an unheld lock.
		     */
		    if (!_graph->dominates(entry, exit) &&
			_graph->postdominates(exit, entry)) {
			warning_line(exit->last_instr(), ("May be releasing"
				" an unheld lock"));
			pn->remove(entry->number());
			px->remove(exit->number());
		    }
		}
	    }
	}

	/* Any remaining nodes in pn and px are unmatched synchronization
	 * operations
	 */
	pn_iter.reset(pn);
	while (!pn_iter.is_empty()) {
	    pfg_node *n = _graph->node(pn_iter.step());
	    warning_line(n->last_instr(), ("Unmatched lock operation\n"));
	}

	bit_set_iter px_iter(px);
	while (!px_iter.is_empty()) {
	    pfg_node *x = _graph->node(px_iter.step());
	    warning_line(x->last_instr(), ("Unmatched unlock operation\n"));
	}
    }
}


//***************************************************************************
//		 Member functions for class StrictInterval
//***************************************************************************
//---------------------------------------------------------------------------
// StrictInterval::StrictInterval()
//---------------------------------------------------------------------------
StrictInterval::StrictInterval(var_sym *var, pfg_node *n, bit_set *pn, pfg_node *x, bit_set *px)
    : _var(var), _entry(n), _pn(pn), _exit(x), _px(px)
{
    pfg *graph = n->parent();
    unsigned num_nodes = graph->num_nodes();
    unsigned i;

    _body = new bit_set(0, num_nodes);
    assert(_body);

    for (i = 0; i < num_nodes; i++) {
	if (graph->dominates(n, graph->node(i)) && 
	    graph->postdominates(x, graph->node(i))) {
	    _body->add(i);
	}
    }
    _body->remove(n->number());		/* Entry is not included in body */
}


//---------------------------------------------------------------------------
// StrictInterval::print()
//---------------------------------------------------------------------------
void
StrictInterval::print(FILE *f)
{
    fprintf(f, "(%d-%d]: ", source_line_num(_entry->last_instr()),
	    source_line_num(_exit->last_instr())
	    );
    _body->print(stdout, "%d ");
    fprintf(f, "\n");
}


//---------------------------------------------------------------------------
// StrictInterval::computeExitRDefs()
// Computes the set of definitions that reach the exit node of the
// interval. We compute this set by traversing all the definitions and
// phi-terms in the body of the interval.
//
// A defintion will reach the exit node if it postdominates every
// definition or phi-term for the same variable inside the body of the
// interval. We keep all such definitions in the set _exitRDefs.
//---------------------------------------------------------------------------
void
StrictInterval::computeExitRDefs()
{
    static char *__FNAME = "StrictInterval::computeExitRDefs";

    _exitRDefs = new VarDef_list;
    assert(_exitRDefs);

    pfg *graph = _entry->parent();

    /* Gather all definitions in the interval. */
    VarDef_list *defs_list = new VarDef_list;
    assert(defs_list);

    bit_set_iter body_iter(_body);
    while (!body_iter.is_empty()) {
	pfg_node *node = graph->node(body_iter.step());

	VarRef_list_iter refs_iter(node->refs());
	while (!refs_iter.is_empty()) {
	    VarRef *ref = refs_iter.step();
	    if (ref->isD()) {
		defs_list->append((VarDef *)ref);
	    }
	}
    }

    /* Determine which defintions for a given variable can reach the end of
     * the interval. A definition can reach the exit node if and only if it
     * is not postdominated by any other definition in the body.
     */
    VarDef_list_iter defs_iter1(defs_list);
    while (!defs_iter1.is_empty()) {
	VarDef *def1 = defs_iter1.step();
	boolean reachesExit = TRUE;

	/* Does any other definition in the body postdominate def1? */
	VarDef_list_iter defs_iter2(defs_list);
	while (!defs_iter2.is_empty()) {
	    VarDef *def2 = defs_iter2.step();
	    if (def2->var()->sym_id() != def1->var()->sym_id() || 
		def1 == def2) {
		continue;
	    }

	    /* If def2 postdominates def1 then def1 cannot reach the end of
	     * the interval.
	     */
	    if (graph->postdominates(def2->node(), def1->node())) {
		reachesExit = FALSE;
		break;
	    }
	}

	/* If the definition reaches the exit node, add it to the list. */
	if (reachesExit) {
	    _exitRDefs->append(def1);
	}
    }

    if (debug(__FNAME, 1)) {
	printf("%s(): List of defintions reaching the exit of\n", __FNAME);
	printf("\tmutex body: ");
	this->print();

	VarDef_list_iter rdefs_iter(_exitRDefs);
	while (!rdefs_iter.is_empty()) {
	    printf("\t");
	    rdefs_iter.step()->print();
	    printf("\n");
	}

	printf("\n");
    }
}


//---------------------------------------------------------------------------
// StrictInterval::rewritePiFunctions()
// For every PI function in the interval, check if any of the arguments
// comes from another mutex body in the same structure. If the argument
// comes from another mutex body and it's not in the list of definitions
// reaching the exit of that mutex body then we can safely remove it.
//---------------------------------------------------------------------------
void
StrictInterval::rewritePiFunctions(StrictInterval_list *mutexStruct)
{
    static char *__FNAME = "StrictInterval::rewritePiFunctions";

    pfg *graph = _entry->parent();

    PhiTerm_list *pi_list = new PhiTerm_list;
    assert(pi_list);

    bit_set_iter body_iter(_body);
    while (!body_iter.is_empty()) {
	pfg_node *node = graph->node(body_iter.step());

	VarRef_list_iter refs_iter(node->refs());
	while (!refs_iter.is_empty()) {
	    VarRef *ref = refs_iter.step();
	    if (ref->isPI()) {
		pi_list->append((PhiTerm *)ref);
	    }
	}
    }


    if (debug(__FNAME, 2)) {
	printf("\n%s(): Checking PI terms in interval ", __FNAME);
	this->print();
    }

    PhiTerm_list_iter pi_iter(pi_list);
    while (!pi_iter.is_empty()) {
	PhiTerm *pi = pi_iter.step();

	if (debug(__FNAME, 2)) {
	    printf("\n\tChecking PI term ");
	    pi->print();
	    printf("\n");
	}

	VarRef_list_iter args_iter(pi->phi_chain());
	if (!args_iter.is_empty()) {
	    /* Skip the first argument of the Pi term because it is the
	     * control reaching definition which do not care about.
	     */
	    args_iter.step();
	}

	while (!args_iter.is_empty()) {
	    VarRef *ref = args_iter.step();

	    /* All the parallel reaching references should be definitions
	     * otherwise this is not a properly formed PI term.
	     */
	    assert(ref->isD());
	    VarDef *arg = (VarDef *)ref;
	    
	    /* Check if the argument comes from a node in another mutex
	     * body in this mutex structure
	     */
	    StrictInterval_list_iter interval_iter(mutexStruct);
	    while (!interval_iter.is_empty()) {
		StrictInterval *interval = interval_iter.step();
		if (interval == this) {
		    continue;
		}

		if (debug(__FNAME, 2)) {
		    printf("\tChecking if argument: ");
		    arg->print();
		    printf("\n\tcomes from mutex body: ");
		    interval->print();
		}

		if (interval->body()->contains(arg->node()->number())) {
		    if (debug(__FNAME, 3)) {
			printf("\tinterval->exitRDefs()\n");
			VarDef_list_iter iter(interval->exitRDefs());
			while (!iter.is_empty()) {
			    printf("\t");
			    iter.step()->print();
			    printf("\n");
			}
			printf("\n");
			printf("\targ is: ");
			arg->print();
			printf("\n");
		    }

		    /* If the control reaching definition for this PI comes
		     * from this body, we can discard the argument.
		     */
		    VarRef *ctrl_rdef = (*(pi->phi_chain()))[0];
		    if (this->body()->contains(ctrl_rdef->node()->number())) {
			pi->phi_chain()->remove(pi->phi_chain()->lookup(arg));
		    } else if (!interval->exitRDefs()->lookup(arg)) {
			/* If the argument comes from another mutex body,
			 * we check if the argument reaches the exit of
			 * that mutex body. If not, we can discard it.
			 */
			pi->phi_chain()->remove(pi->phi_chain()->lookup(arg));

			if (debug(__FNAME, 2)) {
			    printf("\tYes. Removing argument\n");
			}
		    }
		}
	    }
	}
    }
}
