//---------------------------------------------------------------------------
// $Id: graph-dfa.cc,v 1.12 1998/06/03 16:52:38 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>

#define _MODULE_ "libpfg.a"

// #pragma implementation "graph.h"

#include <values.h>
#include <ctype.h>

#include <suif.h>
#include <par.h>
#include <odc-util.h>

#include "pfg.h"

//---------------------------------------------------------------------------
// pfg::addConfSyncEdges()
// Conflict analysis. This pass splits basic blocks so that they comply
// with the following rules:
//
// 1- A conflicting definition must be the last statement of a basic block.
//    A definition for a variable V at block X is conflicting if there is
//    another def or use for the same variable at another block Y such that
//    conc(X, Y) == TRUE.
//
// 2- A conflicting use must be the first statement of a basic block.
//
// If a conflicting def is found, a conflict edge is added from the
// defining block to the other block.
//---------------------------------------------------------------------------
void 
pfg::addConfSyncEdges()
{
    static char *__FNAME = "pfg::addConfSyncEdges";

    /* Find references, conflicts and synchronization statements in the
     * original PFG.
     */
    this->findRefs();
    this->findConflicts();

    if (debug(__FNAME, 1)) {
	printf("%s(): Conflicts found in the original PFG\n", __FNAME);
	RefConflict_list_iter iter(_conflicts);
	while (iter.is_empty() == FALSE) {
	    RefConflict *conf = iter.step();
	    conf->print();
	    printf("\n");
	}
	printf("\n");
    }

    /* Now that we know all the conflicts in the program, we can split the
     * PFG nodes that contain conflicting references. The rules for
     * splitting a PFG node are:
     *
     * 1- A conflicting definition or a call to a tail synch function must
     *    be the last statement of a basic block.
     *
     * 2- A conflicting use must or a call to a head synch function be the
     *    first statement of a basic block.
     */
    RefConflict_list_iter iter(_conflicts);
    while (iter.is_empty() == FALSE) {
	RefConflict *conf = iter.step();
	conf->tail()->node()->split();
    }

    /* Re-scan the PFG looking for references and conflicts to
     * take the new nodes into account.
     */
    this->findRefs();
    this->findConflicts();

    /* Now we can add the sync/conflict edges. We update the
     * _csuccs/_ssuccs and _cpreds/_spreds list for each node. The _csuccs
     * list for node i contains all the nodes j that conflict with i for
     * which i is the tail of the conflict. The _cpreds list is similarly
     * defined.
     */
    iter.reset(_conflicts);
    while (iter.is_empty() == FALSE) {
	RefConflict *conf = iter.step();
	/* If the tail of the conflict is a synchronization instruction, we
	 * update the synchronization successors and predecessors.
	 * Otherwise, it is a regular conflict.
	 */
	if (conf->tail()->isSync()) {
	    if (conf->tail()->isTSync() || conf->tail()->isMutexSync()) {
		conf->tail()->node()->ssuccs()->append(conf->head()->node());
		conf->head()->node()->spreds()->append(conf->tail()->node());
	    }
	} else {
	    conf->tail()->node()->csuccs()->append(conf->head()->node());
	    conf->head()->node()->cpreds()->append(conf->tail()->node());
	}
    }
}


//---------------------------------------------------------------------------
// pfg::findRefs()
//---------------------------------------------------------------------------
void
pfg::findRefs()
{
    unsigned i;

    /* Delete existing references */
    if (_refs) {
	VarRef_list_iter iter(_refs);
	while (!iter.is_empty()) {
	    delete iter.step();
	}
	_refs->erase();
	delete _refs;
    }
    _refs = new VarRef_list;
    assert(_refs);

    /* Traverse all the nodes in the graph looking for references */
    for (i = 0; i < this->num_nodes(); i++) {
	VarRef_list *r = this->node(i)->findRefs();
	if (r) {
	    VarRef_list_iter ref_iter(r);
	    while (!ref_iter.is_empty()) {
		_refs->append(ref_iter.step());
	    }
	}
    }
}

//***************************************************************************
//		       Protected and private methods
//***************************************************************************
//---------------------------------------------------------------------------
// pfg::findConflicts()
// Find all the conflicting references in the program. For every two nodes
// that may execute concurrently determine whether they access the same
// variable in a conflicting manner. The possible conflicts are: DD, DU and
// UD.
//---------------------------------------------------------------------------
void
pfg::findConflicts()
{
    static char *__FNAME = "pfg::findConflicts";

    unsigned i, j;

    /* Delete existing conflict lists */
    if (_conflicts) {
	RefConflict_list_iter iter(_conflicts);
	while (!iter.is_empty()) {
	    delete iter.step();
	}
	_conflicts->erase();
	delete _conflicts;
    }
    _conflicts = new RefConflict_list;
    assert(_conflicts);

    for (i = 0; i < this->num_nodes(); i++) {
	delete this->node(i)->_conflicts;
	this->node(i)->_conflicts = new RefConflict_list;
	assert(this->node(i)->_conflicts);
    }

    /* Compare all the nodes in the graph that can execute concurrently. If
     * they reference the same variable in a conflicting manner, create a
     * new conflict for the reference.
     *
     * Conflicts are appended to the conflict's head node. This makes it
     * easier to append PI terms later on. We also keep a global list of
     * conflicts in the graph and each conflicting reference contains a
     * list of the conflicts it participates in.
     */
    for (i = 0; i < this->num_nodes(); i++) {
	for (j = 0; j < this->num_nodes(); j++) {
	    if (this->conc(i, j) == FALSE) {
		continue;
	    }

	    VarRef_list_iter irefs(this->node(i)->refs());
	    while (irefs.is_empty() == FALSE) {
		VarRef *iref = irefs.step();

		VarRef_list_iter jrefs(this->node(j)->refs());
		while (jrefs.is_empty() == FALSE) {
		    VarRef *jref = jrefs.step();

		    if (jref->_conflicts == 0) {
			jref->_conflicts = new RefConflict_list;
			assert(jref->_conflicts);
		    }

		    /* Check whether the reference at node i (iref)
		     * conflicts with the reference at node j (jref). Two
		     * references conflict if they access the same variable
		     * in a conflicting manner. The possible conflicts are:
		     * DD, DU and UD.
		     */
		    if ( jref->var()->sym_id() == iref->var()->sym_id() &&
			 (  (jref->isD() && iref->isD())   ||
			    (jref->isD() && iref->isU())   ||
			    (jref->isU() && iref->isD())   ||
			    (jref->isHSync() && iref->isTSync()) ||
			    (jref->isTSync() && iref->isHSync()) ||
			    (jref->isUnlock() && iref->isLock())
			 ) 
			) {
			RefConflict *conf = new RefConflict(jref, iref);
			_conflicts->append(conf);
			this->node(i)->_conflicts->append(conf);
			iref->_conflicts->append(conf);
		    }
		}
	    }
	}
    }

    if (debug(__FNAME, 1)) {
	unsigned i;
	for (i = 0; i < this->num_nodes(); i++) {
	    printf("%s(): Conflicts for node %d\n", __FNAME, i);
	    RefConflict_list_iter tmp(this->node(i)->conflicts());
	    while (tmp.is_empty() == FALSE) {
		tmp.step()->print();
	    }
	    printf("\n");
	}
	printf("\n");
    }
}
