/* $Id: graph.cc,v 1.12 1998/06/03 16:52:38 diego Exp diego $ */
/*  Parallel Flow Graph Implementation */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    Copyright (c) 1995,1996 The President and Fellows of Harvard University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file. 
 

    Modified by Diego Novillo (diego@cs.ualberta.ca)
*/

#include <string.h>

#include <suif_copyright.h>

#define _MODULE_ "libpfg.a"

// #pragma implementation "graph.h"

#include <suif.h>
#include <par.h>
#include <odc-util.h>

#include "pfg.h"

/* local function prototypes */
static void add_pfg_edge(tree_node *t, void *x);
static void add_edge(tree_instr *ti, label_sym *target, unsigned n);

const int io_null = -1;		         /* null opcode extension to SUIF
					    op space */

/* Is_label() -- return TRUE if label instruction.  Notice that we don't
 * check in this routine to see if you've got the correct helper function. */
boolean
Is_label(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_lab) return TRUE;

    return FALSE;
}

/* Get_label() -- helper routine that performs the appropriate cast
 * to get at the label_sym on a label instruction. */
label_sym *
Get_label(instruction *i)
{
    assert(Is_label(i));
    return ((in_lab *)i)->label();
}


/* Is_ubr() -- return TRUE if instruction is an unconditional
 * jump or branch. 
 */
boolean
Is_ubr(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_jmp) return TRUE;

    return FALSE;
}

/* Is_null() -- returns TRUE if null instruction.  Notice that we don't
 * check in this routine to see if you've got the correct helper function. */
boolean
Is_null(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_null) return TRUE;		/* SUIF */

    return FALSE;
}

/* Is_cbr() -- return TRUE if instruction uses conditional
 * branch machine resources (independent of whether the branch
 * evaluation is conditional or not). */
boolean
Is_cbr(instruction *i)
{
    int op = (int)i->opcode();
    if ((op == io_btrue) || (op == io_bfalse)) return TRUE;

    return FALSE;
}

/* Is_call() -- return TRUE if instruction is a call. */
boolean
Is_call(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_cal) return TRUE;

    return FALSE;
}

/* Is_mbr() -- return TRUE if instruction is a multi-way branch. */
boolean
Is_mbr(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_mbr) 
	return TRUE;
    else
	return FALSE;
}

/* Is_return() -- return TRUE if instruction is a return. */
boolean
Is_return(instruction *i)
{
    int op = (int)i->opcode();
    if (op == io_ret) return TRUE;

    return FALSE;
}

/* Is_cti() -- return TRUE if control-transfer instruction 
 * This implementation is slow, but is guaranteed to give
 * the correct semantics. 
 */
boolean Is_cti(instruction *i)
{
    return Is_ubr(i) || Is_cbr(i) || Is_mbr(i) 
	|| Is_call(i) || Is_return(i); 
}

/* Get_target() -- helper routine that performs the appropriate cast
 * to get at the target sym_node on a branch/jump instruction.  Note
 * that the machine_instr may return NULL if it is an mbr instruction. */
sym_node *
Get_target(instruction *i)
{
    assert(Is_cti(i));
    return ((in_bj *)i)->target();
}

/*
 *  Build a new flow graph.  Flow graphs represent entire procedures.  
 */
pfg::pfg (tree_block *b, boolean build_blocks, 
    boolean break_at_call, boolean keep_l)
{
    static char *__FNAME = "pfg::pfg";

    /* initialization of private variables */
    tp = (tree_proc *)b; 
    nds = new pfg_node_array;
    en = NULL;
    ex = NULL;
    doms = NULL;
    pdoms = NULL;
    idom = NULL;
    ipdom = NULL;
    df = NULL;
    rdf = NULL;
    lp_depth = NULL;
    _conflicts = NULL;
    _refs = NULL;

    /* first pass -- create the nodes and fallthrough edges */
    graph_add_tree(b, NULL, build_blocks, break_at_call, keep_l, NULL);

    set_entry_node(pfg_begin_node(b));
    set_exit_node(pfg_end_node(b));

    /* second pass -- set the successor(s) of the entry node */
    tree_node_list_iter tnli(b->body());
    while (!tnli.is_empty()) {
	tree_node *tn = tnli.step(); 
	if (tn->is_instr()) {
	    tree_instr *ti = (tree_instr *)tn; 
	    ti->append_annote(k_proc_entry);
	    entry_node()->add_succ(pfg_begin_node(ti));
	    break;
	}
    }

    /* add an edge from the entry to the exit */
    entry_node()->add_succ(exit_node());

    /* third pass -- add the remaining edges.  postorder walk
     * because the tree_node being visited might be deleted */
    b->map(add_pfg_edge, (void *)exit_node(), FALSE);

    /* fourth pass -- set up layout information */
    if (keep_l) {
	for (unsigned i = 0; i < num_nodes(); ++i) {
	    pfg_node *nd = (*this)[i]; 
	    pfg_node *succ; 
	    tree_node_list_e *tnle; 
	    switch (nd->kind()) {
	    case PFG_BEGIN: 
		// grab non-end successor block
		succ = nd->succs()->head()->contents; 
		assert(!succ->is_end()); 
		assert(nd->lsucc == NULL); 
		assert(succ->lpred == NULL); 
		nd->lsucc = succ; 
		succ->lpred = nd; 
		break; 

	    case PFG_END:
		// grab highest numbered successor block
		// DO NOT set pred's layout successor to END; breaks 
		// set_layout_succ()
		assert(nd->lpred == NULL); 
		break; 

	    case PFG_BLOCK:
		tnle = ((pfg_block *)nd)->in_tail()->list_e()->next(); 
		if (tnle == NULL)
		    break; 
		succ = pfg_end_node(tnle->contents); 
		assert(nd->lsucc == NULL); 
		assert(succ->lpred == NULL); 
		nd->lsucc = succ; 
		succ->lpred = nd; 
		break; 

	    case PFG_INSTR:
		tnle = ((pfg_instr *)nd)->instr()->list_e()->next(); 
		if (tnle == NULL)
		    break; 
		succ = pfg_end_node(tnle->contents); 
		assert(nd->lsucc == NULL); 
		assert(succ->lpred == NULL); 
		nd->lsucc = succ; 
		succ->lpred = nd; 
		break; 

	    case PFG_TEST:
	    case PFG_LABEL:
	    default:
		assert(0);
		break; 
	    }
	}

    } else {
	/*
	   Add shadow gotos. This code is broken, it only works with
	   machsuif. Removed for now. 
	   
	   WARNING! The PFG library will not be able to keep track of
	   proper control flow if nodes are moved around!

	for (unsigned i = 0; i < num_nodes(); ++i) {
	    pfg_node *nd = (*this)[i]; 
	    if (!nd->is_block())
		continue; 
	    pfg_block *blk = (pfg_block *)nd; 

	    if (blk->succs()->count()
		&& (blk->succs()->head()->contents)->is_block()) {
		blk->set_shadow(blk->succs()->head()->contents); 
	    } 
	}
	*/
    }


    /* Fifth pass. Perform conflict and synchronization analysis to add
     * conflict and synchronization edges to the PFG.
     */
    this->addConfSyncEdges();

    if (debug(__FNAME, 1)) {
	printf("%s(): Parallel flow graph for function %s\n", __FNAME,
		tp->proc()->name());
	this->print();
    }
}


/* 
 * Destroy the pfg graph and remove the pfg annotations on all of the
 * tree_instrs and label_syms.  Notice that this is not entirely
 * necessary since the pfg annotations will not be written out
 * with the procedure.
 */

static void remove_pfg_annotations(tree_block *tb);

pfg::~pfg ()
{
    for (unsigned i = 0; i < num_nodes(); i++) delete (*nds)[i];
    delete nds;
    if (doms) delete[] doms;
    if (pdoms) delete[] pdoms;
    if (idom) delete[] idom;
    if (ipdom) delete[] ipdom;
    if (df) delete[] df;
    if (rdf) delete[] rdf;
    if (lp_depth) delete[] lp_depth;

    remove_pfg_annotations(tproc());
}

void
remove_pfg_annotations(tree_block *tb)
{
    tree_node_list_iter tnli(tb->body());

    /* walk over the flat list of instructions looking for annotations */
    while (!tnli.is_empty()) {
	tree_instr *ti = (tree_instr *)tnli.step();
	ti->get_annote(k_pfg_node);
	if (Is_label(ti->instr()))
	    Get_label(ti->instr())->get_annote(k_pfg_node);
    }
    tb->get_annote(k_pfg_begin);
    tb->get_annote(k_pfg_end);
}


/* 
 * Return the number of nodes currently in the PFG.  
 */
unsigned
pfg::num_nodes ()
{
    int n = nds->ub();
    return (unsigned)n;
}


/* 
 * Set the entry or exit node of the graph.  These should probably be
 * protected, because users don't get to add entry or end
 * nodes to a PFG.  
 */
void
pfg::set_entry_node (pfg_node *n)
{
    assert(n->is_begin());
    en = n;
}

void
pfg::set_exit_node (pfg_node *n)
{
    assert(n->is_end());
    ex = n;
}


/* attach() -- Link a free-floating pfg_block into an existing
 * pfg.  This is a helper function for clone_into() and new_empty_block(). 
 * nod is assumed to be a pfg_block.  nod's instructions are appended
 * to this' tree_node_list, and nod gets added to this PFG.  
 * All pfg edges and branch targets are cleared, except return blocks 
 * have the exit node set as a successor.  
 */
void 
pfg::attach(pfg_node *nod)
{
    assert(nod->kind() == PFG_BLOCK); 
    pfg_block *blk = (pfg_block *)nod; 
    tree_node_list *pfg_tnl = tproc()->body(); 

    // Make sure this is unattached
    assert(blk->parent() == NULL); 
    assert(blk->tnl != pfg_tnl); 
    assert((tree_instr *)blk->tnl->head()->contents == blk->in_head()); 
    assert((tree_instr *)blk->tnl->tail()->contents == blk->in_tail()); 
    pfg_node_list_iter cnlip(blk->preds()); 
    while (!cnlip.is_empty())
	assert(cnlip.step() == NULL); 
    pfg_node_list_iter cnlis(blk->preds()); 
    while (!cnlis.is_empty())
	assert(cnlis.step() == NULL); 
    assert(blk->layout_succ() == NULL); 
    assert(blk->tiS == NULL); 

    // Transfer instructions from blk->tnl to pfg_tnl
    // Destroy and change over blk->tnl
    blk->set_in_head(NULL); 
    while (!blk->tnl->is_empty()) {
	tree_node *tn = blk->tnl->pop(); 
	assert(tn->is_instr()); 
	if (blk->in_head() == NULL) // First iteration?
	    blk->set_in_head((tree_instr *)tn); 
	pfg_tnl->append(tn); 
    }
    blk->set_in_tail((tree_instr *)pfg_tnl->tail()->contents); 
    delete blk->tnl; 
    blk->tnl = pfg_tnl; 
    assert(blk->tnl->tail()->contents == blk->in_tail()); 

    blk->set_shadow(NULL); 
    blk->remove_explicit_targets(); 
    add_node(nod); 

    // If this is a return node, make the end node a successor
    if (blk->ends_in_return())
	blk->set_succ(0, exit_node()); 
}


/* 
 * clone_into() -- Make a copy of a block and link it into 
 * this pfg.  Return the clone to the user, so he can set its
 * successors and predecessors.  
 */
pfg_block *
pfg::clone_into(pfg_block *base, boolean parallel, pfg_node *par_parent,
	threadBody *tb)
{
    pfg_block *clone = new pfg_block(base, tproc()->proc_syms(), parallel,
	    par_parent, tb); 
    attach(clone); 
    return clone; 
}


/* 
 * new_empty_block() -- Make a new empty block (actually a block with
 * just a label instruction for its contents) and link it into 
 * this pfg.  Return the empty block to the user, so he can set its
 * successors and predecessors.  
 */
pfg_block *
pfg::new_empty_block(boolean parallel, pfg_node *par_parent, threadBody *tb)
{
    pfg_block *empty = new pfg_block(tproc()->proc_syms(), parallel,
	    par_parent, tb);
    attach(empty); 
    return empty; 
}


/*
 *  Remove unreachable nodes from a flow graph.  This simplifies finding
 *  dominators and may also be used after manually removing nodes (e.g.
 *  unwanted BEGIN and END markers, nop instructions, etc.) from the graph.
 *  Note: There should never be unreachable nodes in the reverse graph.
 */

static void unreachable_visit(pfg_node *n, bit_set *mark);

void
pfg::remove_unreachable_blocks (boolean delete_instrs)
{
    unsigned num = num_nodes();
    unsigned i;

    /* find the reachable blocks */
    bit_set mark(0, num);
    unreachable_visit(entry_node(), &mark);

    /* copy the reachable nodes to a new array */
    pfg_node_array *new_nds = new pfg_node_array;
    for (i = 0; i < num; i++) {
	pfg_node *nd = node(i);
	if (mark.contains(i)) {
	    int new_num = new_nds->extend((void *)nd);
	    nd->set_number((unsigned)new_num);
	} else {
	    /* detach the unreachable node from its successors */
	    pfg_node_list_iter succ_iter(nd->succs());
	    while (!succ_iter.is_empty()) {
		pfg_node *succ = succ_iter.step();
		nd->remove_succ(succ);
	    }
	    /* detach the unreachable node from its predecessors -- this is
               needed when there's a back-edge from a higher numbered node
	       to a lower numbered node */
	    pfg_node_list_iter pred_iter(nd->preds());
	    while (!pred_iter.is_empty()) {
		pfg_node *pred = pred_iter.step();
		nd->remove_pred(pred);
	    }
	    assert(nd->succs()->is_empty() && nd->preds()->is_empty());

	    if (delete_instrs) pfg_cleanup(nd); 

	    delete nd;
	}
    }

    delete nds;
    nds = new_nds;
    delete lp_depth; 
    lp_depth = NULL; 
}

static void
unreachable_visit (pfg_node *n, bit_set *mark)
{
    /* mark the node as reachable */
    mark->add(n->number());

    /* visit its successors */
    pfg_node_list_iter succ_iter(n->succs());
    while (!succ_iter.is_empty()) {
	pfg_node *succ = succ_iter.step();
	if (!mark->contains(succ->number())) {
	    unreachable_visit(succ, mark);
	}
    }
}

void 
pfg::pfg_cleanup(pfg_node *nd)
{
    if (nd->is_block()) {
	pfg_block *blk = (pfg_block *)nd; 

	pfg_node *olsucc = blk->lsucc; 
	pfg_node *olpred = blk->lpred; 
	if (blk->lsucc)
	    blk->lsucc->lpred = olpred; 
	if (blk->lpred)
	    blk->lpred->lsucc = olsucc; 

	tree_instr *h = blk->in_head(); 
	tree_instr *t = blk->in_tail(); 
	while (h != t) {
	    tree_node *n = h->list_e()->next()->contents; 
	    assert(n->is_instr()); 
	    delete blk->tnl->remove(h->list_e()); 
	    delete h;
	    h = (tree_instr *)n; 
	}
	delete blk->tnl->remove(h->list_e()); 
	delete h;
	if (blk->tiS) {
	    delete blk->tnl->remove(blk->tiS->list_e());
	    delete blk->tiS;
	}
    }
}


/* 
 * optimize_jumps() -- Replace jumps to jumps.  Works in
 * two passes.  First identifies blocks that end with
 * an unconditional jump, but are otherwise vacuous.
 * (Vacuous means: containing only labels, "line" pseudo-
 * ops, and null instructions.)  Then replaces jumps to
 * such blocks by jumps to their unique successors.
 * 
 * Returns TRUE if some jump was replaced; FALSE otherwise.
 * This allows callers to iterate until all jumps to jumps
 * have been removed. 
 */
boolean 
pfg::optimize_jumps() 
{
    boolean did_work = FALSE; 
    int nn = num_nodes(); 
    int i; 
    label_sym **single_jump_target = new (label_sym *[nn]);
    
    for (i = 0; i < nn; ++i) {
	pfg_node *scn = (*this)[i]; 

	single_jump_target[i] = NULL;
	if (scn->is_block()) {
	    pfg_block *blk = (pfg_block *)scn; 
	    tree_instr *cti; 
	    
	    if ((cti = blk->in_cti()) && (Is_ubr(cti->instr()))
	    ||  (cti = blk->in_shad()))
	    {
		tree_node_list_e *ctile = cti->list_e();
		tree_node_list_e *tnle = blk->in_head()->list_e();

		while (tnle != ctile) {
		    tree_instr *ti = (tree_instr *)tnle->contents;
		    instruction *in = ti->instr();

		    if (Is_label(in) || in->peek_annote(k_line) || Is_null(in))
			tnle = tnle->next();
		    else
			break;
		}
		if (tnle == ctile)
		    single_jump_target[i] =
			(label_sym *)Get_target(cti->instr());
	    }
	}
    }

    for (i = 0; i < nn; ++i) {
	pfg_node *scn = (*this)[i]; 
	if (!scn->is_block()) 
	    continue; 
	
	pfg_block *blk = (pfg_block *)scn; 
	for (int j = 0; j < blk->succs()->count(); ++j) {
	    int targ_num = ((*blk->succs())[j])->number(); 
	    if (single_jump_target[targ_num]) {
		pfg_node *new_succ = (pfg_node *)(single_jump_target[targ_num]
		    ->peek_annote(k_pfg_node)); 
		assert(new_succ); 
		blk->set_succ(j, new_succ); 
		did_work = TRUE; 
	    }
	}
    }

    return did_work; 
}

/* 
 * remove_shadows() -- Walks through the blocks of the
 * procedure and removes shadow instructions if they jump to
 * the immediately following block.  
 *
 * This function should not be used until after all set_succ()
 * and set_layout_succ() calls have been made.  set_succ()
 * and set_layout_succ() expect shadow instructions to be 
 * in place; they may break if the shadows have been removed. 
 */
void 
pfg::remove_shadows()
{
    unsigned i; 
    for (i = 0; i < num_nodes(); ++i) {
	pfg_block *blk = (pfg_block *)((*this)[i]); 
	if (!blk->is_block() || blk->tiS == NULL)
	    continue; 
	tree_node_list_e *tnle = blk->tiS->list_e()->next(); 
	if (!tnle)
	    continue; 
	tree_node *tn = tnle->contents; 
	if (!tn->is_instr())
	    continue; 

	pfg_node *nod = pfg_begin_node(tn); 
	assert(nod); 
	if (nod == blk->succs()->head()->contents) {
	    assert(blk->layout_succ() == NULL); 
	    blk->set_layout_succ(nod); 
	}
    }
}

/* 
 * remove_layout_info() -- Clear all the layout links between
 * pfg_node's.  Uses clear_layout_succ(), so shadow instructions
 * may be inserted.  
 */
void 
pfg::remove_layout_info()
{
    unsigned i; 
    for (i = 0; i < num_nodes(); ++i) {
	pfg_node *scn = (*this)[i]; 
	scn->clear_layout_succ(); 
    }
    for (i = 0; i < num_nodes(); ++i) {
	pfg_node *scn = (*this)[i]; 
	assert(scn->layout_pred() == NULL); 
    }
}


/*
 *  Build a list of the flow graph nodes in reverse postorder.
 *  When forward == FALSE, build it in reverse, giving plain
 *  postorder.
 */

static void rpostord_visit(pfg_node_list *result, pfg_node *n, bit_set *mark,
			   boolean forward);

pfg_node_list *
pfg::reverse_postorder_list (boolean forward)
{
    pfg_node_list *result = new pfg_node_list;
    bit_set mark(0, num_nodes());
    rpostord_visit(result, entry_node(), &mark, forward);
    return result;
}


void
rpostord_visit (pfg_node_list *result, pfg_node *n, bit_set *mark,
		boolean forward)
{
    pfg_node_list_iter iter(n->succs());
    while (!iter.is_empty()) {
	pfg_node *s = iter.step();

	/* check if the node has already been visited */
	if (mark->contains(s->number())) continue;

	mark->add(s->number());
	rpostord_visit(result, s, mark, forward);
    }

    /* add this block to the front (for rpo) or back (plain po)
     * of the result list */
    if (forward)
	result->push(n);
    else
        result->append(n);
}


/*
 *  Find the dominators of each node.  The "remove_unreachable_blocks"
 *  method must be called prior to this or the results may be incorrect.
 */

void
pfg::find_dominators ()
{
    /* delete any old information */
    if (doms) delete[] doms;
    if (idom) delete[] idom;

    /* allocate and initialize the new bit vectors */
    doms = _dominators(TRUE);
    idom = immed_dominators(TRUE);
}


/*
 *  Find the postdominators of each node.  A legal graph should never have
 *  unreachable nodes in the reverse graph.  If that is not true, this will
 *  not work correctly.
 */

void
pfg::find_postdominators ()
{
    /* delete any old information */
    if (pdoms) delete[] pdoms;
    if (ipdom) delete[] ipdom;

    /* compute the dominators */
    pdoms = _dominators(FALSE);
    ipdom = immed_dominators(FALSE);
}


/*
 *  Compute (post)dominators.  This is a protected method that is used
 *  internally to do the actual work of finding dominators.
 */

bit_set *
pfg::_dominators (boolean forward)
{
    pfg_node_list *node_list = reverse_postorder_list(forward);
    boolean changed = FALSE;
    unsigned num = num_nodes();

    /* allocate and initialize the new bit vectors */
    bit_set *d = new bit_set[num];
    for (unsigned i = 0; i < num; i++) {
	d[i].expand(0, num);
	d[i].universal();
    }

    /* set up the first node */
    pfg_node *start = (forward ? entry_node() : exit_node());
    d[start->number()].clear();
    d[start->number()].add(start->number());

    bit_set t(0, num_nodes());

    /* iterate until no changes */
    do {
	changed = FALSE;
	pfg_node_list_iter node_iter(node_list);
	while (!node_iter.is_empty()) {

	    pfg_node *n = node_iter.step();
	    if (n == start) continue;

	    /* get intersection of predecessors of n */
	    t.universal();
	    pfg_node_list_iter pred_iter(forward ? n->preds() : n->succs());
	    while (!pred_iter.is_empty()) {
		pfg_node *p = pred_iter.step();
		t *= d[p->number()];
	    }

	    /* include itself in dominator set */
	    t.add(n->number());

	    /* check if there were any changes */
	    if (d[n->number()] != t) {
		d[n->number()] = t;
		changed = TRUE;
	    }
	}
    } while (changed);

    delete node_list;
    return d;
}


/*
 *  Find the immediate (post)dominators.  This is a protected method that
 *  is only used internally to do the actual work of finding immediate
 *  dominators.
 */

pfg_node **
pfg::immed_dominators (boolean forward)
{
    unsigned num = num_nodes();
    unsigned root = (forward ? entry_node()->number() : exit_node()->number());
    bit_set *d = (forward ? doms : pdoms);

    pfg_node **im = new pfg_node*[num];
    im[root] = NULL;

    for (unsigned n = 0; n < num; n++) {

	/* the root node has no dominators */
	if (n == root) continue;

	/* remove self for comparison */
	d[n].remove(n);

	/* check nodes in dominator set */
	boolean found = FALSE;
	bit_set_iter diter(&d[n]);
	while (!diter.is_empty()) {
	    unsigned i = diter.step();

	    if (i < num && d[n] == d[i]) {
		found = TRUE;
		im[n] = node(i);
	    }
	}

	if (!found) {
	    warning_line(NULL, "can't find idom of node %u", n);
	    im[n] = NULL;
	}

	/* re-add self when done */
	d[n].add(n);
    }

    return im;
}


/*
 *  Find the dominance frontier of each node.  The dominators must have
 *  already been computed.
 */

void
pfg::find_df ()
{
    assert_msg(doms, ("pfg::find_df - need to compute dominators first"));

    /* delete any old information */
    if (df) delete[] df;

    df = new bit_set[num_nodes()];
    dom_frontiers(entry_node(), TRUE);
}


/*
 *  Find the reverse dominance frontier of each node.  The postdominators
 *  must have already been computed.
 */

void
pfg::find_reverse_df ()
{
    assert_msg(pdoms, ("pfg::find_reverse_df - need postdominators first"));

    /* delete any old information */
    if (rdf) delete[] rdf;

    rdf = new bit_set[num_nodes()];
    dom_frontiers(exit_node(), FALSE);
}


/*
 *  Compute dominance frontiers using a postorder traversal of the dominator
 *  tree.  This is a protected method that is only used internally.
 */

void
pfg::dom_frontiers (pfg_node *x, boolean forward)
{
    unsigned n, num = num_nodes();
    pfg_node **xidom;
    bit_set *dom_front;
    pfg_node_list *nodes;

    if (forward) {
	xidom = idom;
	dom_front = df;
	nodes = x->succs();
    } else {
	xidom = ipdom;
	dom_front = rdf;
	nodes = x->preds();
    }

    /* visit all children (i.e. immediate dominatees) first */
    for (n = 0; n < num; n++) {
	if (xidom[n] == x) {
	    dom_frontiers(node(n), forward);
	}
    }

    /* calculate dominance frontier, from paper RCytron_89 */
    dom_front[x->number()].expand(0, num);

    /* local loop, uses PFG */
    pfg_node_list_iter succ_iter(nodes);
    while (!succ_iter.is_empty()) {
	pfg_node *s = succ_iter.step();

	if (x != xidom[s->number()]) {
	    dom_front[x->number()].add(s->number());
	}
    }

    /* up loop, uses dominator tree */
    for (n = 0; n < num; n++) {
	if ((xidom[n] == x) && !dom_front[n].is_empty()) {
	    for (unsigned y = 0; y < num; y++) {
		if (dom_front[n].contains(y) && (x != xidom[y])) {
		    dom_front[x->number()].add(y);
		}
	    }
	}
    }
}


/* Access routines for dominators and dominance frontiers. */

boolean
pfg::dominates (int n_dominator, int n_dominatee)
{
    assert_msg(doms, ("pfg::dominates() -- run find_dominators() first"));
    return doms[n_dominatee].contains(n_dominator);
}

boolean
pfg::dominates (pfg_node *dominator, pfg_node *dominatee)
{
    return dominates(dominator->number(), dominatee->number());
}

boolean
pfg::postdominates (int n_dominator, int n_dominatee)
{
    assert_msg(pdoms, ("pfg::postdominates() -- "
		       "run find_postdominators() first"));
    return pdoms[n_dominatee].contains(n_dominator);
}

boolean
pfg::postdominates (pfg_node *dominator, pfg_node *dominatee)
{
    return postdominates(dominator->number(), dominatee->number());
}

/* Return the set of dominators of node n */
bit_set *
pfg::dominators(int n)
{
    assert_msg(doms, ("pfg::dominators() -- run find_dominators() first"));
    return &(doms[n]);
}

bit_set *
pfg::dominators(pfg_node *n)
{ 
    return dominators(n->number());
}

/* Return the set of post-dominators of node n */
bit_set *
pfg::postdominators(int n)
{
    assert_msg(pdoms, ("pfg::postdominators() -- run find_postdominators() first"));
    return &(pdoms[n]);
}

bit_set *
pfg::postdominators(pfg_node *n)
{ 
    return postdominators(n->number());
}

pfg_node *
pfg::immed_dom (int n)
{
    assert_msg(idom, ("pfg::immed_dom() -- run find_dominators() first"));
    return idom[n];
}

pfg_node *
pfg::immed_dom (pfg_node *n)
{
    return immed_dom(n->number());
}

pfg_node *
pfg::immed_postdom (int n)
{
    assert_msg(ipdom, ("pfg::immed_postdom() -- "
		       "run find_postdominators() first"));
    return ipdom[n];
}

pfg_node *
pfg::immed_postdom (pfg_node *n)
{
    return immed_postdom(n->number());
}

bit_set *
pfg::dom_frontier (int n)
{
    assert_msg(df, ("pfg::dom_frontier() -- run find_df first"));
    return &df[n];
}

bit_set *
pfg::dom_frontier (pfg_node *n)
{
    return dom_frontier(n->number());
}

bit_set *
pfg::reverse_dom_frontier (int n)
{
    assert_msg(rdf, ("pfg::reverse_dom_frontier() -- "
		     "run find_reverse_df() first"));
    return &rdf[n];
}

bit_set *
pfg::reverse_dom_frontier (pfg_node *n)
{
    return reverse_dom_frontier(n->number());
}


void 
pfg::find_natural_loops()
{
    static char *__FNAME = "pfg::find_natural_loops";

    int i;
    int num = num_nodes();
    int *node_stack = new int[num]; 

    if (lp_depth)
	delete[] lp_depth; 
    lp_depth = new int[num]; 

    assert((doms != NULL) && (pdoms != NULL));

    /* initialize */
    for (i = 0; i < num; i++) 
	lp_depth[i] = 0; 

    /* find all the loop entries and loop-back blocks */
    for (i = 0; i < num; i++) {
	pfg_node *cn = node(i);
	if (cn->is_block() || cn->is_test()) {
	    int top_stack = 0; 
	    bit_set this_loop(0, num); 

	    /* if some predecessor induces a back edge, 
	     * then use it to seed the loop
	     */
	    if (debug(__FNAME, 1)) {
		printf("%s(): Checking if predecessors of %d induce a back"
			" edge\n", __FNAME, cn->number()
			);
	    }

	    pfg_node_list_iter pred_iter(cn->preds());
	    while(!pred_iter.is_empty()) {
		pfg_node *pred_cn = pred_iter.step();
		if (debug(__FNAME, 2)) {
		    printf("\tchecking predecessor %d\n", pred_cn->number());
		}
		if (dominates(cn, pred_cn)) {
		    if (debug(__FNAME, 2)) {
			printf("\tYes. %d is dominated by %d. Add it to the"
				" stack\n", pred_cn->number(), cn->number());
		    }
		    node_stack[top_stack++] = pred_cn->number(); 
		}
	    }

	    /* there was some back edge, so this is a header */
	    if (top_stack) {
		++lp_depth[i]; 
		this_loop.add(i); 
	    }

	    /* walk backwards around loop, adding nodes */
	    while (top_stack) {
		int top = node_stack[--top_stack]; 
		if (!this_loop.contains(top)) {
		    if (debug(__FNAME, 2)) {
			printf("%s(): Incrementing lp_depth for node %d\n",
				__FNAME, top);
		    }
		    this_loop.add(top); 
		    ++lp_depth[top]; 
		    pfg_node_list_iter top_iter(node(top)->preds());
		    while (!top_iter.is_empty()) {
			int m = top_iter.step()->number(); 
			node_stack[top_stack++] = m; 
			assert(top_stack < num); 
		    }
		}
	    }
	}
    }

    delete[] node_stack; 
}

int
pfg::loop_depth (int n)
{
    assert_msg(lp_depth, ("pfg::loop_depth() -- "
	"run find_natural_loops() first"));
    return lp_depth[n];
}

int
pfg::loop_depth (pfg_node *n)
{
    return loop_depth(n->number());
}

void 
pfg::set_loop_depth(int n, int d)
{
    assert_msg(lp_depth, ("pfg::loop_depth() -- "
	"run find_natural_loops() first"));
    lp_depth[n] = d;
}

void 
pfg::set_loop_depth(pfg_node *n, int d)
{
    set_loop_depth(n->number(), d);
}


boolean
pfg::is_loop_begin (int n)
{
    return is_loop_begin(node(n));
}

boolean
pfg::is_loop_begin (pfg_node *cn)
{
    pfg_node_list_iter pred_iter(cn->preds());
    while (!pred_iter.is_empty()) {
	pfg_node *pred_cn = pred_iter.step();
	if (dominates(cn, pred_cn)) return TRUE;
    }
    return FALSE;
}

boolean
pfg::is_loop_end (int n)
{
    return is_loop_end(node(n));
}

boolean
pfg::is_loop_end (pfg_node *cn)
{
    pfg_node_list_iter succ_iter(cn->succs());
    while (!succ_iter.is_empty()) {
	pfg_node *succ_cn = succ_iter.step();
	if (dominates(succ_cn, cn)) return TRUE;
    }
    return FALSE;
}

boolean
pfg::is_loop_exit (int n)
{
    return is_loop_exit(node(n));
}

boolean
pfg::is_loop_exit (pfg_node *cn)
{
    pfg_node_list_iter succ_iter(cn->succs());
    while (!succ_iter.is_empty()) {
	pfg_node *succ_cn = succ_iter.step();
	if (loop_depth(succ_cn) != loop_depth(cn)) return TRUE;
    }
    return FALSE;
}

//---------------------------------------------------------------------------
// pfg::outermostLoopHeader()
// Returns the header node for the outermost loop enclosing the given node.
// It works by going up in the dominator tree looking for loop header
// nodes. We stop when we find the first node header with depth 1.
//---------------------------------------------------------------------------
pfg_node *
pfg::outermostLoopHeader(pfg_node *node)
{
    static char *__FNAME = "pfg::outermostLoopHeader";

    assert_msg(lp_depth, ("Run pfg::find_natural_loops() first\n"));

    /* Check trivial cases first. If the node is not inside a loop, return
     * NULL. If the node is already an outermost loop header, return it.
     */
    if (this->loop_depth(node) < 1) {
	return NULL;
    }

    if (this->is_loop_begin(node) && this->loop_depth(node) == 1) {
	return node;
    }

    /* Walk up the dominator tree looking for a loop header with loop depth
     * of 1.
     */
    pfg_node *header = NULL;
    pfg_node *candidate = this->immed_dom(node);

    while (candidate != NULL) {

	if (debug(__FNAME, 2)) {
	    printf("%s(): Checking dominator(%d) = %d\n", __FNAME, 
		    node->number(), candidate->number());
	}

	if (this->is_loop_begin(candidate)) {
	    if (debug(__FNAME, 2)) {
		printf("\tNode %d is a loop header and dominates node %d\n", 
			candidate->number(), node->number());
	    }
	    header = candidate;
	    if (this->loop_depth(header) == 1) {
		break;			/* We found the outermost header */
	    }
	}
	candidate = this->immed_dom(candidate);
    }

    return header;
}



/**
 ** ------------- Protected functions ------------- 
 **/

/* 
 * add_node() -- add a node to a pfg, without any error checking.  
 */
void
pfg::add_node (pfg_node *n)
{
    int num = nds->extend((void *)n);
    n->set_number((unsigned)num);
    n->set_parent(this);
    lp_depth = (int *)realloc(lp_depth, (num + 1) * sizeof(int)); 
    assert_msg(lp_depth != NULL, ("pfg::add_node -- "
	"unable to reallocate lp_depth array")); 
}


/* 
 * graph_add_list() -- pfg constructor helper for handling
 * tree_node_lists.  Has separate cases for maximal basic
 * block and single-instruction basic block cases.  Mutually
 * recursive with graph_add_tree() to support high-SUIF
 * structures.  
 */
pfg_node *
pfg::graph_add_list (tree_node_list *list, pfg_node *prev, 
    boolean block, boolean break_at_call, boolean keep_l, boolean parallel,
    threadBody *thread, pfg_node *par_parent, boolean firstNodeOfThread)
{
    annote *an;

    tree_node_list_iter iter(list);

    /* If this is the first node of a thread, we set the iterator to point
     * to the beginning of the thread. */
    if (firstNodeOfThread) {
	iter.set(thread->first()->list_e());
    }

    while (!iter.is_empty()) {
	tree_node *tn = iter.step();

	/* check for and deal with non-instruction AST nodes */
	if (!tn->is_instr()) {
	    prev = graph_add_tree(tn, prev, block, break_at_call, keep_l,
		    par_parent, thread);
	    continue;
	}

	/* check for a cobegin node which is represented by an mbr instr.
	 * If we have a cobegin node, call graph_add_tree to add the
	 * parallel begin and end nodes, add the threads and the join node
	 * (ie, the default label) at the end.
	 */
	if (isCobegin(tn)) {
	    unsigned i;
	    tree_cobegin *cobegin = new tree_cobegin(tn);
	    pfg_node *cobegin_node, *coend_node;

	    /* Add parallel begin and end markers */
	    cobegin_node = graph_add_tree(tn, prev, block, break_at_call, 
		    keep_l, par_parent, thread);

	    coend_node = pfg_end_node(tn);

	    /* Add node for mbr instruction. Create a parallel PFG node */
	    tree_instr *ti = (tree_instr *)tn;
	    pfg_block *blknode = new pfg_block(list, ti, ti, TRUE,
		    cobegin_node, thread);
	    add_node(blknode);
	    blknode->set_in_cti(ti); 
	    if (cobegin_node) cobegin_node->add_succ(blknode);

	    an = new annote(k_pfg_node, (void *)blknode);
	    ti->annotes()->append(an);

	    /* Add bodies for the different threads */
	    pfg_node *last_thread;
	    for (i = 0; i < cobegin->numThreads(); i++) {
		last_thread = graph_add_list(list, NULL, block,
			break_at_call, keep_l, FALSE, cobegin->thread(i),
			cobegin_node, TRUE);
	    }

	    /* Add a node for the default label. */
	    iter.set(cobegin->last()->list_e());
	    tn = iter.step();
	    ti = (tree_instr *)tn;
	    pfg_block *instrnode = new pfg_block(list, ti, ti, TRUE,
		    cobegin_node, thread);
	    add_node(instrnode);

	    /* record the label position */
	    instruction *instr = ti->instr();
	    label_sym *lab = Get_label(instr);
	    an = new annote(k_pfg_node, (void *)instrnode);
	    lab->annotes()->append(an);

	    /* record the node for the instruction */
	    an = new annote(k_pfg_node, (void *)instrnode);
	    ti->annotes()->append(an);

	    /* Finally, link the default label node to the coend node.
	     * Also, add a fall-through edge from the last thread to the
	     * default label node.
	     */
	    instrnode->add_succ(coend_node);
	    last_thread->add_succ(instrnode);
	    prev = coend_node;

	    continue;
	}

	/* the rest of this code deals with instructions: */

	tree_instr *ti = (tree_instr *)tn;
	tree_instr *tiFirst = ti; 
	instruction *i = ti->instr();

	/* first handle the easy case of nodes for individual instructions */

	if (!block) {

	    /* create a new node for the instruction */
	    pfg_instr *inode = new pfg_instr(ti, parallel, par_parent,
		    thread);
	    an = new annote(k_pfg_node, (void *)inode);
	    ti->annotes()->append(an);

	    add_node(inode);
	    if (prev) 
		prev->add_succ(inode);

	    /* record the positions of labels */
	    if (Is_label(i)) {
		label_sym *lab = Get_label(i);
		an = new annote(k_pfg_node, (void *)inode);
		lab->annotes()->append(an);
	    }

	    /* check if the instruction falls through to its successor. */
	    if (Is_cti(i)) {
	    	if (Is_cbr(i) || Is_call(i))
		    prev = inode;
		else if (Is_ubr(i) || Is_mbr(i) || Is_return(i))
		    prev = NULL;
		else
		    assert(0); 
	    } else {
		prev = inode; 
	    }

	    continue;
	}
	
	/* now comes the harder task of building basic blocks.... */

	boolean end_block = FALSE;
	pfg_block *blknode = new pfg_block(list, ti, ti, parallel,
		par_parent, thread);

	add_node(blknode);
	if (prev) prev->add_succ(blknode);
	prev = blknode;

	/* check for label instructions */
	while (Is_label(i)) {
	    /* record the label position */
	    label_sym *lab = Get_label(i);
	    an = new annote(k_pfg_node, (void *)blknode);
	    lab->annotes()->append(an);

	    /* record the node for the instruction */
	    an = new annote(k_pfg_node, (void *)blknode);
	    ti->annotes()->append(an);

	    /* read the next instruction */
	    if (iter.is_empty() || !iter.peek()->is_instr()) {
		end_block = TRUE;
		break;
	    }
	    ti = (tree_instr *)iter.step();
	    blknode->extend(ti);
	    assert(ti->is_instr());
	    i = ti->instr();
	}

	/* if there's nothing but labels then this is the end */
	if (end_block) continue;

	while (TRUE) {

	    /* record the node for the instruction */
	    an = new annote(k_pfg_node, (void *)blknode);
	    ti->annotes()->append(an);

	    /* break if next instruction is cobegin */
	    if (iter.is_empty() || isCobegin((tree_instr *)iter.peek())) {
		break;
	    }

	    /* break if we are at the end of a cobegin thread */
	    if (thread && thread->last() == (tree_node *)ti) {
		break;
	    }

	    /* check for branch instruction that don't fall through. */
	    if (Is_cti(i) && !Is_cbr(i) && !Is_call(i)) {
		prev = NULL;
		break;
	    }

	    /* check for other branch instructions */
	    if (Is_cbr(i))
		break;

	    /* break basic blocks at calls if flag is set */
	    if (Is_call(i) && break_at_call)
		break; 

	    /* check if there is another non-label instruction after this */
	    if (iter.is_empty() ||
		!(ti = (tree_instr *)iter.peek())->is_instr() ||
		(Is_label(ti->instr())))
		break;

	    /* read the next instruction */
	    ti = (tree_instr *)iter.step();
	    blknode->extend(ti);
	    assert(ti->is_instr());
	    i = ti->instr();
	}

	/* Handle shadow instructions at the end of blocks */
	if (Is_ubr(i)) {
	    if (tiFirst == ti) {
		/* Insert label to make up body of block */
		tree_proc *tp = tproc();
		proc_symtab *par_symtab = tp->proc_syms(); 
		label_sym *new_lab = par_symtab->new_unique_label(); 
		new_lab->annotes()->append(new annote(k_pfg_node,
		    (void *)blknode));
		instruction *new_i = new in_lab(new_lab);
		tree_instr *new_ti = new tree_instr(new_i); 
		new_ti->annotes()->append(new annote(k_pfg_node, 
		    (void *)blknode)); 
		list->insert_before(new_ti, ti->list_e()); 
		blknode->set_in_head(new_ti); 
	    } 
	    if (keep_l) {
		blknode->set_in_cti(ti); 
	    } else {
		blknode->set_in_tail((tree_instr *)ti->list_e()
		    ->prev()->contents); 
		blknode->set_in_cti(NULL); 
		blknode->tiS = ti; 
	    }
	} else if (Is_cbr(i) || Is_call(i)) {
	    blknode->set_in_cti(ti); 
	    /* check for shadow goto */
	    tree_instr *tiPossible; 
	    if (!keep_l
		&& !iter.is_empty() 
		&& (tiPossible = (tree_instr *)iter.peek())->is_instr() 
		&& Is_ubr(i = tiPossible->instr())) {
		iter.step(); 
		an = new annote(k_pfg_node, (void *)blknode);
		tiPossible->annotes()->append(an);
		blknode->tiS = tiPossible; 
	    }
	} else if (Is_mbr(i) || Is_return(i)) {
	    blknode->set_in_cti(ti); 
	} else if (Is_cti(i)) {
	    assert_msg(FALSE, ("graph_add_list() -- "
		"Unknown cti instruction type")); 
	}

	if (thread && thread->last() == (tree_node *)ti) {
	    break;
	}
    }

    return prev;
}


/* 
 * graph_add_tree() -- pfg constructor helper for handling
 * tree_nodes.  Separate cases handle different high-SUIF
 * constructs.  Mutually recursive with graph_add_list()
 * to handle the tree_node_list's embedded in high-SUIF
 * constructs.  
 */
pfg_node *
pfg::graph_add_tree (tree_node *t, pfg_node *prev, 
    boolean block, boolean break_at_call, boolean keep_l, pfg_node
    *par_parent, threadBody *thread)
{
    pfg_node *node;
    pfg_node *save = NULL;
    pfg_node *save_contlab = NULL;
    pfg_node *save_toplab = NULL;
    pfg_node *save_jumpto = NULL;
    annote *an;

    /* Is this a parallel or a sequential AST node? */
    boolean parallel = (isCobegin(t) || isDoall(t));

    /* add the BEGIN marker node */
    node = new pfg_begin(t, parallel, par_parent, thread);
    an = new annote(k_pfg_begin, (void *)node);
    t->annotes()->append(an);
    add_node(node);
    if (prev) prev->add_succ(node);
    prev = node;

    /* add the END marker node. If this is a coend node then its parallel
     * parent is the cobegin node we just created, otherwise its parallel
     * parent is whatever we were told.
     */
    pfg_node *end = new pfg_end(t, parallel, (parallel) ? node :
	    par_parent, thread);
    an = new annote(k_pfg_end, (void *)end);
    t->annotes()->append(an);
    add_node(end);

    switch (t->kind()) {

	case TREE_BLOCK: {
	    tree_block *tb = (tree_block *)t;
	    prev = graph_add_list(tb->body(), prev, block, break_at_call, 
				    keep_l, parallel, thread, par_parent);
	    
	    /* If this is a sub-block inside the procedure then we join the
	     * end node to the last node of the sub-block and return the
	     * end node, so the next pfg node will come after it. This is
	     * not done for the topmost block because the procedure's end
	     * node does not have any successors.
	     */
	    if (tb->parent() != NULL) {
		if (prev) prev->add_succ(end);
		node = end;
	    }
	    break;
	}

	case TREE_FOR: {
	    tree_for *tf = (tree_for *)t;
	    prev = graph_add_list(tf->lb_list(), prev, 
			    block, break_at_call, keep_l, parallel, thread,
			    par_parent); 

	    prev = graph_add_list(tf->ub_list(), prev, 
			    block, break_at_call, keep_l, parallel, thread,
			    par_parent);

	    save = graph_add_list(tf->step_list(), prev, 
			    block, break_at_call, keep_l, parallel, thread,
			    par_parent);

	    prev = graph_add_list(tf->landing_pad(), save, 
			    block, break_at_call, keep_l, parallel, thread,
			    par_parent);


	    /* create a separate node for the non-existent "top" label */
	    if (!block) {
		node = new pfg_label(t, parallel, par_parent, thread);
		an = new annote(k_pfg_toplab, (void *)node);
		t->annotes()->append(an);
		add_node(node);
		if (prev) prev->add_succ(node);
		prev = node;
		save_toplab = node;
	    } else {
		prev = NULL;
	    }

	    prev = graph_add_list(tf->body(), prev, 
		block, break_at_call, keep_l, FALSE, thread, par_parent);

	    /* create a separate node for the "continue" label */
	    if (!block) {
		node = new pfg_label(t, parallel, par_parent, thread);
		add_node(node);
		if (prev) prev->add_succ(node);
		prev = node;
		save_contlab = node;
	    }

	    /* create a new node for the "test" part */
	    node = new pfg_test(tf, parallel, par_parent, thread);
	    an = new annote(k_pfg_test, (void *)node);
	    t->annotes()->append(an);
	    add_node(node);
	    if (prev) prev->add_succ(node);
	    prev = node;

	    break;
	}

	case TREE_LOOP: {
	    tree_loop *tl = (tree_loop *)t;

	    /* create a separate node for the "top" label */
	    if (!block) {
		node = new pfg_label(t, parallel, par_parent, thread);
		add_node(node);
		if (prev) prev->add_succ(node);
		prev = node;
		save_toplab = node;
	    }

	    prev = graph_add_list(tl->body(), prev, 
		block, break_at_call, keep_l, parallel, thread, par_parent);

	    /* create a separate node for the "continue" label */
	    if (!block) {
		node = new pfg_label(t, parallel, par_parent, thread);
		add_node(node);
		if (prev) prev->add_succ(node);
		prev = node;
		save_contlab = node;
	    }

	    prev = graph_add_list(tl->test(), prev, 
		block, break_at_call, keep_l, parallel, thread, par_parent);
	    break;
	}

	case TREE_IF: {
	    tree_if *ti = (tree_if *)t;

	    prev = graph_add_list(ti->header(), prev, block, break_at_call, 
		    keep_l, parallel, thread, par_parent);
	    save = graph_add_list(ti->then_part(), prev, block,
		    break_at_call, keep_l, parallel, thread, par_parent);
	    prev = NULL;

	    /* create a separate node for the "jumpto" label */
	    if (!block && !ti->else_part()->is_empty()) {
		node = new pfg_label(t, parallel, par_parent, thread);
		add_node(node);
		prev = node;
		save_jumpto = node;
	    } else {
		node = end;
	    }

	    prev = graph_add_list(ti->else_part(), prev, block,
		    break_at_call, keep_l, parallel, thread, par_parent);

	    break;
	}

	default: {
	    if (isCobegin(t)) {
		/* Do nothing, ::graph_add_list takes care of adding the
		 * node for the mbr instruction representing the cobegin.
		 */
		break;
	    } else {
		assert(FALSE);
	    }
	}
    }

    // PROBLEM: don't want non-return instructions that are
    // garbage at the end of the procedure to show end as a
    // successor, because they don't actually return.  But we
    // need the next line for TREE_FOR's, etc. where flow
    // falls out of the end.  Commented out for now.  
    // if (prev) prev->add_succ(end);

    switch (t->kind()) {

	case TREE_FOR: {
	    tree_for *tf = (tree_for *)t;
	    
	    /* mark the position of the "continue" label */
	    if (block) save_contlab = pfg_test_node(tf);
	    an = new annote(k_pfg_node, (void *)save_contlab);
	    tf->contlab()->annotes()->append(an);

	    /* the "break" label is at the END node */
	    an = new annote(k_pfg_node, (void *)node);
	    tf->brklab()->annotes()->append(an);

	    /* add an edge from before the landing pad to the END */
	    assert(save);
	    save->add_succ(node);

	    /* add an edge from the test node to the top of the loop */
	    if (block) {
		save_toplab = save_contlab;
		if (!tf->body()->is_empty()) {
		    save_toplab = pfg_begin_node(tf->body()->head()->contents);
		}
	    }
	    pfg_test_node(tf)->add_succ(save_toplab);

	    /* Add an edge from the test node to the end of the loop */
	    pfg_test_node(tf)->add_succ(end);
 	    node = end;

	    break;
	}

	case TREE_LOOP: {
	    if (prev) prev->add_succ(end);

	    tree_loop *tl = (tree_loop *)t;

	    /* find the position of the "continue" label */
	    if (block) {
		save_contlab = node;
		if (!tl->test()->is_empty()) {
		    save_contlab =pfg_begin_node(tl->test()->head()->contents);
		}
	    }
	    an = new annote(k_pfg_node, (void *)save_contlab);
	    tl->contlab()->annotes()->append(an);

	    /* find the position of the "top" label */
	    if (block) {
		save_toplab = save_contlab;
		if (!tl->body()->is_empty()) {
		    save_toplab = pfg_begin_node(tl->body()->head()->contents);
		}
	    }
	    an = new annote(k_pfg_node, (void *)save_toplab);
	    tl->toplab()->annotes()->append(an);

	    /* the "break" label is at the END node */
	    an = new annote(k_pfg_node, (void *)end);
	    tl->brklab()->annotes()->append(an);

	    node = end;

	    break;
	}

	case TREE_IF: {
	    tree_if *ti = (tree_if *)t;

	    /* the "jumpto" label is at the beginning of the ELSE part */
	    if (ti->else_part()->is_empty()) {
		save_jumpto = node;
	    } else if (block) {
		save_jumpto = pfg_begin_node(ti->else_part()->head()->contents);
	    }
	    an = new annote(k_pfg_node, (void *)save_jumpto);
	    ti->jumpto()->annotes()->append(an);

	    /* the THEN part skips over the ELSE part */
	    if (save) save->add_succ(node);

	    /* the ELSE part falls through to the END node */
	    if (!ti->else_part()->is_empty()) {
		pfg_node *else_end;
		else_end = pfg_end_node(ti->else_part()->head()->contents);
		else_end->add_succ(node);
	    }

	    break;
	}

	default: {
	    break;
	}
    }

    return node;
}


/**
 ** ------------- Local functions ------------- 
 **/

/*
 *  This function is mapped across the AST to add the branch edges after
 *  the flow graph nodes have been created.  A pointer to the exit node
 *  should be passed in the "x" parameter.
 * 
 *  This function doesn't add fallthrough edges; that's done in 
 *  graph_add_list(). 
 * 
 *  This function should not call any of the set_succ() functions, 
 *  because it has to work for both block and instr mode, and there
 *  is no current way to set_succ() on a pfg_instr object.  
 *  Unlike set_succ(), we don't have to bother to remove old
 *  predecessor pointers.  We're setting them up for the first time.  
 */

static void
add_pfg_edge (tree_node *t, void *x)
{

    if (t->is_instr()) {
	tree_instr *ti = (tree_instr *)t;
	instruction *i = ti->instr();
	pfg_node *this_node = pfg_holding_node(ti);
	assert_msg(this_node, ("add_pfg_edge - cannot find node for "
	    "instruction (%d)", ti->instr()->number()));

	// Multiway branches
	if (Is_mbr(i)) {
	    assert_msg(this_node->succs()->count() == 0, 
		("add_pfg_edge - mbr already has some successors")); 
	    in_mbr *mbri = (in_mbr *)i;
// 	    if (isCobegin(ti) == FALSE) {
		add_edge(ti, mbri->default_lab(), 0);
// 	    }
	    for (unsigned n = 0; n < mbri->num_labs(); n++) 
		add_edge(ti, mbri->label(n), n + 1);
	}

	// Conditional branches 
	if (Is_cbr(i)) {
// 	    assert_msg(this_node->succs()->count() == 1, 
// 		("add_pfg_edge - expected one cbr successor")); 
	    label_sym *target = (label_sym *)Get_target(i); 
	    pfg_node *target_node = label_pfg_node(target);
	    assert_msg(target_node, ("add_pfg_edge - cannot find node for "
		"label `%s'", target->name()));
	    this_node->succs()->append(target_node); 
	    if (!occurs(this_node, target_node->preds()))
		target_node->preds()->append(this_node); 
	}

	// Unconditional branches 
	if (Is_ubr(i)) {
	    label_sym *target = (label_sym *)Get_target(i); 
	    pfg_node *target_node = label_pfg_node(target);
	    assert_msg(target_node, ("add_pfg_edge - cannot find node for "
		"label `%s'", target->name()));

	    if (this_node->is_block()) {
		pfg_block *blk = (pfg_block *)this_node; 
		if (blk->in_shad() == ti) {
		    this_node->set_fall_succ(target_node); 
		} else {
		    assert_msg(this_node->succs()->count() == 0, 
			("add_pfg_edge - expected no ubr successors")); 
		    this_node->succs()->push(target_node); 
		    if (!occurs(this_node, target_node->preds()))
			target_node->preds()->push(this_node); 
		}
	    } else if (this_node->is_instr()) {
		pfg_instr *ins = (pfg_instr *)this_node; 
		assert_msg(ins->succs()->count() == 0, 
		    ("add_pfg_edge - expected no ubr successors (instr)")); 
		ins->succs()->push(target_node); 
		if (!occurs(ins, target_node->preds()))
		    target_node->preds()->push(this_node); 
	    } else {
		assert_msg(FALSE, ("add_pfg_edge - unknown node type")); 
	    }
	}

	// Link return to exit node 
	if (Is_return(i)) {
	    assert_msg(this_node->succs()->count() == 0, 
		("add_pfg_edge - expected no return successors")); 
	    // add an edge to the exit node 
	    this_node->succs()->append((pfg_node *)x); 
	    if (!occurs(this_node, ((pfg_node *)x)->preds()))
		((pfg_node *)x)->preds()->append(this_node); 
	}
    }	
}


/* 
 * add_edge() -- This function should be removed when we do
 * the low-SUIF implementation.  It currently exists to handle
 * low-SUIF mbr's, and hasn't been tested.  We need to come up with
 * a reasonable way to number the default label in low-SUIF mbr's.  
 */
static void
add_edge (tree_instr *ti, label_sym *target, unsigned n)
{
    pfg_node *target_node = label_pfg_node(target);
    assert_msg(target_node, ("graph_update_tree - cannot find node for "
	"label `%s'", target->name()));

    pfg_node *this_node = pfg_holding_node(ti);
    assert_msg(this_node, ("graph_update_tree - cannot find node for "
	"instruction (%d)", ti->instr()->number()));

//     this_node->set_succ(this_node->succs()->count(), target_node);
    this_node->set_succ(n, target_node);
}


void
pfg::print (boolean show_layout)
{
    printf("%d blocks\n", num_nodes()); 

    if (show_layout) {
	bit_set done(0, num_nodes()); 
	for (unsigned i = 0; i < num_nodes(); i++) {
	    if (done.contains(i))
		continue;
	    pfg_node *curr = node(i); 
	    while (curr->layout_pred())
		curr = curr->layout_pred(); 
	    while (curr) {
		curr->print(); 
		done.add(curr->number()); 
		curr = curr->layout_succ(); 
	    }
	}
    } else {
	for (unsigned i = 0; i < num_nodes(); ++i) 
	    node(i)->print(); 
    }
}

/*
 * conc(a, b): Returns TRUE if blocks a and b can execute concurrently.
 * That is, both a and b are in separate threads and share a common
 * cobegin/doall ancestor.
 */
boolean
pfg::conc(pfg_node *a, pfg_node *b)
{
    /* Check if they're both the same node */
    if (a->number() == b->number()) return FALSE;

    /* Check if one of them is not in a thread */
    if (a->par_parent() == NULL || b->par_parent() == NULL) return FALSE;

    /* Check if they are on the same thread */
    if (a->thread() == b->thread()) return FALSE;

    /* Find out the concurrent ancestors for each node, if they share at
     * least one ancestor then they may execute concurrently.
     */
    bit_set *a_panc = a->par_ancestors();
    bit_set *b_panc = b->par_ancestors();

    return (*a_panc ^ *b_panc);
}

/*
 * par_ancestors(): Returns the set of all the parallel ancestors for the
 * given node.
 */
bit_set *
pfg::par_ancestors(pfg_node *n)
{
    bit_set *panc = new bit_set(0, this->num_nodes());
    pfg_node *par_parent = n->par_parent();

    panc->clear();
    
    while (par_parent) {
	panc->add(par_parent->number());
	par_parent = par_parent->par_parent();
    }

    return panc;
}

/*
 * isPath(): Returns TRUE if there is a path from the first node to the
 * second.
 */
boolean
pfg::isPath(pfg_node *from, pfg_node *to)
{
    /* Simplest case, both nodes are the same */
    if (from == to) return TRUE;

    /* Recursive plunge. Depth first search on our successors. Check if
     * there is a path from each of our successors to the node. Eureka jump
     * on the first one that we find.
     */
    boolean foundPath = FALSE;
    pfg_node_list_iter succ_iter(from->succs());
    while (!succ_iter.is_empty()) {
	pfg_node *succ = succ_iter.step();
	foundPath = this->isPath(succ, to);
	if (foundPath == TRUE) {
	    break;
	}
    }

    return foundPath;
}
