/*  Control Flow Graph */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    Copyright (c) 1995, 1996, 1997 The President and Fellows of Harvard 
    University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file.

    Modified by Diego Novillo (diego@cs.ualberta.ca)
*/

#include <suif_copyright.h>


#ifndef PFG_GRAPH_H
#define PFG_GRAPH_H

// #pragma interface

class pfg_node;
class pfg_block;
class pfg_node_list;
class RefConflict_list;
class VarRef_list;
class cssame;

/*
 *  The control flow graph defined here is intended to be general enough
 *  for almost everyone to use (at least as a base -- other information
 *  can be added on top of it).  It supports nodes containing either
 *  individual tree_instrs or basic blocks of contiguous tree_instrs.
 *  High-level tree nodes are represented by multiple flow graph nodes
 *  corresponding to the code that will be created when they are expanded.
 *  Moreover, each high-level tree node is delimited by a pair of BEGIN
 *  and END nodes to make it easier to match up the flow graph with the
 *  SUIF code.
 */

DECLARE_X_ARRAY(pfg_node_array, pfg_node, 10);

class pfg {
friend class pfg_node;
friend class cssame;

private:
    tree_proc *tp;                      /* proc of this pfg */

    pfg_node_array *nds;                /* xarray of nodes */
    pfg_node *en;                       /* entry node */
    pfg_node *ex;                       /* exit node */

    bit_set *doms, *pdoms;              /* dominators and postdominators */
    pfg_node **idom, **ipdom;           /* immediate dominators and postdoms */
    bit_set *df, *rdf;                  /* dominance frontier and reverse df */
    int *lp_depth;                      /* loop depth */
    RefConflict_list *_conflicts;	/* Conflicting references */
    VarRef_list *_refs;			/* Global list of references */

protected:
    /* internal methods for building the graph */
    void add_node(pfg_node *n);
    pfg_node *graph_add_list(tree_node_list *list, pfg_node *prev,
        boolean block, boolean break_at_call, boolean keep_layout, boolean
	parallel = FALSE, threadBody *thread=0, pfg_node *par_parent=0,
	boolean firstNodeOfThread = FALSE);
    pfg_node *graph_add_tree(tree_node *t, pfg_node *prev, 
        boolean block, boolean break_at_call, boolean keep_layout, pfg_node
	*par_parent = 0, threadBody *thread = 0);

    /* helper routine to clone nodes or create new empty nodes */
    void attach(pfg_node *);

    /* dfa functions used internally */
    bit_set *_dominators(boolean forward);
    pfg_node **immed_dominators(boolean forward);
    void dom_frontiers(pfg_node *x, boolean forward);

    /* Methods to find references and conflicts in the flow graph */
    void addConfSyncEdges();
    void findRefs();
    void findConflicts();

    /* helper routines for clean-up */
    void pfg_cleanup(pfg_node *nd);

public:
    tree_proc *tproc()                  { return tp; }

    unsigned num_nodes();
    pfg_node *node(unsigned i)          { return (*nds)[i]; }
    pfg_node *&operator[](unsigned i)   { return (*nds)[i]; }

    pfg_node *entry_node()              { return en; }
    pfg_node *exit_node()               { return ex; }
    void set_entry_node(pfg_node *n);
    void set_exit_node(pfg_node *n);

    /* node insert/create routine -- must manually set preds/succs */
    pfg_block *clone_into(pfg_block *, boolean parallel=FALSE, pfg_node
	    *par_parent=0, threadBody *tb=0); 
    pfg_block *new_empty_block(boolean parallel=FALSE, pfg_node
	    *par_parent=0, threadBody *tb=0);

    /* built-in analysis routines */
    void find_dominators();
    void find_postdominators();
    void find_df();                     /* dominance frontier */
    void find_reverse_df();

    /* access dfa information */
    boolean dominates(int n_dominator, int n_dominatee);
    boolean dominates(pfg_node *dominator, pfg_node *dominatee);
    boolean postdominates(int n_dominator, int n_dominatee);
    boolean postdominates(pfg_node *dominator, pfg_node *dominatee);
    pfg_node *immed_dom(int n);
    pfg_node *immed_dom(pfg_node *n);
    pfg_node *immed_postdom(int n);
    pfg_node *immed_postdom(pfg_node *n);
    bit_set *dom_frontier(int n);
    bit_set *dom_frontier(pfg_node *n);
    bit_set *reverse_dom_frontier(int n);
    bit_set *reverse_dom_frontier(pfg_node *n);
    bit_set *dominators(int n);
    bit_set *dominators(pfg_node *n);
    bit_set *postdominators(int n);
    bit_set *postdominators(pfg_node *n);

    void find_natural_loops(); 
    int loop_depth(int n);
    int loop_depth(pfg_node *n);
    void set_loop_depth(pfg_node *n, int d); 
    void set_loop_depth(int n, int d); 
    
    /* Accessibility methods */
    boolean isPath(pfg_node *from, pfg_node *to);

    /* Returns the header node of the outermost loop enclosing 'node'. */
    pfg_node *outermostLoopHeader(pfg_node *node);

    boolean is_loop_begin(int n);       /* TRUE if block is loop entry */
    boolean is_loop_begin(pfg_node *cn);
    boolean is_loop_end(int n);         /* TRUE if block jumps to loop entry */
    boolean is_loop_end(pfg_node *cn);
    boolean is_loop_exit(int n);        /* TRUE if block is a loop exit */ 
    boolean is_loop_exit(pfg_node *cn);

    /* list generator routines */
    pfg_node_list *reverse_postorder_list(boolean forward=TRUE);

    /* Parallel information routines */
    boolean conc(pfg_node *a, pfg_node *b);	// TRUE if a and b can
    boolean conc(unsigned a, unsigned b)	// execute concurrently
   	{ return conc(node(a), node(b)); }
    bit_set *par_ancestors(pfg_node *n);
    bit_set *par_ancestors(unsigned n)
	{ return par_ancestors(node(n)); }

    /* Data references and conflicts */
    RefConflict_list *conflicts()	{ return _conflicts; }
    VarRef_list *refs()			{ return _refs; }

    /* clean-up routines */
    void remove_unreachable_blocks(boolean delete_instrs=FALSE); 
    boolean optimize_jumps(); 
    void remove_shadows(); 
    void remove_layout_info(); 

    void print(boolean show_layout = FALSE); 
    pfg(tree_block *b,
        boolean build_blocks = TRUE,    /* if FALSE, 1 instr per block */
        boolean break_at_call = FALSE, 
        boolean keep_layout = FALSE);
    ~pfg();
};

#endif /* PFG_GRAPH_H */
