//---------------------------------------------------------------------------
// $Id: node-dfa.cc,v 1.13 1998/06/03 16:52:38 diego Exp diego $
//
// Data flow methods for basic blocks.
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>

#define _MODULE_ "libpfg.a"

// #pragma implementation "node.h"

#include <string.h>

#include <suif.h>
#include <par.h>
#include <odc-util.h>

#include "pfg.h"

//---------------------------------------------------------------------------
// pfg_instr::findRefs()
// Looks for all variable references in this instruction node.
//---------------------------------------------------------------------------
VarRef_list *
pfg_instr::findRefs()
{
    delete _refs;
    _refs = this->pfg_node::findRefs(this->ti->instr());

    return _refs;
}


//---------------------------------------------------------------------------
// pfg_block::findRefs()
// Looks for all variable references in this basic block node.
//---------------------------------------------------------------------------
VarRef_list *
pfg_block::findRefs()
{
    static char *__FNAME = "pfg_block::findRefs";

    pfg_node_instr_iter iter(this);

    delete _refs;
    _refs = new VarRef_list;
    assert(_refs);

    while (iter.is_empty() == FALSE) {
	tree_instr *instr = iter.step();
	if (debug(__FNAME, 2)) {
	    printf("%s(): Looking for references in\n", __FNAME);
	    instr->print();
	    printf("\n");
	}
	_refs->append(this->pfg_node::findRefs(instr->instr()));
    }

    if (debug(__FNAME, 1)) {
	printf("%s(): References found in node %d\n", __FNAME,
		this->number());
	VarRef_list_iter ref_iter(_refs);
	while (ref_iter.is_empty() == FALSE) {
	    printf("\t");
	    ref_iter.step()->print();
	    printf("\n");
	}
	printf("\n");
    }

    return _refs;
}


//---------------------------------------------------------------------------
// pfg_node::addPhi()
// Adds a phi-term for the given variable to the list of references in this
// node.
//---------------------------------------------------------------------------
PhiTerm * 
pfg_node::addPhi(var_sym *var)
{
    PhiTerm *newref = new PhiTerm(var, this);
    assert(newref);
    if (_refs == NULL) {
	_refs = new VarRef_list;
	assert(_refs);
    }
    _refs->append(newref);

    return newref;
}


//---------------------------------------------------------------------------
// pfg_node::addPi()
// Adds a pi-term for the given variable to the list of references in this
// node.
//---------------------------------------------------------------------------
PhiTerm * 
pfg_node::addPi(var_sym *var)
{
    PhiTerm *newref = new PhiTerm(var, this, PI);
    assert(newref);
    if (_refs == NULL) {
	_refs = new VarRef_list;
	assert(_refs);
    }
    _refs->append(newref);

    return newref;
}


//---------------------------------------------------------------------------
// pfg_node::hasPi()
// Returns the PI term for the given variable. If the node does not have a
// PI term for the variable, it returns NULL.
//---------------------------------------------------------------------------
PhiTerm * 
pfg_node::hasPi(var_sym *var)
{
    VarRef_list_iter iter(_refs);
    
    while (iter.is_empty() == FALSE) {
	VarRef *ref = iter.step();
	if (ref->isPI() && ref->var()->sym_id() == var->sym_id()) {
	    return (PhiTerm *)ref;
	}
    }

    return NULL;
}


//---------------------------------------------------------------------------
// pfg_block::split()
// Splits a PFG basic block at the conflicting statements and
// synchronization instructions.
//---------------------------------------------------------------------------
void
pfg_block::split()
{
    static char *__FNAME = "pfg_block::split";

    if (_refs->count() == 0 || _splitup == TRUE) {
	return;
    }

    Partition *splitArray;
    splitArray = new Partition[_refs->count()];
    assert(splitArray);
    memset(splitArray, 0, _refs->count() * sizeof(Partition));

    unsigned numPart = this->initialSplit(splitArray);

    this->fixLowerBounds(splitArray, numPart);

    this->fixUpperBounds(splitArray, numPart);

    this->fixGaps(&splitArray, &numPart);

    splitArray[0].ti1           = this->ti1;
    splitArray[numPart - 1].ti2 = this->ti2;

    if (debug(__FNAME, 1)) {
	printf("%s(): Final partitions for basic block %d\n", __FNAME, 
		this->number());
	unsigned i;
	for (i = 0; i < numPart; i++) {
	    printf("Partition %d: ti1 = %d\tti2 = %d\n", i,
		(splitArray[i].ti1) ? splitArray[i].ti1->instr()->number() : 0,
		(splitArray[i].ti2) ? splitArray[i].ti2->instr()->number() : 0
		);
	}
	printf("\n");
    }

    /* Once partition boundaries have been computed, we create new basic
     * blocks. We don't have to create 'numPart' basic blocks. We update
     * this block's ti2 with the first partition's ti2.
     */
    this->ti2 = splitArray[0].ti2;

    pfg_block *newblk, *prevblk;
    unsigned i;
    prevblk = this;
    for (i = 1; i < numPart; i++) {
	newblk = new pfg_block(prevblk->tnl, splitArray[i].ti1,
		splitArray[i].ti2, prevblk->_parallel, prevblk->_par_parent,
		prevblk->_thread);
	newblk->insert_node_after(prevblk);
	newblk->splitup(TRUE);
	prevblk = newblk;
    }
     
    /* Finally, move the shadow and cti instructions from this block
     * to the last block that we created.
     */
    prevblk->set_in_cti(this->in_cti());
    this->set_in_cti(NULL);

    prevblk->set_in_shad(this->in_shad());
    this->set_in_shad(NULL);

    _splitup = TRUE;

    delete splitArray;
}


//---------------------------------------------------------------------------
// pfg_block::initialSplit()
// Helper function for pfg_block::split(). It traverses the conflicting
// references in this basic block to set up split points based on the kinds
// of references it finds. It only sets the limits based on the reference,
// it doesn't try to adjust for preceding labels or unknown bounds.
//---------------------------------------------------------------------------
unsigned
pfg_block::initialSplit(Partition *splitArray)
{
    static char *__FNAME = "pfg_block::initialSplit";

    VarRef_list_iter iter(_refs);
    unsigned numPart;

    numPart = 0;
    while (iter.is_empty() == FALSE) {
	VarRef *ref = iter.step();

	if (debug(__FNAME, 2)) {
	    printf("%s(): Checking reference ", __FNAME);
	    ref->print();
	    printf("\n");
	}

	if (ref->conflicts()->is_empty() && !ref->isSync()) {
	    if (debug(__FNAME, 2)) {
		printf("%s(): No conflicts for this reference\n", __FNAME);
	    }
	    continue;
	}

	/* Check if the statement for this reference contains another
	 * conflicting reference of a different kind. If so, the statement
	 * must go in its own basic block because it has to be both the
	 * first and the last statement of the block.
	 */
	boolean isDefUse = FALSE;
	if ((ref->isD() || ref->isU()) && iter.is_empty() == FALSE) {
	    VarRef *nextRef = iter.peek();
	    while (nextRef && nextRef->stmt() == ref->stmt()) {
		iter.step();	/* Consume the reference */
		if (nextRef->conflicts()->is_empty() == FALSE && 
		    nextRef->kind() != ref->kind()) {
		    isDefUse = TRUE;
		    if (debug(__FNAME, 2)) {
			printf("%s(): The reference is DefUse\n", __FNAME);
		    }
		}
		if (iter.is_empty()) {
		    nextRef = NULL;
		} else {
		    nextRef = iter.peek();
		}
	    }
	}

	/* Must check whether it's def-use first! */
	if (isDefUse || ref->isUSync()) {
	    splitArray[numPart].ti1 = ref->stmt();
	    splitArray[numPart].ti2 = ref->stmt();
	} else if (ref->isD() || ref->isHSync()) {
	    splitArray[numPart].ti2 = ref->stmt();
	} else if (ref->isU() || ref->isTSync()) {
	    splitArray[numPart].ti1 = ref->stmt();
	} else {
	    error_line(1, ref->stmt(), "Unrecognized reference kind %d\n",
		    ref->kind());
	}

	numPart++;
    }

    if (debug(__FNAME, 1)) {
	printf("%s(): Initial partitions for basic block %d\n", __FNAME, 
		this->number());
	unsigned i;
	for (i = 0; i < numPart; i++) {
	    printf("Partition %d: ti1 = %d\tti2 = %d\n", i,
		(splitArray[i].ti1) ? splitArray[i].ti1->instr()->number() : 0,
		(splitArray[i].ti2) ? splitArray[i].ti2->instr()->number() : 0
		);
	}
	printf("\n");
    }

    return numPart;
}

//---------------------------------------------------------------------------
// pfg_block::fixLowerBounds()
// Patch all lower bounds in the split array to include any immediately
// preceding 'lab' or 'mrk' instructions.
//---------------------------------------------------------------------------
void
pfg_block::fixLowerBounds(Partition *splitArray, unsigned numPart)
{
    unsigned i;

    for (i = 1; i < numPart; i++) {
	tree_instr *new_ti1 = (splitArray[i].ti1) 
	    			? splitArray[i].ti1 
				: splitArray[i].ti2;
	boolean foundAll = FALSE;
	while (!foundAll && new_ti1) {
	    new_ti1 = (tree_instr *)new_ti1->list_e()->prev()->contents;
	    if_ops op = new_ti1->instr()->opcode();
	    if (op != io_mrk && op != io_lab) {
		foundAll = TRUE;
		new_ti1 = (tree_instr *)new_ti1->list_e()->next()->contents;
	    }
	}
	splitArray[i].ti1 = new_ti1;
    }
}


//---------------------------------------------------------------------------
// pfg_block::fixUpperBounds()
// Once all the lower bounds have been reset, we fix the upper bounds to be
// the statement immediately preceding the next partition's lower bound.
//---------------------------------------------------------------------------
void
pfg_block::fixUpperBounds(Partition *splitArray, unsigned numPart)
{
    unsigned i;

    for (i = 0; i < numPart - 1; i++) {
	if (splitArray[i].ti2 == NULL) {
	    splitArray[i].ti2 = (tree_instr *)splitArray[i + 1].ti1->list_e()->prev()->contents;
	}
    }
}

//---------------------------------------------------------------------------
// pfg_block::fixGaps()
// Traverse the split array looking for gaps caused by the splitting
// process. We are looking for cases in which splitArray[i].ti1 is not the
// instruction following splitArray[i-1].ti2. If this is the case, we
// insert a new partition between i-1 and i covering the gap.
//---------------------------------------------------------------------------
void
pfg_block::fixGaps(Partition **splitArray, unsigned *numPart)
{
    static char *__FNAME = "pfg_block::fixGaps";

    unsigned i;

    for (i = 1; i < *numPart; i++) {
	tree_instr *first = (*splitArray)[i].ti1;
	tree_instr *next = (tree_instr *)(*splitArray)[i - 1].ti2->list_e()->next()->contents;

	if (first != next) {
	    /* Create a new partition. The first instruction is the
	     * instruction following splitArray[i-1].ti2 and the last
	     * instruction is the instruction preceding splitArray[i].ti1.
	     */
	    tree_instr *new_ti1 = next;
	    tree_instr *new_ti2 = (tree_instr *)first->list_e()->prev()->contents;

	    if (debug(__FNAME, 2)) {
		printf("%s(): There is a gap between partitions %d and %d\n", 
			__FNAME, i - 1, i);
		printf("\tFirst instruction of the gap: %d\n",
			new_ti1->instr()->number());
		printf("\tLast instruction of the gap: %d\n",
			new_ti2->instr()->number());
	    }

	    /* Insert the new partition before position 'i' of the
	     * splitArray.
	     */
	    (*numPart)++;

	    Partition *newSplit = new Partition[*numPart];
	    assert(newSplit);

	    unsigned j;
	    for (j = 0; j < i; j++) {
		newSplit[j].ti1 = (*splitArray)[j].ti1;
		newSplit[j].ti2 = (*splitArray)[j].ti2;
	    }

	    newSplit[i].ti1 = new_ti1;
	    newSplit[i].ti2 = new_ti2;

	    for (j = i + 1; j < *numPart; j++) {
		newSplit[j].ti1 = (*splitArray)[j - 1].ti1;
		newSplit[j].ti2 = (*splitArray)[j - 1].ti2;
	    }

	    delete[] *splitArray;
	    *splitArray = newSplit;
	}
    }
}



//***************************************************************************
//			     Protected methods
//***************************************************************************
//---------------------------------------------------------------------------
// pfg_node::findRefs()
// Looks for a variable reference in the given instruction. A references
// may be a definition or a use. Definitions may be found in the
// destination operand or in pointers to variables. Uses are found in
// source operands to the instruction.
//---------------------------------------------------------------------------
VarRef_list *
pfg_node::findRefs(instruction *in)
{
    static char *__FNAME = "pfg_node::findRefs";
    VarRef *newref;
    VarRef_list *refs = new VarRef_list;
    assert(refs);

    if (debug(__FNAME, 2)) {
	printf("%s(): Checking instruction\n", __FNAME);
	in->print();
    }

    /* First check if the instruction is one of the synchronization
     * functions. If it is, add the synch reference to the list of
     * references and return.
     */
    if (Is_head_sync(in) || Is_tail_sync(in) || Is_mutex_sync(in)) {
	assert( ((in_cal *)in)->argument(0).is_symbol() );
	var_sym *var = ((in_cal *)in)->argument(0).symbol();
	if (Is_head_sync(in)) {
	    HeadSyncRef *ref = new HeadSyncRef(var, in->parent(), this);
	    refs->append(ref);
	} else if (Is_tail_sync(in)){
	    TailSyncRef *ref = new TailSyncRef(var, in->parent(), this);
	    refs->append(ref);
	} else if (Is_lock_sync(in)) {
	    LockRef *ref = new LockRef(var, in->parent(), this);
	    refs->append(ref);
	} else if (Is_unlock_sync(in)) {
	    UnlockRef *ref = new UnlockRef(var, in->parent(), this);
	    refs->append(ref);
	} else {
	    error_line(1, in, ("Unknown synchronization instruction\n"));
	}

	return refs;
    }
	
    /* Special case. Check store, load and memcpy instructions for pointer,
     * structures and array references.
     */
    if_ops op = in->opcode();
    if (op == io_lod || op == io_str || op == io_memcpy) {
	refs->append(this->findPtrRefs(in));
    }
	

    /* Now check the destination and source operands for the instruction.
     * Note that we must check all the source operands first because if the
     * instruction defines and uses the same variable, the use should
     * appear first so that the FUD chain algorithm doesn't get confused by
     * linking the use to the def in this instruction.
     *
     * Check all the source operands to the instruction. If the operand is
     * a symbol, add it to the list of uses. If it is an instruction
     * recurse into it.
     */
    unsigned i;
    for (i = 0; i < in->num_srcs(); i++) {
	if (in->src_op(i).is_symbol()) {
	    if (debug(__FNAME, 1)) {
		printf("%s(): Line %d. Found use for variable: ",
			__FNAME, source_line_num(in));
		in->src_op(i).symbol()->print();
		printf("\n");
	    }
	    newref = new VarUse(in->src_op(i).symbol(), in->parent(), this);
	    refs->append(newref);
	} else if (in->src_op(i).is_instr()) {
	    refs->append(this->findRefs(in->src_op(i).instr()));
	}
    }


    /* Check the destination operand. If the destination operand is a
     * symbol add it to the list of definitions.
     */
    if (in->dst_op().is_symbol()) {
	if (debug(__FNAME, 1)) {
	    printf("%s(): Line %d. Found definition for variable: ",
		    __FNAME, source_line_num(in));
	    in->dst_op().symbol()->print();
	    printf("\n");
	}
	newref = new VarDef(in->dst_op().symbol(), in->parent(), this);
	refs->append(newref);
    }

    return refs;
}



//---------------------------------------------------------------------------
// pfg_node::findPtrRefs()
// Finds symbol references in memory copy instructions (str, lod and
// memcpy). These instructions move data from memory addresses which can be
// pointers, arrays and fields within data structures.
//---------------------------------------------------------------------------
VarRef_list *
pfg_node::findPtrRefs(instruction *in)
{
    static char *__FNAME = "pfg_node::findPtrRefs";

    if_ops op = in->opcode();
    assert(op == io_lod || op == io_str || op == io_memcpy);

    VarRef *newref;
    var_sym *sym;
    VarRef_list *refs = new VarRef_list;
    assert(refs);


    if (debug(__FNAME, 1)) {
	printf("%s(): Checking instruction for pointer references\n", __FNAME);
	in->print();
    }

    /* Look for uses in the src_addr_op() of 'lod' and 'memcpy'
     * instructions. Note again that we must check uses before the defs.
     */
    if (op == io_lod || op == io_memcpy) {
	sym = this->symRefBy(((in_rrr *)in)->src_addr_op());
	assert(sym);

	/* Create the new use and add it to the list */
	if (debug(__FNAME, 2)) {
	    printf("%s(): Found source address: ", __FNAME);
	    sym->print();
	    printf("\n");
	}
	newref = new VarUse(sym, in->parent(), this);
	refs->append(newref);
    }

    /* Look for definitions in the dst_addr_op() of 'str' and 'memcpy'
     * instructions.
     */
    if (op == io_str || op == io_memcpy) {
	sym = this->symRefBy(((in_rrr *)in)->dst_addr_op());
	assert(sym);

	/* Create the new definition and add it to the list */
	if (debug(__FNAME, 2)) {
	    printf("%s(): Found destination address: ", __FNAME);
	    sym->print();
	    printf("\n");
	}
	newref = new VarDef(sym, in->parent(), this);
	refs->append(newref);
    }	


    return refs;
}


//---------------------------------------------------------------------------
// pfg_node::symRefBy()
// Helper method for findPtrRefs(), it returns the variable symbol
// referenced by the given address operand. It assumes that the operand is
// an array instruction, an ldc instruction or a symbol.
//---------------------------------------------------------------------------
var_sym *
pfg_node::symRefBy(operand addr)
{
    static char *__FNAME = "pfg_node::symRefBy";

    if (debug(__FNAME, 2)) {
	printf("%s(): Looking for a symbol in operand:\n", __FNAME);
	addr.print();
	printf("\n");
    }

    /* Ignore null operands */
    if (addr.is_null()) {
	return NULL;
    }

    /* If the operand is already a symbol, return it */
    if (addr.is_symbol()) {
	return addr.symbol();
    }


    /* The destination address must be an instruction */
    assert(addr.is_instr());

    instruction *addr_i = addr.instr();
    in_ldc  *ptr;
    var_sym *sym;

    if (addr_i->opcode() == io_array) {
	/* The base operand of the array is an 'ldc' instruction that
	 * references to the base address of the array.
	 */
	ptr = (in_ldc *)(((in_array *)addr_i)->base_op().instr());
	assert(ptr->value().is_symbol());
	sym = (var_sym *)ptr->value().symbol();
    } else if (addr_i->opcode() == io_ldc) {
	/* 'ldc' instructions, check if it's a symbol */
	ptr = (in_ldc *)addr_i;
	if (ptr->value().is_symbol()) {
	    sym = (var_sym *)ptr->value().symbol();
	} else {
	    sym = NULL;
	}
    } else {
	/* If it's neither an array or an ldc, check the source operands,
	 * one of them should have a symbol. If there is more than one
	 * symbol in the source operands, we are in trouble.
	 */
	unsigned i;
	var_sym *tmpsym = NULL;
	sym = NULL;
	for (i = 0; i < addr_i->num_srcs(); i++) {
	    tmpsym = this->symRefBy(addr_i->src_op(i));
	    if (tmpsym) {
		if (sym == NULL) {
		    sym = tmpsym;
		} else {
		    /* Sanity check, we should only find one symbol here */
		    printf("%s(): Mierda! Encontre dos simbolos en\n", __FNAME);
		    addr_i->print();
		    printf("\n");
		    exit(1);
		}
	    }
	}
    }


    return sym;
}


//---------------------------------------------------------------------------
// pfg_node::findAliases()
// Looks for aliases in the given instruction. The instruction type is
// assumed to be a pointer. The analysis is very simple for now, we only
// check for ampersand operators.
//---------------------------------------------------------------------------
var_sym_list *
pfg_node::findAliases(instruction *in)
{
    assert(in->result_type()->is_ptr());

    var_sym_list *list = new var_sym_list;
    assert(list);

    if (in->opcode() == io_ldc) {
	in_ldc *load = (in_ldc *)in;
	immed value = load->value();

	if (value.is_symbol()) {
	    sym_node *sym = value.symbol();
	    if (sym->is_var()) {
		list->append((var_sym *)sym);
	    }
	}
    }

    return list;
}
