/*  Control Flow Graph Node Implementation */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    Copyright (c) 1995,1996 The President and Fellows of Harvard University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file.


    Modified by Diego Novillo (diego@cs.ualberta.ca)
*/
#include <suif_copyright.h>

#define _MODULE_ "libpfg.a"

// #pragma implementation "node.h"

#include <stdlib.h>

#include <suif.h>
#include <par.h>

#include "pfg.h"


/* local function prototypes */
static void remove_node(pfg_node_list *l, pfg_node *n);
static void print_node_list(FILE *fp, pfg_node_list *l);


/* 
 * General constructor for pfg_node's.  The number and
 * parent should be overwritten when the pfg_node is 
 * attached to a pfg.  
 */
pfg_node::pfg_node (boolean parallel, pfg_node *par_parent, threadBody *t)
{
    _parallel = parallel;
    _par_parent = par_parent;
    _thread = t;
    _refs = NULL;
    _conflicts = NULL;
    _splitup = FALSE;
    set_number((unsigned)-1);
    set_parent(NULL);
    lpred = lsucc = NULL;
}

boolean
pfg_node::is_parbegin() const
{ 
    return (this->is_begin() && ((pfg_begin *)this)->is_parallel());
}


boolean
pfg_node::is_parend() const
{ 
    return (this->is_end() && ((pfg_end *)this)->is_parallel());
}


/* 
 * pfg_node::clear_layout_succ() -- Just manipulate the
 * layout pointers.  The actual work of re-inserting
 * shadow instructions is done in the subclass
 * methods.  
 */
void
pfg_node::clear_layout_succ ()
{ 
    if (lsucc) {
	assert(lsucc->lpred == this); 
	lsucc->lpred = NULL; 
	lsucc = NULL; 
    }
}


/* pfg_node::set_layout_succ() -- Just manipulate the
 * layout pointers.  The actual work of inverting polarity
 * and inserting shadow instructions is done in the
 * subclass methods.  
 * 
 * Always return FALSE because this routine never inverts
 * branch polarity. 
 */
boolean 
pfg_node::set_layout_succ (pfg_node *nod) 
{ 
    if (nod == NULL) {
	clear_layout_succ();
	return FALSE; 
    }
    assert(nod->lpred == NULL); 
    lsucc = nod; 
    lsucc->lpred = this; 
    return FALSE; 
}


/* 
 * pfg_node::set_layout_succ(label_sym) -- wrapper routine
 * that translates a label successor into the corresponding
 * pfg_block, then calls the other set_layout_succ() routine.
 */
boolean 
pfg_node::set_layout_succ(label_sym *lab)
{
    pfg_node *new_succ = label_pfg_node(lab); 
    assert_msg(new_succ, ("set_layout_succ() -- "
	"cannot find pfg_node for specified label")); 
    return set_layout_succ(new_succ); 
}


/*
 * Landing pads are ....  They are useful for ....
 */

void 
pfg_node::insert_empty_block(unsigned n)
{
    assert((int)n < succs()->count()); 
    pfg_node *dst = (*succs())[n]; 
    assert(dst->is_block()); 
    pfg_block *dest = (pfg_block *)dst; 

    proc_symtab *par_symtab = parent()->tproc()->proc_syms(); 
    label_sym *new_lab = par_symtab->new_unique_label(); 
    label_sym *dst_lab = dest->get_label(); 

    // Create and attach the block 
    tree_node_list *lp_tnl = new tree_node_list; 
    tree_instr *lp_lab, *lp_jmp; 
    lp_lab = new tree_instr(new in_lab(new_lab)); 
    lp_jmp = new tree_instr(new in_bj(io_jmp, dst_lab)); 

    lp_tnl->append(lp_lab); 
    lp_tnl->append(lp_jmp); 
    pfg_block *lp_blk = new pfg_block(lp_tnl, lp_lab, lp_jmp); 
    lp_lab->annotes()->append(new annote(k_pfg_node, (void *)lp_blk)); 
    lp_jmp->annotes()->append(new annote(k_pfg_node, (void *)lp_blk)); 
    new_lab->annotes()->append(new annote(k_pfg_node, (void *)lp_blk));
    lp_blk->set_in_cti(lp_jmp); 
    parent()->attach(lp_blk); 

    // Link to the block
    assert_msg(is_block(), 
	("insert_empty_block: unsupported source node type"));
    set_succ(n, lp_blk); 
    lp_blk->set_succ(0, dest); 
}

void 
pfg_node::insert_empty_block(pfg_node *dst)
{
    pfg_node_list_iter cnli(succs()); 
    int i; 
    for (i = 0; !cnli.is_empty(); ++i) {
	pfg_node *scan = cnli.step(); 
	if (scan == dst)
	    break; 
    }
    while (!cnli.is_empty()) {
	pfg_node *scan = cnli.step(); 
	assert_msg((scan != dst), 
	    ("insert_empty_block: ambiguous destination node")); 
    }
    insert_empty_block(i); 
}


/*
 * ---------- Implementation of PFG_BEGIN methods ----------
 */

/* 
 * pfg_begin::set_succ() -- This function actually sets
 * the entry point into the procedures.  It does this by
 * moving the k_proc_entry annotation from the old entry
 * point to the new entry point.  Code generators are
 * expected to place this annotation on the first instruction
 * of the entry block and printmachine is expected to handle 
 * the difference * between the _beginning_ of the procedure 
 * and the _entry point_ of the procedure appropriately.  
 *
 * Find and remove any old k_proc_entry annotation.  Then 
 * stick the new annotation on the new entry point.  
 */
void
pfg_begin::set_succ (unsigned n, pfg_node *nod)
{
    annote *an = NULL; 
    assert(n == 0); 
    assert(nod->is_block());
    pfg_block *new_succ = (pfg_block *)nod; 
    pfg_block *old_succ = NULL; 

    if (succs()->count() > 1) {
	old_succ = (pfg_block *)(*succs())[0]; 
	pfg_node_instr_iter cnii(old_succ);  
	while (!an && !cnii.is_empty()) {
	    tree_instr *ti = cnii.step(); 
	    instruction *i = ti->instr(); 
	    if (i->are_annotations()) 
		an = i->annotes()->get_annote(k_proc_entry); 
	}
	assert_msg(an, ("pfg_begin::set_succ() -- "
	    "could not find k_proc_entry annotation on old entry point")); 
    }
    if (an == NULL)
	an = new annote(k_proc_entry); 
    new_succ->in_head()->instr()->annotes()->append(an); 

    if (old_succ)
	succs()->pop(); 
    succs()->push(nod); 

    // Clean up pred pointers
    if (old_succ && !occurs(old_succ, succs()))
	remove_node(old_succ->preds(), this); 
    if (!occurs(this, new_succ->preds()))
	new_succ->preds()->append(this);
}


/* 
 * pfg_begin::set_layout_succ() -- Move a block up to the front
 * of the SUIF tree_node_list.  
 *     We really care about the _clump_ that the target block
 * represents.  A clump is a set of blocks that have been joined together
 * by layout pointers (and are thus contiguous in the tree_node_list).  
 * The driver program has requested that nod's clump appear at the
 * beginning of the SUIF procedure.  
 *     Always returns FALSE because the pfg_begin node has no
 * control instruction to invert. 
 */
boolean
pfg_begin::set_layout_succ (pfg_node *bot)
{
    assert_msg(lpred == NULL, ("set_layout_succ() -- "
	"begin block has a layout predecessor")); 

    if (bot == NULL) {
	clear_layout_succ();
	return FALSE; 
    }

    assert(bot->is_block()); 
    pfg_block *bot_start = (pfg_block *)bot; 

    // Verify layout pointers OK, same tnl in blocks
    assert_msg(this->lsucc == NULL, ("set_layout_succ() -- "
	"begin block already has a layout successor")); 
    assert_msg(bot_start->layout_pred() == NULL, ("set_layout_succ() -- "
	"bottom block already has a layout predecessor")); 
    assert_msg(this->parent() == bot_start->parent(), 
	("set_layout_succ: requested blocks do not belong to the same pfg")); 

    // Check for layout cycles; find end of bottom clump
    pfg_block *bot_end = bot_start; 
    while (bot_end->layout_succ()) {
	bot_end = (pfg_block *)bot_end->layout_succ(); 
	assert(bot_end->is_block()); 
    }

    // Move instructions from bottom clump 
    tree_node_list_e *ins_point = NULL; 
    tree_node_list_e *del_point = bot_start->in_head()->list_e(); 
    tree_node_list_e *end_point = bot_end->in_shad() 
	? bot_end->in_shad()->list_e()->next()
	: bot_end->in_tail()->list_e()->next(); 
    tree_node_list *tnl = parent()->tproc()->body(); 
    do {
	tree_node_list_e *next = del_point->next(); 
	tree_node_list_e *temp = tnl->remove(del_point); 
	if (ins_point == NULL) 
	    tnl->push(temp); 
	else
	    tnl->insert_after(temp, ins_point); 
	del_point = next; 
	ins_point = temp; 
    } while (del_point != end_point); 

    // Link the lsucc and lpred fields. 
    pfg_node::set_layout_succ(bot); 

    return FALSE; 
}


/*
 * ---------- Implementation of PFG_BLOCK methods ----------
 */

/* This constructor creates an unattached copy of pfg_block orig.  This
 * routine needs to be fixed to work with low-suif mbr instructions. */
pfg_block::pfg_block(pfg_block *orig, base_symtab *dst_scope, boolean
	parallel, pfg_node *par_parent, threadBody *t)
    : pfg_node(parallel, par_parent, t)
{
    int num_tgts = 0; 

    // Clone the tree node list
    // Need to take special care with mbr jump tables
    replacements r; 
    if (orig->ends_in_mbr()) {
	instruction *mbri = orig->in_cti()->instr(); 
	immed_list *il = (immed_list *)mbri->peek_annote("instr_mbr_tgts");
	num_tgts = (*il)[0].integer(); 
	assert((*il)[1].symbol()->is_var()); 
	var_sym *old_dtable = (var_sym *)(*il)[1].symbol(); 

	base_symtab *targ_scope = dst_scope; 
	if (targ_scope == NULL)
	    targ_scope = orig->parent()->tproc()->proc_syms(); 

	// Generate new type for dispatch table
	int entry_bits = target.size[C_ptr]; 
	type_node *dtable_entry_type = type_ptr; 
	type_node *dtable_type = new array_type(dtable_entry_type, 
	    array_bound(0), array_bound(num_tgts - 1));
	dtable_type = targ_scope->install_type(dtable_type);

	// Install var_sym into current procedure symbol table 
	var_sym *new_dtable = targ_scope->new_unique_var(dtable_type, 
	    old_dtable->name());
	new_dtable->reset_userdef();        /* not user defined */

	// Define data for table 
	var_def *dtable_def = targ_scope->define_var(new_dtable, 32);
	for (int i = 0; i < num_tgts; i++) {
	    il = new immed_list();
	    il->append(immed(1));                   // no. of copies 
	    il->append(immed(entry_bits));          // total bits in value 
	    il->append(immed((label_sym *)NULL));   // data value 
	    dtable_def->append_annote(k_repeat_init, il);
	}

	r.oldsyms.push(old_dtable); 
	r.newsyms.push(new_dtable); 
    }

    pfg_node_instr_iter cnii(orig); 
    while (!cnii.is_empty()) {
	tree_instr *ti = cnii.step(); 
	ti->find_exposed_refs(dst_scope, &r); 
    }

    r.resolve_exposed_refs(dst_scope); 

    tiC = NULL; 
    tnl = new tree_node_list; 
    cnii.reset(); 
    while (!cnii.is_empty()) {
	tree_instr *tiSrc = cnii.step(); 
	tree_instr *tiDst = (tree_instr *)tiSrc->clone_helper(&r); 
	delete tiDst->get_annote(k_pfg_node); 
	annote *an2 = new annote(k_pfg_node, (void *)this); 
	tiDst->annotes()->append(an2); 
	tnl->append(tiDst); 
	if (orig->tiC == tiSrc)
	    this->tiC =  tiDst; 
    }

    // Set ti1 and ti2
    ti1 = (tree_instr *)tnl->head()->contents; 
    ti2 = (tree_instr *)tnl->tail()->contents; 
    tiS = NULL; 

    // number set to -1, parent set to NULL by pfg_node::pfg_node()
    // prs, scs constructed by pfg_node::pfg_node()

    // If an mbr, set up edges to NULL. 
    if (ends_in_mbr()) {
	for (int i = 0; i < num_tgts; ++i)
	    succs()->append((pfg_node *)NULL); 
    }
}


/* This constructor creates an unattached empty block. */
pfg_block::pfg_block(block_symtab *dst_scope, boolean parallel, pfg_node
	*par_parent, threadBody *t)
    : pfg_node(parallel, par_parent, t)
{
    assert(dst_scope != NULL); 

    tnl = new tree_node_list; 
    label_sym *ls = dst_scope->new_unique_label(); 

    instruction *in; 
    in = new in_lab(ls); 
    tnl->append(new tree_instr(in)); 

    ti1 = (tree_instr *)tnl->head()->contents; 
    ti2 = (tree_instr *)tnl->tail()->contents; 
    tiC = NULL; 
    tiS = NULL; 
}


/* pfg_block::set_succ() -- set the pfg_node suc to be the nth
 * successor of this pfg_block. */
void
pfg_block::set_succ(unsigned n, pfg_node *suc)
{
    if (suc->is_end()) {
	/* assert(ends_in_return());  */
	if (!succs()->is_empty())
	    succs()->pop(); 
	assert(succs()->is_empty()); 
	succs()->push(suc); 
	if (!occurs(this, suc->preds()))
	    suc->preds()->append(this);
	return; 
    } 

    if (suc->is_block() == FALSE) {
	printf("\n");
	printf("pfg_block::set_succ(): Warning. Trying to set my successor "
		"to a non-block node\n");
	printf("pfg_block::set_succ(): source: ");
	this->print();

	printf("pfg_block::set_succ(): target: ");
	suc->print();
	printf("\n");
// 	assert(suc->is_block()); 
    }

    pfg_block *new_succ = (pfg_block *)suc; 
    pfg_node *old_succ = NULL; // until proven otherwise
    instruction *last = in_cti() ? in_cti()->instr() : in_tail()->instr(); 

    // 0. fall-through
    if (n == 0 && !ends_in_mbr()) {
	if (lsucc) {
	    assert_msg(suc == (*succs())[0], ("set_succ() -- "
		"node has layout successor "
		"but requested different fall-through successor")); 
	    return; 
	}

	if (!succs()->is_empty()) {
	    old_succ = succs()->head()->contents; 
	    succs()->pop(); 
	}
	succs()->push(new_succ); 

	set_shadow(suc); 

    // 1. cbr target
    } else if (n == 1 && !ends_in_mbr()) {
	assert(ends_in_cbr()); 

	switch (succs()->count()) {
	case 0:
	    succs()->append((pfg_node *)NULL); 
	    succs()->append(new_succ); 
	    break; 
	case 1:
	    succs()->append(new_succ); 
	    break; 
	default:
	    assert(succs()->count() >= 0); 
	    old_succ = succs()->head()->next()->contents; 
	    succs()->remove(succs()->head()->next()); 
	    succs()->insert_after(new_succ, succs()->head()); 
	    break; 
	}

	((in_bj *)last)->set_target(new_succ->get_label()); 

    // 2. mbr target
    } else {
	assert(ends_in_mbr()); 

// 	immed_list *iml = (immed_list *)
// 	    last->peek_annote("instr_mbr_tgts");
// 	assert_msg(iml, ("add_pfg_edge - no mbr targets"));
	int num = ((in_mbr *)last)->num_labs();
	assert((int)n <= num); 
// 	sym_node *dtable_sym = (*iml)[1].symbol(); 
// 	assert(dtable_sym->is_var()); 
// 	var_sym *dtable_var = (var_sym *)dtable_sym; 
// 	var_def *dtable_def = dtable_var->definition(); 

	unsigned j; 
	if (succs()->count() == 0) 
	    for (j = 0; j < (unsigned)num; ++j)
		succs()->append(new pfg_node_list_e(NULL)); 

	pfg_node_list_e *cnle = succs()->head(); 
// 	immed_list **tgts = new immed_list*[num]; 
	for (j = 1; j <= (unsigned)num; j++) {
// 	    tgts[j] = (immed_list *)dtable_def->get_annote(k_repeat_init); 
// 	    assert(tgts[j]); 
	    if (j == n) {
// 		immed_list *il = new immed_list; 
// 		il->append((*tgts[n])[0].integer()); 
// 		il->append((*tgts[n])[1].integer()); 
// 		il->append(new_succ->get_label()); 
// 		delete tgts[n]; 
// 		tgts[n] = il; 

		pfg_node_list_e *new_elem = new pfg_node_list_e(new_succ);
		if (cnle == NULL) {
		    succs()->append(new_elem); 
		} else {
		    old_succ = cnle->contents; 
		    succs()->insert_before(new_elem, cnle); 
		    succs()->remove(cnle); 
		}
		cnle = new_elem; 
	    }
	    cnle = cnle->next(); 
	}
// 	for (j = 0; j < (unsigned)num; j++)
// 	    dtable_def->append_annote(k_repeat_init, tgts[j]); 
// 	delete tgts; 
    }

    // Clean up pred pointers
    if (old_succ && !occurs(old_succ, succs()))
	remove_node(old_succ->preds(), this); 
    if (!occurs(this, new_succ->preds()))
	new_succ->preds()->append(this);

    return; 
}


/* pfg_block::clear_layout_succ() -- Disconnect two blocks in the layout
 * chain.  This may involve demoting unconditional branches into 
 * shadows or inserting a new shadow goto instruction.
 * Just call set_shadow() to do the real work. 
 */
void
pfg_block::clear_layout_succ (void)
{
    pfg_node::clear_layout_succ(); 
    if (succs()->count())
	set_shadow(succs()->head()->contents); 
}


/* 
 * pfg_block::set_layout_succ() -- Connect two blocks in the layout
 * chain, and update the underlying SUIF tree_node_list so that these
 * two blocks are now adjacent.  
 *     We really care about the _clump_s that the two blocks
 * represent.  A clump is a set of blocks that have been joined together
 * by layout pointers (and are thus contiguous in the tree_node_list).  
 * In the procedure below, we will talk about the top block (this) and
 * the bottom block (bot); the driver program has requested that 
 * bot follow this in layout order.  
 *     The top block should be the last block in the top clump; the 
 * bottom block should be the first block in the bottom clump.  If not, 
 * then we are trying to link into the middle of a clump, which is an
 * error.  
 *     The top and bottom clump should be different.  If they are the
 * same clump, then setting the layout link will create a layout cycle, 
 * which is also an error.  
 *     We optimize control instructions when the layout successor
 * is set.  If the top block is an unconditional or conditional (2-way)
 * branch/jump and the bottom block is a graph successor of the top 
 * block, then we can make the top block fall through to the bottom
 * block on some path.  Doing this may require removing shadow instructions
 * and/or inverting the polarity of the top block's conditional branch. 
 *     Returns TRUE if this's conditional branch was inverted because
 * of this layout request.  
 */
boolean 
pfg_block::set_layout_succ(pfg_node *bot)
{
    boolean inverted = FALSE; 

    if (bot == NULL) {
	clear_layout_succ();
	return FALSE; 
    }

    assert(bot->is_block()); 

    pfg_block *bot_start = (pfg_block *)bot; 

    // Verify layout pointers OK, same tnl in blocks
    assert_msg(this != bot, 
	("set_layout_succ: cannot place a block after itself")); 
    assert_msg(this->lsucc == NULL, 
	("set_layout_succ: top block already has a layout successor")); 
    assert_msg(bot_start->lpred == NULL, 
	("set_layout_succ: bottom block already has a layout predecessor")); 
    assert_msg(this->tnl == bot_start->tnl, 
	("set_layout_succ: requested blocks do not belong to the same pfg")); 

    // Check for layout cycles; find end of bottom clump
    pfg_block *bot_end = bot_start; 
    while (bot_end->lsucc) {
	assert_msg(bot_end != this, ("set_layout_succ: layout cycle")); 
	bot_end = (pfg_block *)bot_end->lsucc; 
	assert(bot_end->is_block()); 
    }
    assert_msg(bot_end != this, ("set_layout_succ: layout cycle")); 

    // Check layout successor is a cbr control successor
    instruction *top_lasti = in_tail()->instr(); 
    assert_msg(!Is_cbr(top_lasti) 
	|| ((*succs())[0] == bot)
	|| ((*succs())[1] == bot), 
	("set_layout_succ() -- layout successor of a cbr block "
	"must be either the fall-through or taken successor")); 

    // Check layout successor is a call control successor 
    assert_msg(!Is_call(top_lasti) 
	|| ((*succs())[0] == bot), 
	("set_layout_succ() -- layout successor of a call block "
	"must be the fall-through successor")); 

    // Fix up end of top clump (invert cbrs, remove shadows)
    if (Is_cbr(top_lasti) && (*succs())[1] == bot_start) {
	invert_branch(); 
	inverted = TRUE; 

	// swap succs(0) and succs(1)
	pfg_block *new_targ = (pfg_block *)succs()->pop(); 
	pfg_block *new_fall = (pfg_block *)succs()->pop(); 
	assert(new_targ->is_block()); 
	assert(new_fall->is_block()); 
	succs()->push(new_targ); 
	succs()->push(new_fall); 

	// set the new branch target
	((in_bj *)top_lasti)->set_target(new_targ->get_label()); 
    }
    if (succs()->count()) {
	if ((*succs())[0] == bot_start || ends_in_cti()) {
	    set_shadow(NULL); 
	} else {
	    // promote shadow. 
	    assert(succs()->count() == 1); 
	    assert(tiC == NULL); 
	    ti2 = tiC = tiS; 
	    tiS = NULL; 
	}
    } 

    // Move instructions from bottom clump 
    tree_node_list_e *ins_point = this->tiS
	? this->tiS->list_e()
	: this->ti2->list_e(); 
    tree_node_list_e *del_point = bot_start->ti1->list_e(); 
    tree_node_list_e *end_point = bot_end->tiS 
	? bot_end->tiS->list_e()->next()
	: bot_end->ti2->list_e()->next(); 
    do {
	tree_node_list_e *next = del_point->next(); 
	tree_node_list_e *temp = tnl->remove(del_point); 
	tnl->insert_after(temp, ins_point); 
	del_point = next; 
	ins_point = temp; 
    } while (del_point != end_point); 

    // Link the lsucc and lpred fields. 
    pfg_node::set_layout_succ(bot); 

    return inverted; 
}

/* Invert_cbr_op() -- architecture-specific helper routine that returns
 * the conditional branch opcode corresponding to the inversion of the
 * conditional branch passed as the 2nd parameter. */
if_ops
Invert_cbr_op(int orig_op)
{
    if (orig_op == io_btrue) return io_bfalse;
    else if (orig_op == io_bfalse) return io_btrue;
    else assert_msg(FALSE, ("Invert_cbr_op() -- unexpected CBR op %s",
			    orig_op));
    return io_btrue;
}

/* 
 * invert_branch() -- Invert the branch by changing the opcode. 
 * This is usually used in conjunction with set_succ() to 
 * swap the targets of a conditional branch while maintaining
 * program correctness.  
 */
void 
pfg_block::invert_branch()
{
    assert(tiC); 
    assert(Is_cbr(tiC->instr())); 

    int orig_op = tiC->instr()->opcode(); 
//     archinfo *target_arch = this->parent()->arch();
    if_ops new_op = (if_ops)Invert_cbr_op(orig_op);
    tiC->instr()->set_opcode(new_op); 
}

/* 
 * set_shadow() -- if necessary, add a shadow instruction to a block.
 */
void 
pfg_block::set_shadow(pfg_node *nod)
{
    // Remove any old ubr, if any
    if (ends_in_ubr()) {
	assert_msg(ti1 != ti2, ("set_shadow() -- "
	    "Deleted old ubr is entire block")); 
	if (tiC == ti2) 
	    ti2 = (tree_instr *)ti2->list_e()->prev()->contents; 
	delete tnl->remove(tiC->list_e()); 
	delete tiC;
	tiC = NULL; 
    }

    // Remove the old shadow, if any
    if (tiS != NULL) {
	delete tnl->remove(tiS->list_e());
	delete tiS;
	tiS = NULL;
    }

    if (nod == NULL || ends_in_mbr() || ends_in_return())
	return; 

    if (nod->is_block() == FALSE) {
	printf("\n");
	printf("pfg_block::set_shadow(): Warning. Trying to set shadow "
		"to a non-block node\n");
	printf("pfg_block::set_shadow(): source: ");
	this->print();
	printf("pfg_block::set_shadow(): target: ");
	nod->print();
	printf("\n");
// 	assert(nod->is_block());
	return;
    }

    pfg_block *new_succ = (pfg_block *)nod; 
    instruction *new_shadow;
    new_shadow = new in_bj(io_jmp, new_succ->get_label()); 
    tiS = new tree_instr(new_shadow); 
    tiS->annotes()->append(new annote(k_pfg_node, (void *)this));
    tnl->insert_after(tiS, in_tail()->list_e()); 
}

/* 
 * get_label() -- Find the label_sym that starts a pfg_block.
 * If there isn't one, add one and return it.  
 */
label_sym *
pfg_block::get_label(void)
{
    if (Is_label(in_head()->instr()))
	return Get_label(in_head()->instr());
    
    tree_proc *tp = parent()->tproc();
    proc_symtab *par_symtab = tp->proc_syms(); 
    label_sym *new_lab = par_symtab->new_unique_label(); 
    new_lab->annotes()->append(new annote(k_pfg_node, (void *)this));
    instruction *new_i; 
    new_i = new in_lab(new_lab); 

    tree_instr *new_ti = new tree_instr(new_i); 
    new_ti->annotes()->append(new annote(k_pfg_node, (void *)this)); 
    tnl->insert_before(new_ti, in_head()->list_e()); 
    set_in_head(new_ti); 
    return new_lab; 
}


/* set of simple test routines */
boolean 
pfg_block::ends_in_ubr() const
{
    return tiC && Is_ubr(tiC->instr()); 
}

boolean 
pfg_block::ends_in_cbr() const
{
    return tiC && Is_cbr(tiC->instr()); 
}

boolean 
pfg_block::ends_in_mbr() const
{
    return tiC && Is_mbr(tiC->instr()); 
}

boolean 
pfg_block::ends_in_call() const
{
    return tiC && Is_call(tiC->instr()); 
}

boolean 
pfg_block::ends_in_return() const
{
    return tiC && Is_return(tiC->instr()); 
}


/* 
 * pfg_block::first_non_label() -- returns pointer to the first
 * tnl element that is not a label.  Routine returns NULL if the
 * block is all label instructions. 
 */
tnle *
pfg_block::first_non_label (void) const
{
    tree_node_list_e *tnle = in_head()->list_e(); 
    while (tnle && Is_label(((tree_instr *)tnle->contents)->instr())
	&& tnle != in_tail()->list_e()) 
	tnle = tnle->next(); 
    if (tnle == NULL)
	return NULL; // block is all labels
    else
	return tnle; 
}

/* 
 * pfg_block::last_non_cti() -- returns pointer to the last
 * tnl element that is not a control instruction.  Routine returns NULL 
 * if the block is just a single control instruction. 
 */
tnle *
pfg_block::last_non_cti (void) const
{
    if (Is_cti(in_tail()->instr())) {
	if (in_tail() == in_head())
	    return NULL; // block is one cti instr
	else 
	    return in_tail()->list_e()->prev(); 
    } else {
	return in_tail()->list_e(); 
    }
}

/* pfg_block::last_exec()
 * Returns the last statement of the block that is not a label, mrk or nop
 * instruction.
 */
tree_instr *
pfg_block::last_exec() const
{
    tree_instr *ti2_exec = in_tail();
    if_ops op = ti2_exec->instr()->opcode();
    while ( (ti2_exec != in_head()) && 
	    (op == io_nop || op == io_mrk || op == io_lab) ) {
	ti2_exec = (tree_instr *)ti2_exec->list_e()->prev()->contents;
	op = ti2_exec->instr()->opcode();
    }

    /* If we reached the head of the block and we didn't find an executable
     * statement, return the last instruction.
     */
    if (ti2_exec == in_head() &&
	(op == io_nop || op == io_mrk || op == io_lab) ) {
	return in_tail();
    } else {
	return ti2_exec;
    }
}


/* pfg_block::first_exec()
 * Returns the first statement of the block that is not a label, mrk or nop
 * instruction.
 */
tree_instr *
pfg_block::first_exec() const
{
    tree_instr *ti1_exec = in_head();
    if_ops op = ti1_exec->instr()->opcode();
    while ( (ti1_exec != in_tail()) && 
	    (op == io_nop || op == io_mrk || op == io_lab) ) {
	ti1_exec = (tree_instr *)ti1_exec->list_e()->next()->contents;
	op = ti1_exec->instr()->opcode();
    }

    /* If we reached the end of the block and we didn't find an executable
     * statement, return the first instruction.
     */
    if (ti1_exec == in_tail() &&
	(op == io_nop || op == io_mrk || op == io_lab) ) {
	return in_head();
    } else {
	return ti1_exec;
    }
}



tnle *
pfg_block::push(tnle *new_e)
{
    return this->insert_before(new_e, in_head()->list_e()); 
}

tnle *
pfg_block::pop()
{
    return this->remove(in_head()->list_e()); 
}

tnle *
pfg_block::append(tnle *new_e)
{
    return this->insert_after(new_e, in_tail()->list_e()); 
}

tnle *
pfg_block::insert_before(tnle *new_e, tnle *pos)
{
    assert_msg(new_e->contents->is_instr(), 
	("pfg_block::insert_before() -- new_e not a TREE_INSTR")); 
    assert_msg(pfg_begin_node(new_e->contents) == NULL, 
	("pfg_block::insert_before() -- new_e has a pfg_node annotation")); 
    annote *an = new annote(k_pfg_node, (void *)this);
    new_e->contents->annotes()->append(an);

    if (Is_cti(((tree_instr *)new_e->contents)->instr())) {
	set_in_cti((tree_instr *)(new_e->contents)); 
    }
    tree_instr *new_head = pos->contents == in_head() 
	? (tree_instr *)(new_e->contents) : in_head(); 
    tnle *ret = tnl->insert_before(new_e, pos); 
    set_in_head(new_head); 
    return ret; 
}

/* pfg_block::insert_after() -- insert the single instruction new_e
 * after the instruction pos in the tree_node_list. */
tnle *
pfg_block::insert_after(tnle *new_e, tnle *pos)
{
    assert_msg(new_e->contents->is_instr(), 
	("pfg_block::insert_after() -- new_e not a TREE_INSTR")); 
    assert_msg(pfg_begin_node(new_e->contents) == NULL, 
	("pfg_block::insert_after() -- new_e has a pfg_node annotation")); 
    annote *an = new annote(k_pfg_node, (void *)this);
    new_e->contents->annotes()->append(an);

    if (Is_cti(((tree_instr *)new_e->contents)->instr())) {
	set_in_cti((tree_instr *)(new_e->contents)); 
    }
    tree_instr *new_tail = pos->contents == in_tail() 
	? (tree_instr *)(new_e->contents) : in_tail(); 
    tnle *ret = tnl->insert_after(new_e, pos); 
    set_in_tail(new_tail); 
    return ret; 
}

/* pfg_block::insert_after() -- insert the list of instructions mtnl
 * after the instruction pos in the tree_node_list. */
void
pfg_block::insert_after(tree_node_list *mtnl, tnle *pos)
{
    assert_msg(pos, ("pfg_block::insert_after() -- push not implemented"));
    while (!mtnl->is_empty()) 
	pos = insert_after(new tree_node_list_e(mtnl->pop()), pos); 
}

tnle *
pfg_block::remove(tnle *rem)
{
    assert_msg(rem->contents->is_instr(), 
	("pfg_block::remove() -- rem not a TREE_INSTR")); 
    assert_msg(this == pfg_begin_node(rem->contents), 
	("pfg_block::remove() -- instruction does not belong to this block")); 
    rem->contents->get_annote(k_pfg_node);

    if (in_head() == in_tail()) {
	assert(in_head() == rem->contents); 
	tree_proc *tp = parent()->tproc();
	proc_symtab *par_symtab = tp->proc_syms(); 
	label_sym *new_lab = par_symtab->new_unique_label(); 
	new_lab->annotes()->append(new annote(k_pfg_node, (void *)this));
	instruction *new_i; 
	new_i = new in_lab(new_lab);

	tree_instr *new_ti = new tree_instr(new_i); 
	new_ti->annotes()->append(new annote(k_pfg_node, (void *)this)); 
	tnl->insert_after(new_ti, in_head()->list_e()); 
	tnle *ret = tnl->remove(in_head()->list_e()); 
	set_in_head(new_ti); 
	set_in_tail(new_ti); 
	set_in_cti(NULL); 
	return ret; 
    } 

    tree_instr *new_head = rem->contents == in_head() 
	? (tree_instr *)(in_head()->list_e()->next()->contents)
	: in_head(); 
    tree_instr *new_tail = rem->contents == in_tail()
	? (tree_instr *)(in_tail()->list_e()->prev()->contents)
	: in_tail(); 
    if (rem->contents == in_cti()) {
	/* hope the user knows what he or she is doing when
	 * they remove the CTI instruction */
	set_in_cti(NULL); 
    }

    tnle *ret = tnl->remove(rem); 
    set_in_head(new_head); 
    set_in_tail(new_tail); 
    return ret; 
}


/**
 ** ------------- Print methods ------------- 
 **/

static void
print_node_list (FILE *fp, pfg_node_list *l)
{
    pfg_node_list_iter cnli(l);
    while (!cnli.is_empty()) {
        fprintf(fp, "%d", cnli.step()->number());
        if (!cnli.is_empty())
            fprintf(fp, " ");
    }
}


void
pfg_node::print_base (FILE *fp)
{
    fprintf(fp, "%s ", (this->is_parallel()) ? "par" : "seq");
    fprintf(fp, "p ");
    print_node_list(fp, preds());
    fprintf(fp, "\ts ");
    print_node_list(fp, succs());
    if (layout_succ())
	fprintf(fp, "\tl %d", layout_succ()->number());
}


void
pfg_begin::print (FILE *fp)
{
    if (this->is_parbegin()) {
	fprintf(fp, "    %2d: cobegin       ", number());
    } else {
	fprintf(fp, "    %2d: begin         ", number());
    }
    print_base(fp);
    putc('\n', fp);
}


void
pfg_end::print (FILE *fp)
{
    if (this->is_parend()) {
	fprintf(fp, "    %2d: coend         ", number());
    } else {
	fprintf(fp, "    %2d: end           ", number());
    }
    print_base(fp);
    putc('\n', fp);
}


void
pfg_label::print (FILE *fp)
{
    fprintf(fp, "    %2d: label         ", number());
    print_base(fp);
    putc('\n', fp);
}


void
pfg_instr::print (FILE *fp)
{
    fprintf(fp, "    %2d: instr         ", number());
    print_base(fp);
    putc('\n', fp);
}


void
pfg_block::print (FILE *fp)
{
    fprintf(fp, "    %2d: block ", number());
    char *terms = "      ";
    if (ends_in_ubr())
	terms = "ubr   ";
    else if (ends_in_cbr())
	terms = "cbr   ";
    else if (ends_in_mbr())
	terms = "mbr   ";
    else if (ends_in_call())
	terms = "call  ";
    else if (ends_in_return())
	terms = "ret   ";
    else if (ends_in_cti())
	terms = "CTI?  ";
    char *shads = (in_shad()) ? "S " : "  ";
    fprintf(fp, "%s%s", terms, shads);

    print_base(fp);
    putc('\n', fp);
}


void
pfg_test::print (FILE *fp)
{
    fprintf(fp, "    %2d: test          ", number());
    print_base(fp);
    putc('\n', fp);
}


/**
 ** ------------- Protected functions ------------- 
 **/

void
pfg_node::add_succ (pfg_node *n)
{
    /* avoid duplicate edges */
    pfg_node_list_iter succ_iter(succs());
    while (!succ_iter.is_empty()) {
	pfg_node *succ = succ_iter.step();
	if (succ == n) return;
    }

    /* add the forward and backward edges */
    succs()->append(n);
    n->preds()->append(this);
}


void
pfg_node::remove_pred (pfg_node *n)
{
    remove_node(preds(), n);
    remove_node(n->succs(), this);
}


void
pfg_node::remove_succ (pfg_node *n)
{
    remove_node(succs(), n);
    remove_node(n->preds(), this);
}


/* 
 * remove_explicit_targets() -- This procedure leaves control
 * transfer instructions in place, but strips off their targets.
 * This is done when we clone basic blocks, so that users of
 * the library will be forced to explicitly set the targets
 * of the cloned block.  
 */
void
pfg_block::remove_explicit_targets()
{
    /* Remove any explicit targets */
    instruction *lasti = ti2->instr(); 
    if (1 /* lasti->architecture() == k_suif */) {
	unsigned nl; 
	in_mbr *mbr2 = (in_mbr *)lasti; 

	switch (lasti->opcode()) {
	  case io_btrue: case io_bfalse: case io_jmp:
	    ((in_bj *)lasti)->set_target(NULL); 
	    break; 

	  case io_mbr:
	    mbr2->set_default_lab(NULL); 
	    nl = mbr2->num_labs(); 
	    for (unsigned i = 0; i < nl; ++i)
		mbr2->set_label(i, NULL); 
	    break; 

	  case io_cal: case io_ret:
	    break;

	  default:
	    break; 
	}

    } else if (Is_cti(lasti)) {
	if (Is_ubr(lasti) || Is_cbr(lasti)) {
	    ((in_bj *)lasti)->set_target(NULL); 

	} else if (Is_mbr(lasti)) {
	    /* Do nothing.  This is handled in pfg::attach(). */

	} else if (Is_call(lasti) || Is_return(lasti)) {

	} else {
	    assert(0); 
	}
    }
}


pfg_block *
pfg_block::tnl_pred(void)
{
    tree_node_list_e *tnle; 
    pfg_node *nod; 
    if ((tnle = ti1->list_e()->prev())
	&& (nod = pfg_begin_node(tnle->contents))
	&& nod->is_block())
	return (pfg_block *)nod; 
    return NULL; 
}


pfg_block *
pfg_block::tnl_succ(void)
{
    tree_node_list_e *tnle; 
    pfg_node *nod; 
    tree_instr *ti = tiS ? tiS : ti2; 
    if ((tnle = ti->list_e()->next())
	&& (nod = pfg_begin_node(tnle->contents))
	&& nod->is_block())
	return (pfg_block *)nod; 
    return NULL; 
}


//---------------------------------------------------------------------------
// pfg_node::insert_node_after()
// Insert this node after the given PFG node.
//---------------------------------------------------------------------------
void
pfg_node::insert_node_after(pfg_node *n)
{
    /* Add the node to the PFG */
    n->parent()->add_node(this);

    /* Update the annotation for the instructions in the node. This may not
     * always be needed, but we need it when we are splitting up basic
     * nodes at conflicting statements. It doesn't hurt anyway.
     */
    pfg_node_instr_iter instr(this);
    while (instr.is_empty() == FALSE) {
	/* Remove the previous annotation */
	tree_instr *ti = instr.step();
	annote *an = ti->annotes()->get_annote(k_pfg_node);
	delete an;

	/* Add a new one pointing to the new node */
	an = new annote(k_pfg_node, (void *)this);
	ti->annotes()->append(an);
    }

    /* Delete node n from the predecessor list in its successors and add
     * ourselves to that list
     */
    pfg_node_list_iter succ_iter(n->succs());
    while (succ_iter.is_empty() == FALSE) {
	pfg_node *succ = succ_iter.step();
	remove_node(succ->preds(), n);
	succ->preds()->append(this);
    }

    /* Set up our successors to be node n's successors */
    this->succs()->clear();
    this->succs()->append(n->succs());

    /* Set node n's successor to be us */
    n->succs()->clear();
    n->succs()->append(this);

    /* Set our predecessor to be node n */
    this->preds()->append(n);
}


//---------------------------------------------------------------------------
// pfg_marker::last_instr()
// Return the last instruction for this marker node. Note that we are
// actually interested in the last instruction in the header of the
// associated control structure, not the body. For instance, the last_instr
// of a while loop is the expression at the head of the loop.
//---------------------------------------------------------------------------
tree_instr *
pfg_marker::last_instr() const
{
    switch (tn->kind()) {
	case TREE_LOOP: return last_ti(((tree_loop *)tn)->test());
	case TREE_FOR:  return last_ti(((tree_for *)tn)->ub_list());
	case TREE_IF:   return last_ti(((tree_if *)tn)->header());
	case TREE_BLOCK: return last_ti(((tree_block *)tn)->body());
	default: {
	    if (this->is_parbegin() || this->is_parend()) {
		return (tree_instr *)tn;
	    }

	    assert_msg(FALSE, ("Invalid tree_node kind %d\n", tn->kind()));
	}
    }

    return NULL;
}

//---------------------------------------------------------------------------
// pfg_test::last_instr()
//---------------------------------------------------------------------------
tree_instr *
pfg_test::last_instr() const
{ 
	return last_ti(tf->ub_list());
}


/**
 ** ------------- Static functions ------------- 
 **/

/* remove_node() -- removes pfg_node n from the pfg_node_list l */
static void
remove_node (pfg_node_list *l, pfg_node *n)
{
    pfg_node_list_iter node_iter(l);
    while (!node_iter.is_empty()) {
	pfg_node *node = node_iter.step();
	if (node == n) {
	    delete l->remove(node_iter.cur_elem());
	    return;
	}
    }
    assert_msg(FALSE, ("remove_node - node '%u' not found", n->number()));
}
