/*  Control Flow Graph Nodes */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    Copyright (c) 1995, 1996, 1997 The President and Fellows of Harvard 
    University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file.

    Modified by Diego Novillo (diego@cs.ualberta.ca)
*/

#include <suif_copyright.h>


#ifndef PFG_NODE_H
#define PFG_NODE_H

class VarRef;
class PhiTerm;
class VarRef_list;
class RefConflict_list;
class cssame;

/*
 *  Control flow graph nodes are divided into several subclasses
 *  depending on the sort of SUIF object(s) to which they correspond.
 */

enum pfg_node_kind {
    PFG_BEGIN,                          /* beginning marker for AST node */
    PFG_END,                            /* ending marker for AST node */
    PFG_INSTR,                          /* instruction or expression tree */
    PFG_BLOCK,                          /* basic block of instructions */
    PFG_TEST,                           /* "test" node of a FOR loop */
    PFG_LABEL                           /* implicit AST node label */
};

/*
 *  Each flow graph node contains only an ID number and information
 *  related to the flow graph.  Other information can be associated with
 *  a node by storing it in a separate array in the pfg structure where
 *  the node ID number is used to index into the array.
 *
 *  This is a virtual class that cannot be instantiated.
 */

class pfg_node;
DECLARE_LIST_CLASS(pfg_node_list, pfg_node*);

class pfg_begin;
class pfg_end;

class pfg_node {
    friend class pfg;
    friend class cssame;

private:
    unsigned nnum;                      /* node number */
    pfg *par;                           /* parent flow graph */
    pfg_node_list prs;                  /* pfg predecessors */
    pfg_node_list scs;                  /* pfg successors */
    pfg_node_list _cpreds;		/* Conflict predecessors */
    pfg_node_list _csuccs;		/* Conflict successors */
    pfg_node_list _spreds;		/* Synchronization predecessors */
    pfg_node_list _ssuccs;		/* Synchronization successors */
    bit_set _prec;			/* Nodes guaranteed to precede me */
    bit_set _prec_ct;			/* Nodes guaranteed to precede me
					   through control flow */
    bit_set _prec_sy;			/* Nodes guaranteed to precede me
					   through synchronization flow */

  protected:
    pfg_node *lpred;                    /* layout predecessor */
    pfg_node *lsucc;                    /* layout successor */
    boolean _parallel;			/* Is this a parallel node? */
    pfg_node *_par_parent;		/* Parallel parent node (ie,
					   enclosing cobegin node) */
    threadBody *_thread;		/* Thread body for this node. NULL
					   if the node is not inside a
					   thread. */
    VarRef_list *_refs;			/* List of variables referenced in
					   this node. */
    RefConflict_list *_conflicts;	/* List of conflicts that have this
					   node as its head. */
    boolean _splitup;			/* Has this node been split up? */

    void set_number(unsigned n)         { nnum = n; }
    void set_parent(pfg *p)             { par = p; }

    void add_succ(pfg_node *n);
    void remove_pred(pfg_node *n);
    void remove_succ(pfg_node *n);

    void print_base(FILE *fp);

    /* Data flow methods */
    VarRef_list *findRefs(instruction *in);
    VarRef_list *findPtrRefs(instruction *in);
    var_sym *symRefBy(operand addr);
    var_sym_list *findAliases(instruction *in);

  public:
    pfg_node(boolean parallel=FALSE, pfg_node *par_parent=0, threadBody *t=0);
    virtual ~pfg_node() { }

    virtual pfg_node_kind kind() const = 0;
    boolean is_begin() const            { return (kind() == PFG_BEGIN); }
    boolean is_end() const              { return (kind() == PFG_END); }
    boolean is_instr() const            { return (kind() == PFG_INSTR); }
    boolean is_block() const            { return (kind() == PFG_BLOCK); }
    boolean is_test() const             { return (kind() == PFG_TEST); }
    boolean is_label() const            { return (kind() == PFG_LABEL); }
    boolean is_parbegin() const;		
    boolean is_parend() const;
    boolean is_parallel() const		{ return _parallel; }

    unsigned number()                   { return nnum; }
    pfg *parent()                       { return par; }

    pfg_node_list *preds()              { return &prs; }
    pfg_node_list *succs()              { return &scs; }
    pfg_node *fall_succ()               { return (*succs())[0]; }
    pfg_node *take_succ()               { return (*succs())[1]; }
    pfg_node *layout_pred() const       { return lpred; }
    pfg_node *layout_succ() const       { return lsucc; }

    virtual void set_succ(unsigned n, pfg_node *) = 0; 
    void set_fall_succ(pfg_node *nod)   { set_succ(0, nod); }
    void set_take_succ(pfg_node *nod)   { set_succ(1, nod); }

    virtual void clear_layout_succ();
    virtual boolean set_layout_succ(pfg_node *);
    boolean set_layout_succ(label_sym *); 

    void insert_empty_block(unsigned n); 
    void insert_empty_block(pfg_node *dst); 
    void insert_node_after(pfg_node *n);

    pfg_node *par_parent() const	{ return _par_parent; }
    bit_set *par_ancestors() const	{ return par->par_ancestors(nnum); }
    threadBody *thread() const		{ return _thread; }

    /* Data flow information */
    virtual VarRef_list *findRefs()	{ return NULL; }
    VarRef_list *refs()			{ return _refs; }
    RefConflict_list *conflicts()	{ return _conflicts; }
    virtual void split()		{ }
    boolean splitup()			{ return _splitup; }
    void splitup(boolean flag)		{ _splitup = flag; }


    /* Methods to access instructions contained in the node */
    virtual tree_instr *last_instr() const = 0;

    /* SSA helper methods */
    PhiTerm *addPhi(var_sym *var);
    PhiTerm *addPi(var_sym *var);
    PhiTerm *hasPi(var_sym *var);

    pfg_node_list *cpreds()		{ return &_cpreds; }
    pfg_node_list *csuccs()		{ return &_csuccs; }

    pfg_node_list *spreds()		{ return &_spreds; }
    pfg_node_list *ssuccs()		{ return &_ssuccs; }

    bit_set prec()		 	{ return _prec; }
    bit_set prec_ct()		 	{ return _prec_ct; }
    bit_set prec_sy()		 	{ return _prec_sy; }

    virtual void print(FILE *fp=stdout) = 0;
};

/*
 *  The BEGIN and END flow graph nodes are used to indicate the beginning
 *  and ending of tree_nodes (except for tree_instrs). Parallel section
 *  markers (cobegins and doalls) have the _parallel flag set to true.
 */
class pfg_marker : public pfg_node {
    friend class pfg;
    friend class cssame;

  private:
    tree_node *tn;

  public:
    pfg_marker(tree_node *t, boolean parallel=FALSE, pfg_node
	    *par_parent=0, threadBody *tb=0)
	: pfg_node(parallel, par_parent, tb)
    	{ tn = t; }

    tree_node *node()                   { return tn; }
    virtual tree_instr *last_instr() const;
};

class pfg_begin : public pfg_marker {
  public:
    pfg_begin(tree_node *t, boolean parallel=FALSE, pfg_node *par_parent=0,
	    threadBody *tb = 0) 
	: pfg_marker(t, parallel, par_parent, tb) {}

    pfg_node_kind kind() const                { return PFG_BEGIN; }

    void set_succ(unsigned n, pfg_node *); 
    boolean set_layout_succ(pfg_node *); 

    void print(FILE *fp=stdout);
};

class pfg_end : public pfg_marker {
  public:
    pfg_end(tree_node *t, boolean parallel=FALSE, pfg_node *par_parent=0,
	    threadBody *tb=0)
	: pfg_marker(t, parallel, par_parent, tb) {}

    pfg_node_kind kind() const                { return PFG_END; }

    void set_succ(unsigned, pfg_node *) { assert(0); }

    void print(FILE *fp=stdout);
};

/*
 *  LABEL nodes are only used in flow graphs built with separate nodes for
 *  each instruction (i.e. no basic blocks).  They are used for the implicit
 *  labels associated with high-level AST nodes.  The result is that only
 *  nodes associated with labels will have multiple predecessors -- in other
 *  words, any node with multiple predecessors will be empty.  That is a
 *  nice property for solving data flow problems.
 */

class pfg_label : public pfg_marker {
  public:
    pfg_label(tree_node *t, boolean parallel=FALSE, pfg_node *par_parent=0,
	    threadBody *tb=0)
	: pfg_marker(t, parallel, par_parent, tb) {}

    pfg_node_kind kind() const                { return PFG_LABEL; }

    void set_succ(unsigned, pfg_node *) { assert(0); }

    virtual tree_instr *last_instr() const	{ return NULL; }

    void print(FILE *fp=stdout);
};

/*
 *  The PFG_INSTR flow graph nodes correspond to individual SUIF instructions
 *  (not basic blocks).
 */

class pfg_instr : public pfg_node {
    friend class pfg;
    friend class cssame;

  private:
    tree_instr *ti;

  public:
    pfg_instr(tree_instr *t, boolean parallel=FALSE, pfg_node
	    *par_parent=0, threadBody *tb=0)
	: pfg_node(parallel, par_parent, tb)
    	{ ti = t; }

    pfg_node_kind kind() const                { return PFG_INSTR; }

    tree_instr *instr()                 { return ti; }
    void set_succ(unsigned, pfg_node *) { assert(0); }

    /* Data flow information */
    virtual VarRef_list *findRefs();

    virtual tree_instr *last_instr() const { return ti; }

    void print(FILE *fp=stdout);
};

/*
 *  The BLOCK flow graph nodes represent basic blocks.  Their boundaries
 *  are identified by the starting and ending tree_instr nodes.  Only other
 *  tree_instrs should be in the block -- no high-level tree_nodes.
 */
typedef tree_node_list_e tnle;  /* just shorthand for this declaration */

/* Structure to hold partition bounds for basic blocks */
struct Partition {
    tree_instr *ti1;
    tree_instr *ti2;
};

class pfg_block : public pfg_node {
    friend class pfg;
    friend class cssame;

  private:
    tree_node_list *tnl; 
    tree_instr *ti1;
    tree_instr *ti2;
    tree_instr *tiC;            /* CTI instruction, if any */
    tree_instr *tiS;            /* shadow goto, if needed */

  protected:
    void extend(tree_instr *t)          { ti2 = t; }
    void remove_explicit_targets(); 
    pfg_block *tnl_pred(void); 
    pfg_block *tnl_succ(void); 
    
    /* Helper functions for splitting up basic blocks at conflicting
     * def/uses and synchronization instructions
     */
    unsigned initialSplit(Partition *splitArray);
    void fixLowerBounds(Partition *splitArray, unsigned numPart);
    void fixUpperBounds(Partition *splitArray, unsigned numPart);
    void fixGaps(Partition **splitArray, unsigned *numPart);


  public:
    pfg_block(tree_node_list *tn, tree_instr *t1, tree_instr *t2, boolean
	    parallel=FALSE, pfg_node *par_parent=0, threadBody *tb = 0)
	: pfg_node(parallel, par_parent, tb)
        { tnl = tn; ti1 = t1; ti2 = t2; tiC = NULL; tiS = NULL; }
    pfg_block(pfg_block *orig, base_symtab *dst_scope=NULL, boolean
	    parallel=FALSE, pfg_node *par_parent=0, threadBody *tb=0); 
    pfg_block(block_symtab *dst_scope, boolean parallel=FALSE, pfg_node
	    *par_parent=0, threadBody *tb=0); 

    pfg_node_kind kind() const                { return PFG_BLOCK; }

    tree_instr *in_head() const         { return ti1; }
    tree_instr *in_tail() const         { return ti2; }
    tree_instr *in_cti() const          { return tiC; }
    tree_instr *in_shad() const         { return tiS; }
    void set_in_head(tree_instr *ti)    { ti1 = ti; } 
    void set_in_tail(tree_instr *ti)    { ti2 = ti; } 
    void set_in_cti(tree_instr *ti)     { tiC = ti; } // CLIFF: need more here
    void set_in_shad(tree_instr *ti)	{ tiS = ti; }

    void clear_layout_succ();
    void set_succ(unsigned n, pfg_node *); 
    boolean set_layout_succ(pfg_node *); 

    /* Opcode-changing utility routine.  Does not change succs. */
    void invert_branch(); 
    void set_shadow(pfg_node *); 
    label_sym *get_label(); 

    /* Information routines to assist in code layout. */
    boolean ends_in_cti() const         { return tiC != NULL; }
    boolean ends_in_ubr() const;
    boolean ends_in_cbr() const;
    boolean ends_in_mbr() const;
    boolean ends_in_call() const;
    boolean ends_in_return() const;
        
    tnle *first_non_label() const;
    tnle *last_non_cti() const;
    tree_instr *last_exec() const;
    tree_instr *first_exec() const;

    virtual tree_instr *last_instr() const { return last_exec(); }

    /* Routines to add/delete non-CTI instructions to/from a block.
     * Modeled on tree_node_list routines (dlist.h). */
    tnle *push(tnle *new_e);    /* new_e inserted before head */
    tnle *pop();                /* pop this */
    tnle *append(tnle *new_e);  /* new_e inserted after tail */
    tnle *insert_before(tnle *e, tnle *pos);
    tnle *insert_after(tnle *e, tnle *pos);
    void insert_after(tree_node_list *mtnl, tnle *pos);
    tnle *remove(tnle *rem);    /* remove this from list */

    /* Data flow information */
    virtual VarRef_list *findRefs();
    virtual void split();

    void print(FILE *fp=stdout);
};

/*
 *  The test part of a tree_for node is generated automatically when the
 *  tree_for is expanded, yet it must still be represented in flow graphs.
 *  The TEST flow graph nodes are used for this special purpose.
 */

class pfg_test : public pfg_node {
    friend class pfg;
    friend class cssame;

  private:
    tree_for *tf;

  public:
    pfg_test(tree_for *t, boolean parallel=FALSE, pfg_node *par_parent=0,
	    threadBody *tb=0) 
	: pfg_node(parallel, par_parent, tb)
    	{ tf = t; }

    pfg_node_kind kind() const                { return PFG_TEST; }

    tree_for *for_loop()                { return tf; }
    void set_succ(unsigned, pfg_node *) { assert(0); }

    virtual tree_instr *last_instr() const;

    void print(FILE *fp=stdout);
};


#endif /* PFG_NODE_H */
