/*  Control Flow Graph Utility Implementations */

/*  Copyright (c) 1994 Stanford University

    All rights reserved.

    Copyright (c) 1995 The President and Fellows of Harvard University

    All rights reserved.

    This software is provided under the terms described in
    the "suif_copyright.h" include file. 


    Modified by Diego Novillo (diego@cs.ualberta.ca)
*/

#include <suif_copyright.h>

#define _MODULE_ "libpfg.a"

#include <suif.h>
#include <par.h>
#include <odc-util.h>

#include "pfg.h"

/*  annotation names */
char *k_pfg_node;
char *k_pfg_begin;
char *k_pfg_end;
char *k_pfg_test;
char *k_pfg_toplab;
char *k_proc_entry;

/* Annotations for data references and synchronization instructions */
char *k_refs;
char *k_defs;
char *k_uses;
char *k_phi_terms;
char *k_pi_terms;
char *k_head_sync;
char *k_tail_sync;
char *k_mutex_sync;
char *k_entry_sync_part;
char *k_exit_sync_part;


/*
 *  Initialization and finalization functions for the PFG library.
 */
void
init_pfg (int & /* argc */, char * /* argv */ [])
{
    /* control flow graph annotations are unregistered */
    k_pfg_node = lexicon->enter("pfg_node")->sp;
    k_pfg_begin = lexicon->enter("pfg_begin")->sp;
    k_pfg_end = lexicon->enter("pfg_end")->sp;
    k_pfg_test = lexicon->enter("pfg_test")->sp;
    k_pfg_toplab = lexicon->enter("pfg_toplab")->sp;
    k_proc_entry = lexicon->enter("proc_entry")->sp;

    /* Data and synchronization annotations */
    k_refs = lexicon->enter("refs")->sp;
    k_defs = lexicon->enter("defs")->sp;
    k_uses = lexicon->enter("uses")->sp;
    k_phi_terms = lexicon->enter("phi_terms")->sp;
    k_pi_terms = lexicon->enter("pi_terms")->sp;
    k_head_sync = lexicon->enter("head_sync")->sp;
    k_tail_sync = lexicon->enter("tail_sync")->sp;
    k_mutex_sync = lexicon->enter("mutex_sync")->sp;
    k_entry_sync_part = lexicon->enter("entry_sync_part")->sp;
    k_exit_sync_part = lexicon->enter("exit_sync_part")->sp;
}


void
exit_pfg ()
{
}


/*
 *  Retrieve the node corresponding to the position of a label in a control
 *  flow graph.  This function will return NULL if the flow graph has not yet
 *  been constructed.
 */

pfg_node *
label_pfg_node (label_sym *l)
{
    if (l->are_annotations()) {
	annote *an = l->annotes()->peek_annote(k_pfg_node);
	if (an) return (pfg_node *)an->data();
    }
    return NULL;
}


/*
 *  Retrieve the nodes corresponding to the beginning or end of an AST node
 *  in a control flow graph.  For instruction nodes, the result is the
 *  flow graph node that holds the instruction; for other nodes, it is
 *  the begin/end marker node.  These functions will return NULL if the
 *  flow graph has not yet been constructed.
 */

pfg_node *
pfg_begin_node (tree_node *n)
{
    char *an_name = k_pfg_begin;
    if (n->is_instr() && isCobegin(n) == FALSE) an_name = k_pfg_node;

    if (n->are_annotations()) {
	annote *an = n->annotes()->peek_annote(an_name);
	if (an) return (pfg_node *)an->data();
    }
    return NULL;
}


pfg_node *
pfg_end_node (tree_node *n)
{
    char *an_name = k_pfg_end;
    if (n->is_instr() && isCobegin(n) == FALSE) an_name = k_pfg_node;

    if (n->are_annotations()) {
	annote *an = n->annotes()->peek_annote(an_name);
	if (an) return (pfg_node *)an->data();
    }
    return NULL;
}


/*
 *  Retrieve the "test" and "top" nodes for a FOR loop in a control flow
 *  graph.  The separate node for the invisible "top" label is only used
 *  when the graph is built with nodes for individual instructions instead
 *  of basic blocks.  These functions will return NULL if the flow graph has
 *  not yet been constructed.
 */

pfg_test *
pfg_test_node (tree_for *n)
{
    if (n->are_annotations()) {
	annote *an = n->annotes()->peek_annote(k_pfg_test);
	if (an) return (pfg_test *)an->data();
    }
    return NULL;
}


pfg_label *
pfg_toplab (tree_for *n)
{
    if (n->are_annotations()) {
	annote *an = n->annotes()->peek_annote(k_pfg_toplab);
	if (an) return (pfg_label *)an->data();
    }
    return NULL;
}

pfg_node *
pfg_holding_node(tree_node *n)
{
    char *an_name = k_pfg_node;

    if (n->are_annotations()) {
	annote *an = n->annotes()->peek_annote(an_name);
	if (an) return (pfg_node *)an->data();
    }
    return NULL;
}


/*
 * Generate vcg format file for viewing with xvcg
 */

void
generate_vcg(FILE *f, pfg *pfg)
{
    /* Use minbackward layout algorithm in an attempt to make all backedges  */
    /* true backedges. Doesn't always work, but the back algorithm for this. */
    /* Change these parameters as needed, but they have worked fine so far.  */
    fprintf(f,"graph: { title: \"PFG_GRAPH\"\n");
    fprintf(f,"\n");    
    fprintf(f,"x: 30\n");
    fprintf(f,"y: 30\n");
    fprintf(f,"height: 800\n");
    fprintf(f,"width: 500\n");
    fprintf(f,"stretch: 60\n");
    fprintf(f,"shrink: 100\n");
    fprintf(f,"layoutalgorithm: minbackward\n");
    fprintf(f,"node.borderwidth: 3\n");
    fprintf(f,"node.color: white\n");
    fprintf(f,"node.textcolor: black\n");
    fprintf(f,"node.bordercolor: black\n");
    fprintf(f,"edge.color: black\n");
    fprintf(f,"\n");

    /* Put down nodes */
    int num = pfg->num_nodes();	
    int i; 
    for (i=0; i<num; i++)
	fprintf(f,"node: { title:\"%d\" label:\"%d\" }\n", \
		pfg->node(i)->number(),pfg->node(i)->number());
    fprintf(f,"\n");

    /* Put down edges and backedges */
    for (i=0; i<num; i++) {
	pfg_node_list_iter pfgnli(pfg->node(i)->preds());
	pfg_node *itnd;
	while (!pfgnli.is_empty()) {
	    itnd = pfgnli.step();
	    if (itnd->number()<pfg->node(i)->number())
		fprintf(f,"edge: {sourcename:\"%d\" targetname:\"%d\" }\n",\
			itnd->number(), pfg->node(i)->number());
	    else
		fprintf(f,"backedge: {sourcename:\"%d\" targetname:\"%d\" }\n"\
			,itnd->number(), pfg->node(i)->number());
    }
    }

    /* Wrap up at the end */
    fprintf(f,"\n");
    fprintf(f, "}\n");
}


/*
 * pfg_node_instr_iter - iterator over pfg_block instructions
 */

/* Constructor */

pfg_node_instr_iter::pfg_node_instr_iter(pfg_node *nod, boolean reverse,
	boolean exec)
{
    if (nod->is_block()) {
	pfg_block *blk = (pfg_block *)nod; 
	/* If the 'exec' flag is set to TRUE, we will skip any nop
	 * instructions at the beginning or at the end of the node
	 */
	if (exec) {
	    first = blk->first_exec()->list_e();
	    last = blk->last_exec()->list_e();
	} else {
	    first = blk->in_head()->list_e();
	    last = blk->in_tail()->list_e();
	}
    } else if (nod->is_instr()) {
	pfg_instr *ins = (pfg_instr *)nod; 
	first = ins->instr()->list_e(); 
	last = ins->instr()->list_e(); 
    } else {
	assert_msg(FALSE, ("pfg_node_instr_iter -- Unimplemented node type")); 
    }
    sentinel = reverse ? first : last; 
    rev = reverse; 
    reset(); 
}

/* Reset the iter to the beginning of the basic block */

void
pfg_node_instr_iter::reset()
{
    cur = NULL;
    nxt = rev ? last : first;
}

/* Step the iter to the next instruction */

tree_instr *
pfg_node_instr_iter::step()
{
    /* Return NULL if at end of basic block */
    if (cur == sentinel)
	return NULL;

    cur = nxt;
    nxt = rev ? nxt->prev() : nxt->next();
    return (tree_instr *)cur->contents;
}

/* Look at the next instruction, but don't step the iter */

tree_instr *
pfg_node_instr_iter::peek()
{
    /* Return NULL if at end of basic block */
    if (cur == sentinel)
	return NULL;

    return (tree_instr *)nxt->contents;
}

/* Am I at the end of the instructions in this basic block? */

boolean
pfg_node_instr_iter::is_empty()
{
    return (cur == sentinel);
}

/**
 ** ------------- Helper functions ------------- 
 **/

/* occurs() -- returns TRUE if cn is an element of cnl */ 
boolean
occurs (pfg_node *cn, pfg_node_list *cnl)
{
    pfg_node_list_iter pred_iter(cnl); 
    while (!pred_iter.is_empty()) 
	if (pred_iter.step() == cn) 
	    return TRUE; 
    return FALSE; 
}

//---------------------------------------------------------------------------
// first_ti()
// Returns the first tree instruction node from the given tree_node list.
//---------------------------------------------------------------------------
tree_instr *
first_ti(tree_node_list *tnl)
{
    tree_node_list_iter iter(tnl);
    while (iter.is_empty() == FALSE) {
	tree_node *node = iter.step();
	if (node->is_instr()) {
	    return (tree_instr *)node;
	}
    }

    return NULL;
}

//---------------------------------------------------------------------------
// last_ti()
// Returns the last tree instruction node from the given tree_node list.
//---------------------------------------------------------------------------
tree_instr *
last_ti(tree_node_list *tnl)
{
    int j;
    for (j = tnl->count() - 1; j >= 0; j--) {
	tree_node *node = (*tnl)[j];
	if (node->is_instr()) {
	    return (tree_instr *)node;
	} else if (node->is_block()) {
	    return last_ti( ((tree_block *)node)->body());
	}
    }

    return NULL;
}
