//---------------------------------------------------------------------------
// $Id: varref.cc,v 1.3 1998/06/03 16:52:38 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>

#define _MODULE_ "libcssame.a"

#include <string.h>

#include <suif.h>
#include <par.h>
#include <odc-util.h>

#include "pfg.h"

//---------------------------------------------------------------------------
// VarRef::VarRef()
//---------------------------------------------------------------------------
VarRef::VarRef(var_sym *var, tree_instr *stmt, pfg_node *node)
{ 
    _var	= var; 
    _stmt	= stmt; 
    _node	= node; 
    _conflicts	= NULL;
    _chain	= NULL;
    _save_chain	= NULL;
    _rdefs	= NULL;
    _ruses	= NULL;
    _mark	= NULL;
}


//---------------------------------------------------------------------------
// VarRef::~VarRef()
//---------------------------------------------------------------------------
VarRef::~VarRef()
{
    delete _conflicts;
}


//---------------------------------------------------------------------------
// VarRef::set_reachingDefs()
// Sets the list of definitions reaching this reference to the given list.
//---------------------------------------------------------------------------
void
VarRef::set_reachingDefs(VarRef_list *rdefs)
{
    if (_rdefs) {
	_rdefs->erase();
	delete _rdefs;
    }
    _rdefs = rdefs;
}


//---------------------------------------------------------------------------
// VarRef::set_reachedUses()
// Sets the list of uses reached by this reference to the given list.
//---------------------------------------------------------------------------
void
VarRef::set_reachedUses(VarRef_list *ruses)
{
    if (_ruses) {
	_ruses->erase();
	delete _ruses;
    }
    _ruses = ruses;
}


//---------------------------------------------------------------------------
// VarRef::print()
//---------------------------------------------------------------------------
void
VarRef::print(FILE *f)
{
    fprintf(f, "%s %s at line %d", _var->name(), this->ckind(), this->lineno());

    if (_chain && this->isU()) {
	fprintf(f, " -> ");
	_chain->print(f);
    }

    fflush(f);
}


//---------------------------------------------------------------------------
// VarRef::print_brief()
//---------------------------------------------------------------------------
void
VarRef::print_brief(FILE *f)
{
    fprintf(f, "%s_%s(%d)", this->ckind(), _var->name(), this->lineno());

    if (_chain && this->isU()) {
	fprintf(f, " -> ");
	_chain->print_brief(f);
    }

    fflush(f);
}


//***************************************************************************
//			 Methods for class PhiTerm
//***************************************************************************
//---------------------------------------------------------------------------
// PhiTerm::PhiTerm()
//---------------------------------------------------------------------------
PhiTerm::PhiTerm(var_sym *var, pfg_node *node, RefKind kind)
    : VarRef(var, NULL, node)
{
    static char *__FNAME = "PhiTerm::PhiTerm";

    _kind = kind;
    _phi_chain = new VarRef_list;
    assert(_phi_chain);

    /* Attach the term to the first instruction of the given PFG node */
    if (node->is_begin() || node->is_end() || node->is_label()) {
	tree_node *tn = ((pfg_marker *)node)->node();
	if (tn->is_instr()) {
	    _stmt = (tree_instr *)tn;
	    _lineno = source_line_num(_stmt);
	} else if (tn->is_loop()) {
	    if (node->is_begin()) {
		_stmt = first_ti( ((tree_loop *)tn)->body());
		_lineno = source_line_num(_stmt);
	    } else {
		_stmt = last_ti( ((tree_loop *)tn)->test());
		_lineno = source_line_num(_stmt);
	    }
	} else if (tn->is_for()) {
	    if (node->is_begin()) {
		_stmt = first_ti( ((tree_for *)tn)->lb_list());
		_lineno = source_line_num(_stmt);
	    } else {
		_stmt = last_ti( ((tree_for *)tn)->body());
		_lineno = source_line_num(_stmt) + 1;
	    }
	} else if (tn->is_if()) {
	    if (node->is_begin()) {
		_stmt = first_ti( ((tree_if *)tn)->header() );
		_lineno = source_line_num(_stmt);
	    } else {
		tree_if *tif = (tree_if *)tn;
		if (tif->else_part()->is_empty() == FALSE) {
		    _stmt = last_ti(tif->else_part());
		} else {
		    _stmt = last_ti(tif->then_part());
		}
		_lineno = source_line_num(_stmt) + 1;
	    }
	} else if (tn->is_block() || tn->is_proc()) {
	    if (node->is_begin()) {
		_stmt = first_ti( ((tree_block *)tn)->body() );
		_lineno = source_line_num(_stmt);
	    } else {
		_stmt = last_ti( ((tree_block *)tn)->body() );
		_lineno = source_line_num(_stmt);
	    }
	} else {
	    printf("%s(): Unknown tree node kind %d at PFG marker node %d\n",
		    __FNAME, tn->kind(), node->number());
	    assert(FALSE);
	}
    } else if (node->is_instr()) {
	_stmt = ((pfg_instr *)node)->instr();
	_lineno = source_line_num(_stmt);
    } else if (node->is_block()) {
	_stmt = ((pfg_block *)node)->in_head();
	_lineno = source_line_num(_stmt);
    } else if (node->is_test()) {
	_stmt = first_ti(((pfg_test *)node)->for_loop()->lb_list());
	_lineno = source_line_num(_stmt);
    } else {
	printf("%s(): Unknown node kind %d. This is bad.\n",  __FNAME,
		node->kind());
	assert(FALSE);
    }

    if (_stmt == NULL) {
	printf("\n%s(): Warning. Could not find a tree_instr for "
		"reference\n\t", __FNAME);
	this->print();
	printf("\n%s(): in PFG node:\n\t", __FNAME);
	node->print();
	printf("\n");
	_lineno = 0;
    }

    char *antype;
    switch (_kind) {
	case PHI: antype = k_phi_terms; break;
	case PI: antype = k_pi_terms; break;
	default: assert_msg(FALSE, ("Unkown reference kind %d\n", 
			 _kind));
    }

    this->addTo(_var, antype);
    if (_stmt) {
	this->addTo(_stmt->instr(), antype);
    }
}

//---------------------------------------------------------------------------
// PhiTerm::~PhiTerm()
//---------------------------------------------------------------------------
PhiTerm::~PhiTerm()
{
    delete _phi_chain;

    char *antype;
    switch (_kind) {
	case PHI: antype = k_phi_terms; break;
	case PI: antype = k_pi_terms; break;
	default: assert_msg(FALSE, ("Unkown reference kind %d\n", 
			 _kind));
    }

    this->removeFrom(_var, antype);
    if (_stmt) {
	this->removeFrom(_stmt->instr(), antype);
    }
}


//---------------------------------------------------------------------------
// PhiTerm::print()
//---------------------------------------------------------------------------
void
PhiTerm::print(FILE *f)
{
    int i, last;

    fprintf(f, "%s_%s(%d) = %s(", this->ckind(), _var->name(),
	    this->lineno(), this->ckind()
	    );

    last = _phi_chain->count();
    for (i = 0; i < last; i++) {
	if ((*_phi_chain)[i]) {
	    (*_phi_chain)[i]->print_brief(f);
	    if (i != last - 1) {
		fprintf(f, ", ");
	    }
	}
    }
    fprintf(f, ")");

    fflush(f);
}


//***************************************************************************
//			 Methods for class VarDef
//***************************************************************************
//---------------------------------------------------------------------------
// VarDef::VarDef()
//---------------------------------------------------------------------------
VarDef::VarDef(var_sym *var, tree_instr *stmt, pfg_node *node)
    : VarRef(var, stmt, node)
{
    this->addTo(_var, k_defs);
    assert(_stmt);
    this->addTo(_stmt->instr(), k_defs);
}

//---------------------------------------------------------------------------
// VarDef::~VarDef()
//---------------------------------------------------------------------------
VarDef::~VarDef()
{
    this->removeFrom(_var, k_defs);
    assert(_stmt);
    this->removeFrom(_stmt->instr(), k_defs);
}


//***************************************************************************
//			 Methods for class VarUse
//***************************************************************************
//---------------------------------------------------------------------------
// VarUse::VarUse()
//---------------------------------------------------------------------------
VarUse::VarUse(var_sym *var, tree_instr *stmt, pfg_node *node)
    : VarRef(var, stmt, node)
{
    this->addTo(_var, k_uses);
    assert(_stmt);
    this->addTo(_stmt->instr(), k_uses);
}

//---------------------------------------------------------------------------
// VarUse::~VarUse()
//---------------------------------------------------------------------------
VarUse::~VarUse()
{
    this->removeFrom(_var, k_uses);
    assert(_stmt);
    this->removeFrom(_stmt->instr(), k_uses);
}


//***************************************************************************
//			Methods for class SyncRef
//***************************************************************************
//---------------------------------------------------------------------------
// SyncRef::SyncRef()
//---------------------------------------------------------------------------
SyncRef::SyncRef(var_sym *var, tree_instr *stmt, pfg_node *node)
    : VarRef(var, stmt, node)
{
    instruction *instr = stmt->instr();
    if_ops op = instr->opcode();
    assert(op == io_cal);

    /* Retrieve the name of the function being called */
    in_cal *call = (in_cal *)instr;
    operand addr = call->addr_op();
    assert(addr.is_instr() && addr.instr()->opcode() == io_ldc);
    sym_node *fn = ((in_ldc *)addr.instr())->value().addr().symbol();
    _name = fn->name();
}

//---------------------------------------------------------------------------
// SyncRef::print()
//---------------------------------------------------------------------------
void
SyncRef::print(FILE *f)
{
    fprintf(f, "%s(%s) at line %d", _name, _var->name(), this->lineno());
    fflush(f);
}


//***************************************************************************
//		       Methods for class HeadSyncRef
//***************************************************************************
//---------------------------------------------------------------------------
// HeadSyncRef::HeadSyncRef()
//---------------------------------------------------------------------------
HeadSyncRef::HeadSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node)
    : SyncRef(var, stmt, node)
{
    this->addTo(_var, k_head_sync);
    assert(_stmt);
    this->addTo(_stmt->instr(), k_head_sync);
}


//---------------------------------------------------------------------------
// HeadSyncRef::~HeadSyncRef()
//---------------------------------------------------------------------------
HeadSyncRef::~HeadSyncRef()
{
    this->removeFrom(_var, k_head_sync);
    assert(_stmt);
    this->removeFrom(_stmt->instr(), k_head_sync);
}


//***************************************************************************
//		       Methods for class TailSyncRef
//***************************************************************************
//---------------------------------------------------------------------------
// TailSyncRef::TailSyncRef()
//---------------------------------------------------------------------------
TailSyncRef::TailSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node)
    : SyncRef(var, stmt, node)
{
    this->addTo(_var, k_tail_sync);
    assert(_stmt);
    this->addTo(_stmt->instr(), k_tail_sync);
}


//---------------------------------------------------------------------------
// TailSyncRef::~TailSyncRef()
//---------------------------------------------------------------------------
TailSyncRef::~TailSyncRef()
{
    this->removeFrom(_var, k_tail_sync);
    assert(_stmt);
    this->removeFrom(_stmt->instr(), k_tail_sync);
}


//***************************************************************************
//		      Methods for class MutexSyncRef
//***************************************************************************
//---------------------------------------------------------------------------
// MutexSyncRef::MutexSyncRef()
//---------------------------------------------------------------------------
MutexSyncRef::MutexSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node)
    : UndirSyncRef(var, stmt, node) 
{
    /* Create new partition bit sets for the variable, if it doesn't have
     * one already. These bit sets are used to group all the synchronization
     * operations for the same variable. There are two kinds of partitions,
     * entry and exit. 
     * 
     * An entry partition is the set of nodes in the graph that try to
     * acquire control on the variable (e.g., lock operations)
     *
     * An exit partition is the set of nodes in the graph that try to
     * release control from the variable (e.g., unlock operations).
     *
     * Once the partition has been created, add this node into the entry
     * or exit set depending on the specific synchronization instruction.
     */
    bit_set *npart = (bit_set *)_var->peek_annote(k_entry_sync_part);
    if (npart == NULL) {
	npart = new bit_set(0, node->parent()->num_nodes());
	assert(npart);
	_var->append_annote(k_entry_sync_part, (void *)npart);
    }

    bit_set *xpart = (bit_set *)_var->peek_annote(k_exit_sync_part);
    if (xpart == NULL) {
	xpart = new bit_set(0, node->parent()->num_nodes());
	assert(xpart);
	_var->append_annote(k_exit_sync_part, (void *)xpart);
    }

    if (strcmp(_name, "suif_lock") == 0) {
	npart->add(node->number());
    } else if (strcmp(_name, "suif_unlock") == 0) {
	xpart->add(node->number());
    } else {
	assert_msg(FALSE, ("MutexSyncRef::MutexSyncRef(): Unrecognized "
		"synchronization instruction %s\n", _name));
    }

    this->addTo(_var, k_mutex_sync);
    assert(_stmt);
    this->addTo(_stmt->instr(), k_mutex_sync);
}


//---------------------------------------------------------------------------
// MutexSyncRef::~MutexSyncRef()
//---------------------------------------------------------------------------
MutexSyncRef::~MutexSyncRef()
{
    bit_set *part = (bit_set *)_var->get_annote(k_entry_sync_part);
    delete part;

    part = (bit_set *)_var->get_annote(k_exit_sync_part);
    delete part;

    this->removeFrom(_var, k_mutex_sync);
    assert(_stmt);
    this->removeFrom(_stmt->instr(), k_mutex_sync);
}



//***************************************************************************
//		       Methods for class RefConflict
//***************************************************************************
//---------------------------------------------------------------------------
// RefConflict::print()
//---------------------------------------------------------------------------
void
RefConflict::print(FILE *f)
{
    _tail->print(f);
    fprintf(f, " conflicts with ");
    _head->print(f);
    fprintf(f, "\n");
}


//***************************************************************************
//			      Free functions
//***************************************************************************
//---------------------------------------------------------------------------
// Is_sync()
// Returns TRUE if the given instruction is a call to one of the given
// function names.
//---------------------------------------------------------------------------
static boolean
Is_sync(instruction *instr, char **names)
{
    static char *__FNAME = "Is_sync";

    if_ops op = instr->opcode();
    if (op != io_cal) return FALSE;

    /* Retrieve the name of the function being called */
    in_cal *call = (in_cal *)instr;
    operand addr = call->addr_op();
    assert(addr.is_instr() && addr.instr()->opcode() == io_ldc);
    sym_node *fn = ((in_ldc *)addr.instr())->value().addr().symbol();

    unsigned i = 0;
    while (names[i] != NULL) {
	if (strcmp(fn->name(), names[i]) == 0) {
	    if (debug(__FNAME, 1)) {
		printf("\n");
		printf("%s(): Found sync function %s in instruction\n",
			__FNAME, names[i]);
		instr->print();
		printf("\n");
	    }
	    return TRUE;
	}
	i++;
    }

    return FALSE;
}


//---------------------------------------------------------------------------
// Is_head_sync()
// Returns TRUE if the given instruction is a call to a synchronization
// function that is supposed to be at the beginning of basic blocks (eg,
// _Wait())
//---------------------------------------------------------------------------
boolean
Is_head_sync(instruction *instr)
{
    
    static char *_head_sync_names[] = { "_wait", NULL };
    return Is_sync(instr, _head_sync_names);
}


//---------------------------------------------------------------------------
// Is_tail_sync()
// Returns TRUE if the given instruction is a call to a synchronization
// function that is supposed to be at the end of basic blocks (eg,
// _Set())
//---------------------------------------------------------------------------
boolean
Is_tail_sync(instruction *instr)
{
    static char *_tail_sync_names[] = { "_set", NULL };
    return Is_sync(instr, _tail_sync_names);
}

//---------------------------------------------------------------------------
// Is_mutex_sync()
// Returns TRUE if the given instruction is a call to a mutex
// synchronization function.
//---------------------------------------------------------------------------
boolean
Is_mutex_sync(instruction *instr)
{
    static char *_mutex_sync_names[] = { "suif_lock", "suif_unlock", NULL };
    return Is_sync(instr, _mutex_sync_names);
}


//---------------------------------------------------------------------------
// Is_lock_sync()
// Returns TRUE if the given instruction is a call to suif_lock()
//---------------------------------------------------------------------------
boolean
Is_lock_sync(instruction *instr)
{
    static char *_lock_sync_names[] = { "suif_lock", NULL };
    return Is_sync(instr, _lock_sync_names);
}


//---------------------------------------------------------------------------
// Is_unlock_sync()
// Returns TRUE if the given instruction is a call to suif_unlock()
//---------------------------------------------------------------------------
boolean
Is_unlock_sync(instruction *instr)
{
    static char *_unlock_sync_names[] = { "suif_unlock", NULL };
    return Is_sync(instr, _unlock_sync_names);
}


//***************************************************************************
//			 Private helper methods
//***************************************************************************
//---------------------------------------------------------------------------
// VarRef::addTo()
// Adds this reference to the list of references for the given SUIF object.
// This method creates or updates an annotation on the object based on the
// kind of this reference.
//---------------------------------------------------------------------------
void
VarRef::addTo(suif_object *obj, char *antype)
{
    VarRef_list *refs;

    assert(obj);

    /* Add the reference to a particular list of references depending on
     * the reference kind. If the object contains the annotation we are
     * looking for, we retrieve it. Otherwise, we create a new one.
     */
    refs = (VarRef_list *)obj->peek_annote(antype);
    if (refs == NULL) {
	refs = new VarRef_list;
	assert(refs);
	obj->append_annote(antype, (void *)refs);
    }
    refs->append(this);

    /* Add the new reference to the global list of references to the
     * object.
     */
    refs = (VarRef_list *)obj->peek_annote(k_refs);
    if (refs == NULL) {
	refs = new VarRef_list;
	assert(refs);
	obj->append_annote(k_refs, (void *)refs);
    }
    refs->append(this);
}


//---------------------------------------------------------------------------
// VarRef::removeFrom()
// Removes this reference from the appropriate list of references for the
// given SUIF object.
//---------------------------------------------------------------------------
void
VarRef::removeFrom(suif_object *obj, char *antype)
{
    VarRef_list *refs;

    assert(obj);

    /* Remove this reference from the lists of references to the object. */
    refs = (VarRef_list *)obj->peek_annote(antype);
    assert(refs);
    VarRef_list_e *elem = refs->lookup(this);
    refs->remove(elem);

    /* Now remove it from the global list of references. */
    refs = (VarRef_list *)obj->peek_annote(k_refs);
    assert(refs);
    elem = refs->lookup(this);
    refs->remove(elem);
}
