//---------------------------------------------------------------------------
// $Id: varref.h,v 1.3 1998/06/03 16:52:38 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#ifndef __VAR_REF_H
#define __VAR_REF_H

#include <suif_copyright.h>

enum RefKind { D, U, PHI, PI, HSYNC, TSYNC, USYNC, MUTEXSYNC, LOCK, UNLOCK };

class RefConflict;
class VarRef;
class VarDef;
class PhiTerm;

DECLARE_DLIST_CLASS(RefConflict_list, RefConflict *);
DECLARE_DLIST_CLASS(VarRef_list, VarRef *);
DECLARE_DLIST_CLASS(VarDef_list, VarDef *);
DECLARE_DLIST_CLASS(PhiTerm_list, PhiTerm *);

//---------------------------------------------------------------------------
// Miscellaneous functions.
//---------------------------------------------------------------------------
Is_head_sync(instruction *instr);
Is_tail_sync(instruction *instr);
Is_mutex_sync(instruction *instr);
Is_lock_sync(instruction *instr);
Is_unlock_sync(instruction *instr);

//---------------------------------------------------------------------------
// class VarRef
//---------------------------------------------------------------------------
class VarRef {
    friend class pfg;

protected:
    var_sym *_var;
    tree_instr *_stmt;
    pfg_node *_node;
    RefConflict_list *_conflicts;
    VarRef_list *_rdefs;
    VarRef_list *_ruses;
    VarRef *_chain;
    VarRef *_save_chain;
    VarRef *_mark;

    void addTo(suif_object *obj, char *antype);
    void removeFrom(suif_object *obj, char *antype);
    
public:
    VarRef(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~VarRef();

    var_sym *var() const		{ return _var; }
    tree_instr *stmt() const		{ return _stmt; }
    pfg_node *node() const		{ return _node; }
    virtual RefKind kind() const = 0;
    virtual boolean isD() const		{ return FALSE; }
    virtual boolean isU() const		{ return FALSE; }
    virtual boolean isPHI() const	{ return FALSE; }
    virtual boolean isPI() const	{ return FALSE; }
    virtual boolean isHSync() const	{ return FALSE; }
    virtual boolean isTSync() const	{ return FALSE; }
    virtual boolean isUSync() const	{ return FALSE; }
    virtual boolean isMutexSync() const	{ return FALSE; }
    virtual boolean isLock() const	{ return FALSE; }
    virtual boolean isUnlock() const	{ return FALSE; }
    virtual boolean isSync() const	{ return FALSE; }

    virtual char *ckind() const = 0;
    RefConflict_list *conflicts() const { return _conflicts; }
    virtual int lineno() const		{ return source_line_num(_stmt); }

    /* chain() is a link from a use of a variable at this reference to the
     * reaching definition or phi-term.
     */
    VarRef *chain() const		{ return _chain; }
    void set_chain(VarRef *def) 	{ _chain = def; }

    /* save_chain() is a temporary placeholder to save the old reaching
     * definition when a new definition or phi-term is reached.
     */
    VarRef *save_chain() const		{ return _save_chain; }
    void set_save_chain(VarRef *def) 	{ _save_chain = def; }

    /* Reaching definition information. Generated by
     * cssame::computeReachingDefs()
     */
    VarRef_list *reachingDefs() const	{ return _rdefs; }
    void set_reachingDefs(VarRef_list *rdefs);

    /* Reached uses information. Generated by cssame::computeReachingDefs() */
    VarRef_list *reachedUses() const	{ return _ruses; }
    void set_reachedUses(VarRef_list *rdefs);

    /* Marks set by cssame::followChain() when computing reaching definitions */
    boolean isMarkedBy(VarRef *u)	{ return _mark == u; }
    void markedBy(VarRef *u)		{ _mark = u; }

    virtual void print(FILE *f=stdout);
    virtual void print_brief(FILE *f=stdout);
};

//---------------------------------------------------------------------------
// class VarDef
//---------------------------------------------------------------------------
class VarDef : public VarRef {
public:
    VarDef(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~VarDef();

    virtual RefKind kind() const	{ return D; }
    virtual char *ckind() const		{ return "D"; }
    virtual boolean isD() const		{ return TRUE; }
};


//---------------------------------------------------------------------------
// class VarUse
//---------------------------------------------------------------------------
class VarUse : public VarRef {
public:
    VarUse(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~VarUse();

    virtual RefKind kind() const	{ return U; }
    virtual char *ckind() const		{ return "U"; }
    virtual boolean isU() const		{ return TRUE; }
};


//---------------------------------------------------------------------------
// class PhiTerm
//---------------------------------------------------------------------------
class PhiTerm : public VarRef {
protected:
    int _lineno;
    RefKind _kind;
    VarRef_list *_phi_chain;

public:
    PhiTerm(var_sym *var, pfg_node *node, RefKind kind = PHI);
    virtual ~PhiTerm();

    /* phi_chain() is a vector of links from a phi-term at this reference
     * to the reaching definitions along each CFG predecessor
     */
    VarRef_list *phi_chain()	{ return _phi_chain; }

    virtual int lineno() const	{ return _lineno; }
    virtual void print(FILE *f=stdout);
    virtual char *ckind() const	{ return (_kind == PHI) ? "Phi" : "Pi"; }
    virtual RefKind kind() const	{ return _kind; }
    virtual boolean isPHI() const	{ return _kind == PHI; }
    virtual boolean isPI() const	{ return _kind == PI; }
};


//---------------------------------------------------------------------------
// class SyncRef
//---------------------------------------------------------------------------
class SyncRef : public VarRef {
protected:
    char *_name;

public:
    SyncRef(var_sym *var, tree_instr *stmt, pfg_node *node);

    char *name() const { return _name; }
    virtual void print(FILE *f=stdout);

    virtual boolean isSync() const	{ return TRUE; }
};

//---------------------------------------------------------------------------
// class HeadSyncRef
//---------------------------------------------------------------------------
class HeadSyncRef : public SyncRef {
protected:
public:
    HeadSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~HeadSyncRef();

    virtual boolean isHSync() const	{ return TRUE; }
    virtual char *ckind() const		{ return "HSync"; }
    virtual RefKind kind() const	{ return HSYNC; }
};


//---------------------------------------------------------------------------
// class TailSyncRef
//---------------------------------------------------------------------------
class TailSyncRef : public SyncRef {
protected:
public:
    TailSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~TailSyncRef();

    virtual boolean isTSync() const	{ return TRUE; }
    virtual char *ckind() const		{ return "TSync"; }
    virtual RefKind kind() const	{ return TSYNC; }
};

//---------------------------------------------------------------------------
// class UndirSyncRef
//---------------------------------------------------------------------------
class UndirSyncRef : public SyncRef {
protected:
public:
    UndirSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node) :
	SyncRef(var, stmt, node) {}

    virtual boolean isUSync() const	{ return TRUE; }
};


//---------------------------------------------------------------------------
// class MutexSyncRef
//---------------------------------------------------------------------------
class MutexSyncRef : public UndirSyncRef {
protected:
public:
    MutexSyncRef(var_sym *var, tree_instr *stmt, pfg_node *node);
    virtual ~MutexSyncRef();

    virtual boolean isMutexSync() const	{ return TRUE; }
    virtual boolean isLock() const	{ return FALSE; }
    virtual boolean isUnlock() const	{ return FALSE; }
    virtual char *ckind() const		{ return "Mutex"; }
    virtual RefKind kind() const	{ return MUTEXSYNC; }
};

//---------------------------------------------------------------------------
// class LockRef
//---------------------------------------------------------------------------
class LockRef : public MutexSyncRef {
protected:
public:
    LockRef(var_sym *var, tree_instr *stmt, pfg_node *node) :
	MutexSyncRef(var, stmt, node) {}
    virtual ~LockRef() {}

    virtual boolean isLock() const	{ return TRUE; }
    virtual char *ckind() const		{ return "Lock"; }
    virtual RefKind kind() const	{ return LOCK; }
};

//---------------------------------------------------------------------------
// class UnlockRef
//---------------------------------------------------------------------------
class UnlockRef : public MutexSyncRef {
protected:
public:
    UnlockRef(var_sym *var, tree_instr *stmt, pfg_node *node) :
	MutexSyncRef(var, stmt, node) {}
    virtual ~UnlockRef() {}

    virtual boolean isUnlock() const	{ return TRUE; }
    virtual char *ckind() const		{ return "Unlock"; }
    virtual RefKind kind() const	{ return UNLOCK; }
};

//---------------------------------------------------------------------------
// class RefConflict
//---------------------------------------------------------------------------
class RefConflict {
protected:
    VarRef *_tail;
    VarRef *_head;

public:
    RefConflict(VarRef *tail, VarRef *head) : _tail(tail), _head(head) {}

    VarRef *tail() const { return _tail; }
    VarRef *head() const { return _head; }

    void print(FILE *f=stdout);
};

#endif	// __VAR_REF_H
