//---------------------------------------------------------------------------
// $Id: cobegin.cc,v 1.6 1998/06/03 16:52:36 diego Exp diego $
//
// Copyright (C) 1997 University of Alberta
//
// Author: Diego Novillo (diego@cs.ualberta.ca)
//
// This software is provided under the terms described in the
// "suif_copyright.h" include file.
//---------------------------------------------------------------------------
#include <suif_copyright.h>
#include <string.h>

#include <suif.h>

#include <odc-util.h>

#include "cobegin.h"

//---------------------------------------------------------------------------
// isCobegin()
// Returns TRUE if the given instruction is a cobegin.
//---------------------------------------------------------------------------
boolean
isCobegin(instruction *instr)
{
    /* NOTE: Should assert that 'findpar' has been executed (find out how) */

    if (instr == NULL) return FALSE;

    return (instr->peek_annote(k_cobegin) != NULL);
}


//---------------------------------------------------------------------------
// isCobegin()
// Returns TRUE if the given tree node is a cobegin. Note: this assumes
// that the program has been processed by 'findpar' already!
//---------------------------------------------------------------------------
boolean
isCobegin(tree_node *tn)
{
    /* NOTE: Should assert that 'findpar' has been executed (find out how) */

    if (tn == NULL) return FALSE;

    if (tn->is_instr()) {
	return isCobegin(((tree_instr *)tn)->instr());
    }

    return FALSE;
}

//---------------------------------------------------------------------------
// tree_cobegin::tree_cobegin()
//---------------------------------------------------------------------------
tree_cobegin::tree_cobegin(tree_node *mbr)
{
    static char *__FNAME = "tree_cobegin::tree_cobegin";

    assert(isCobegin(mbr));

    _mbr = (tree_instr *)mbr;
    _instr = (in_mbr *)(_mbr->instr());
    _numThreads = _instr->num_labs();

    if (debug(__FNAME, 1)) {
	printf("%s(): Building cobegin tree with %d threads\n", __FNAME,
		_numThreads);
    }

    /* Now we build the list of threads that are included in the cobegin.
     * The pass 'findpar' has transformed the original switch statement
     * into an mbr that looks like this:
     *		mbr main.suif_tmp1, low 1, L:main.L3, (L:main.L4, L:main.L5)
     *		lab L:main.L4
     *		...
     *		jmp L:main.L3
     *		lab L:main.L5
     *		...
     *		lab L:main.L3
     *
     * Each thread begins with a unique label and ends with a jump to the
     * default label. We iterate until we find the default label and
     * collect the threads as we go down.
     */
    _threadList = new threadBody *[_numThreads];
    assert(_threadList);

    tree_node_list_iter iter(_mbr->parent());
    unsigned i;
    tree_node *tn;
    label_sym *default_lab = _instr->default_lab();

    /* Skip over the mbr itself. */
    iter.set(_mbr->list_e());
    iter.step();

    for (i = 0; i < _numThreads; i++) {
        boolean endOfThread;
        threadBody *thread; 

        /* We should not run out of nodes in the program */
        assert(iter.is_empty() == FALSE);

	thread = new threadBody(this, i, iter.peek(), iter.peek());
	assert(thread);

	if (debug(__FNAME, 1)) {
	    printf("%s(): Found start of thread #%d at tree node:\n",
		    __FNAME, i);
	    iter.peek()->print();
	}

        endOfThread = FALSE;
        while (endOfThread == FALSE && iter.is_empty() == FALSE) {
            tn = iter.step();
	    thread->extend(tn);
            if (tn->is_instr()) {
                instruction *instr = ((tree_instr *)tn)->instr();

		/* We found the end of this thread if we find a jump to the
		 * default label or if we find the default label itself.
		 * This second test is required because 'porky -ucf-opt'
		 * removes idiotic jumps like:
		 *	jmp L:main.L3
		 *	lab L:main.L3
		 * which is exactly what scc generates for the last case
		 * statement of the switch.
		 */
                if ( (instr->opcode() == io_jmp && ((in_bj *)instr)->target() ==
			default_lab) ||
		     (instr->opcode() == io_lab && ((in_lab *)instr)->label() ==
		        default_lab) ) {
		    endOfThread = TRUE;
		    if (debug(__FNAME, 1)) {
			printf("%s(): Found end of thread #%d at tree node:\n",
				__FNAME, i);
			tn->print();
		    }
		}
            }   
        }   
        _threadList[i] = thread;
    }

    /* The last instruction of the cobegin is the last instruction of the
     * last thread
     */
    assert(tn->is_instr());
    _last = (tree_instr *)tn;
}


//---------------------------------------------------------------------------
// tree_cobegin::thread()
// Returns the ith thread
//---------------------------------------------------------------------------
threadBody *
tree_cobegin::thread(unsigned num) const
{
    assert(num < _numThreads);
    return _threadList[num];
}

//---------------------------------------------------------------------------
// tree_cobegin::print()
//---------------------------------------------------------------------------
void
tree_cobegin::print(FILE *fp, int depth)
{
    unsigned i;

    for (i = 0; i < _numThreads; i++) {
	fprintf(fp, "Thread #%d:\n", i);
	_threadList[i]->print(fp, depth);
	fprintf(fp, "\n");
    }
}

//***************************************************************************
//		    Methods for class threadBody
//***************************************************************************
void
threadBody::print(FILE *fp, int depth)
{
    thread_node_iter iter(this);

    while (iter.is_empty() == FALSE) {
	iter.step()->print(fp, depth);
    }
}


//***************************************************************************
//		    Methods for class thread_node_iter
//***************************************************************************
thread_node_iter::thread_node_iter(threadBody *t, boolean reverse)
{
    _first = t->first()->list_e();
    _last = t->last()->list_e();
    _sentinel = reverse ? _first : _last; 
    _rev = reverse; 
    this->reset(); 
}


//---------------------------------------------------------------------------
// thread_node_iter::reset()
// Reset the iter to the beginning of the basic block.
//---------------------------------------------------------------------------
void
thread_node_iter::reset()
{
    _cur = NULL;
    _nxt = _rev ? _last : _first;
}


//---------------------------------------------------------------------------
// thread_node_iter::step()
// Step the iter to the next instruction.
//---------------------------------------------------------------------------
tree_node *
thread_node_iter::step()
{
    /* Return NULL if at end of thread */
    if (_cur == _sentinel) {
	return NULL;
    }

    _cur = _nxt;
    _nxt = _rev ? _nxt->prev() : _nxt->next();
    return (tree_node *)_cur->contents;
}


//---------------------------------------------------------------------------
// thread_node_iter::peek()
// Look at the next instruction, but don't step the iter.
//---------------------------------------------------------------------------
tree_node *
thread_node_iter::peek()
{
    /* Return NULL if at end of basic block */
    if (_cur == _sentinel) {
	return NULL;
    }

    return (tree_node *)_nxt->contents;
}


//---------------------------------------------------------------------------
// thread_node_iter::is_empty()
// Am I at the end of the instructions in this thread?
//---------------------------------------------------------------------------
boolean
thread_node_iter::is_empty()
{
    return (_cur == _sentinel);
}
